-- can package
--
-- Provides an API to a CAN channel
--
-- API Functions:
--  can:new(devname, firmware_path, master_freq)
--  can:poll()
--  can:ts_offset(ts)
--  can:setup_sync_ts()
--  can:get_sync_ts()
--  can:get_current_ts() [DEBUG]
--  can:txdr_passthru_power_on()
--  can:txdr_passthru_power_off()
--  can:bus_setup(args)
--  can:enable_terminator()
--  can:disable_terminator()
--  can:rx_start(start_time, heartbeat_interval, msgs_per_bd, num_bds)
--  can:rx_stop()
--  can:tx_start(ring_buffer_id, msgs_per_bd, num_bds)
--  can:tx_msg(id, data) [DEBUG]
--  can:allocate_msg_array(num_msgs)
--  can:insert_msg(msg_array, index, ts, id, data)
--  can:tx_append_msgs(ring_buffer_id, msg_array, start_index)
--  can:tx_stop(ring_buffer_id)
--  can:event_setup(args)
--  can:event_start(start_time, heartbeat_interval, events_per_bd, num_bds)
--  can:event_stop()
--
--
-- Notes on tx_start and tx_stop ring_buffer_id parameter.
--   This is an integer from 0 to MAX_TX_STARTS - 1.  The MAX_TX_STARTS
--   value is accessed by the FFI interface: ffi.C.MAX_TX_STARTS
--
--
-- Notes on bus_setup(...) parameters:
--   Set the following parameter names to the required values.
--           baud_rate         (12,500 to 1,000,000)
--           sample_point      (0 to 100)
--           allowable_error   (0 to 1.5)
--           sjw               (1 to 4)
--           tseg1             (1 to 16)
--           tseg2             (1 to 8)
--           samples_per_bit   (1 or 3)
--           listen_only       true or not listed
--
--   This module will calculate missing tseg1/tseg2 values if provided with a
--   sample_point percentage.  If sample_point is also missing, a default of
--   75% is used.  The sjw parameter is defaulted to 1 if missing.
--   If samples_per_bit is missing, it is derived from the baud_rate
--   parameter.  Rates above 250,000 set samples_per_bit to 1, otherwise it
--   is set to 3.  If a setting cannot be found that yields a baud rate within
--   the specified allowable_error parameter, an error is returned.  The default
--   for allowable_error is 0.5, i.e. 0.5%.
--
--   For example: bus_setup({baud_rate = 250000})
--
--
-- Notes on event_setup(...) parameters:
--   Set the following parameter names to true or non-zero to enable.
--       bus_off_deassert
--       bus_off
--
--   For example: event_setup({bus_off = true})
--   All parameters not listed are disabled by default.
--


local cc = require 'connctrl'
local ffi = require 'ffi'
local rb = require 'ring_buffers'
local ch = require 'channel'
local cjson = require 'cjson'
local config = require 'system_config'

local can = setmetatable({}, { __index = ch })
local f = assert(io.open("/usr/include/fpga-firmware/can_structures.h", "r"))
local h = f:read("*all")
f:close()
ffi.cdef(h)

local can_controller_frequency = 40000000
local function generate_timing_set(index)   -- index from 0 to 8191
    -- split index into constituent parts
    local tseg2 = index % 8 + 1
    index = math.floor(index / 8)
    local tseg1 = index % 16 + 1
    local brp = math.floor(index / 16) + 1
    local btq = tseg1 + tseg2 + 1
    local baud_rate = can_controller_frequency / brp / 2 / btq
    local sample_point = (tseg1 + 1) / btq * 100
    return { brp = brp, tseg1 = tseg1, tseg2 = tseg2,
             baud_rate = baud_rate, sample_point = sample_point }
end

local function find_setting(baud_rate, sample_point, allowable_error)
    if not sample_point then sample_point = 75 end
    if not allowable_error then allowable_error = 0.5 end
    local index
    local percent_error = 100
    local final_error
    for i = 0,8191 do
        local set = generate_timing_set(i)
        local err = math.abs((baud_rate - set.baud_rate) / baud_rate * 100)
        if err <= allowable_error then
            if math.abs(sample_point - set.sample_point) < percent_error then
                index = i
                percent_error = math.abs(sample_point - set.sample_point)
                final_error = err
            end
        end
    end
    return index, final_error
end

function can:new(devname, firmware_path, master_freq)
    local can_ch = self:new_channel(devname, firmware_path, "can_cmd_s")
    can_ch.master_freq = master_freq
    can_ch.tx_rb_msgs = {}
    return can_ch
end

function can:ts_offset(ts)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_TS_OFFSET
    cancmd.prm.ts_offset.ts_adjust = ts
    return self:send_command_buffer(cancmd)
end

function can:setup_sync_ts()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_SETUP_SYNC_TS
    return self:send_command_buffer(cancmd)
end

function can:get_sync_ts()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_GET_SYNC_TS
    return self:send_command_buffer(cancmd)
end

function can:get_current_ts()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_GET_TS
    return self:send_command_buffer(cancmd)
end

function can:txdr_passthru_power_on()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_PASSTHRU_POWER_ON
    return self:send_command_buffer(cancmd)
end

function can:txdr_passthru_power_off()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_PASSTHRU_POWER_OFF
    return self:send_command_buffer(cancmd)
end

--[[ CAN Bit Timing Diagram

    |<-------------------- Bit Time ----------------------->|
    |
    | SYNC_ |<---------- TSEG1 ------------>|<--- TSEG2 --->| SYNC_ |
    | SEG   |                               |               | SEG   |
    |---------------------------------------------------------------|
    |       |       |       |       |       |       |       |       |
    |       |       |<----->|       |       |       |       |       |
                        |                   ^               ^
                        |                   |               |
                        |                   |               |
                    1 Clk Cycle       Sample Point    Transmit Point

--]]
function can:bus_setup(args)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_SETUP

    if not args.samples_per_bit then
        if args.baud_rate > 250000 then
            args.samples_per_bit = 1
        else
            args.samples_per_bit = 3
        end
    end

    if not args.allowable_error then args.allowable_error = 0.5 end
    if not args.sample_point then args.sample_point = 75 end
    if not args.sjw then args.sjw = 1 end

    local brp, percent_error
    if not args.tseg1 or not args.tseg2 then
        local index
        index, percent_error = find_setting(args.baud_rate,
                                            args.sample_point,
                                            args.allowable_error)
        if index then
            local set = generate_timing_set(index)
            args.tseg1 = set.tseg1
            args.tseg2 = set.tseg2
            brp = set.brp - 1
        else
            percent_error = 100     -- Should be flagged as an error
        end
    else
        local cycles_per_bit_time = 1 + args.tseg1 + args.tseg2
        local fq = self.master_freq / cycles_per_bit_time / args.baud_rate
        brp = (fq / 2) - 1
        brp = math.floor(brp + 0.5)
        local achieved_baud_rate = self.master_freq / cycles_per_bit_time /
                                   (2 * (brp + 1))
        percent_error = (achieved_baud_rate / args.baud_rate - 1) * 100
    end

    -- Parameter check
    assert(args.baud_rate >= 10000 and args.baud_rate <= 1000000)
    assert(args.samples_per_bit == 1 or args.samples_per_bit == 3)
    assert(args.sample_point > 0 and args.sample_point < 100)
    assert(args.sjw >= 1 and args.sjw <= 4)
    assert(args.tseg1 >= 1 and args.tseg1 <= 16)
    assert(args.tseg2 >= 1 and args.tseg2 <= 8)

    -- Handle 1 Mbps baud_rate explicitly
    -- SJA1000 CAN (Redmine - can_baud.xlsx)
    -- Baud Rate:       1000000 Mbps
    -- CAN Clock Rate: 40000000 Hz
    if args.baud_rate == 1000000 and brp == 0 then
        brp = 1
        args.tseg1 = 7
        args.tseg2 = 2
    end

    if arg.listen_only then cancmd.prm.can_setup.listen_only = 1 end
    cancmd.prm.can_setup.sjw  = args.sjw
    cancmd.prm.can_setup.tseg1 = args.tseg1
    cancmd.prm.can_setup.tseg2 = args.tseg2
    cancmd.prm.can_setup.samples_per_bit = args.samples_per_bit
    cancmd.prm.can_setup.brp = brp
    return self:send_command_buffer(cancmd), percent_error
end

function can:enable_terminator()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_ENABLE_TERMINATOR
    return self:send_command_buffer(cancmd)
end

function can:disable_terminator()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_DISABLE_TERMINATOR
    return self:send_command_buffer(cancmd)
end

function can:rx_start(start_time, heartbeat_interval, msgs_per_bd, num_bds)
    if config.device_type == 'edaqxr' then self:remove_aux_ring_buffers() end
    local pa = self:add_ring_buffer('RX', "can_rxtx_buffer_t["..msgs_per_bd.."]",
                                    "can_rxtx_buffer_t", num_bds)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_RX_START
    cancmd.prm.rx_start.phys_address = pa
    cancmd.prm.rx_start.start_time = start_time
    self:calculate_scg_parameters(cancmd.prm.rx_start.heartbeat_interval,
                                  heartbeat_interval, 1, self.master_freq)
    return self:send_command_buffer(cancmd)
end

function can:rx_stop()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_RX_STOP
    return self:send_command_buffer(cancmd)
end

function can:tx_start()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_TX_START
    return self:send_command_buffer(cancmd)
end

-- Lookup table of numeric query rates to enum values from the
-- enum can_query_rate_t which the FPGA understands.
local query_rate_enums = {}
query_rate_enums[0.1] = ffi.C.QUERY_DEC_0p1HZ
query_rate_enums[0.2] = ffi.C.QUERY_DEC_0p2HZ
query_rate_enums[0.5] = ffi.C.QUERY_DEC_0p5HZ
query_rate_enums[1] = ffi.C.QUERY_DEC_1HZ
query_rate_enums[2] = ffi.C.QUERY_DEC_2HZ
query_rate_enums[5] = ffi.C.QUERY_DEC_5HZ
query_rate_enums[10] = ffi.C.QUERY_DEC_10HZ
query_rate_enums[20] = ffi.C.QUERY_DEC_20HZ
query_rate_enums[50] = ffi.C.QUERY_DEC_50HZ
query_rate_enums[100] = ffi.C.QUERY_DEC_100HZ
query_rate_enums[200] = ffi.C.QUERY_DEC_200HZ

function can:add_query_msgs(queries)

    -- Create a ring buffer to hold commanded Active Query messages.
    local pa = self:add_ring_buffer('QUERY',
                                    "can_query_msg_s[1]",
                                    "can_query_msg_s", 128)

    -- Create a local buffer so we can send it into the ring buffer.
    local msg_array = ffi.new('can_query_msg_s[128]')
    local msg_index = 0

    -- We fill out message structs here and then copy to the array in
    -- order to enforce alignment on ARM (eDAQXR); otherwise we get
    -- alignment errors for elements beyond element 0.
    local this_msg = ffi.new('can_query_msg_s[1]')

    -- Fill in the FPGA command struct for each of our Active Query msgs.
    for _, query in ipairs(queries) do
        local enum_value = query_rate_enums[tonumber(query.rate)]
        if enum_value then
            this_msg[0].sample_rate = enum_value
        else
            -- XXX TODO: Gracefully report error
            this_msg[0].sample_rate = 0xff
        end

        local dlc = (query.msg:len() - 8) / 2
        this_msg[0].can_id = tonumber('0x' .. query.msg:sub(1, 8))
        this_msg[0].data_length = dlc
        for i = 1, dlc do
            local offset = 9 + (i - 1) * 2
            this_msg[0].can_message[i - 1] =
                tonumber('0x' .. query.msg:sub(offset, offset + 1))
        end

        ffi.copy(msg_array[msg_index], this_msg[0],
            ffi.sizeof('can_query_msg_s'))

        msg_index = msg_index + 1

    end

    -- Fill in all unused structs with sample_rate 0xff, which signals
    -- the FPGA that this struct does not contain valid Active Query data.
    while msg_index < 128 do
        msg_array[msg_index].sample_rate = 0xff
        msg_index = msg_index + 1
    end

    -- Send it into the ring buffer for later consumption by the FPGA.
    local ptr = ffi.cast('can_query_msg_s *', msg_array)
    self:send_buffer(ptr, 'QUERY', ffi.sizeof(msg_array))

    -- Send the command to the FPGA to read the ring buffer and set up
    -- its active query magic.
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_ADD_QUERY_MSG
    cancmd.prm.add_query_msg.phys_address = pa
    return self:send_command_buffer(cancmd)

end

function can:clear_query_msgs()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_CLEAR_QUERY_MSG
    return self:send_command_buffer(cancmd)
end

local function fill_in_can_rxtx_structure(rxtx_struct, ts, id, data)
    assert(#data <= 8)
    local ff = 0
    rxtx_struct.can_id = tonumber(id)
    if rxtx_struct.can_id >= 0x80000000 then ff = 128 end
    ff = ff + #data
    for i=1,#data do rxtx_struct.can_message[i - 1] = string.byte(data, i) end
    rxtx_struct.frame_format = ff
    rxtx_struct.ts = ts
end

function can:tx_msg(id, data)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_TX_MSG
    fill_in_can_rxtx_structure(cancmd.prm.tx_msg, 0, id, data)
    return self:send_command_buffer(cancmd)
end

function can:allocate_msg_array(num_msgs)
    return ffi.new("can_rxtx_buffer_t["..num_msgs.."]")
end

function can:insert_msg(msg_array, index, ts, id, data)
    local total_msgs = ffi.sizeof(msg_array) / ffi.sizeof("can_rxtx_buffer_t")
    assert(index > 0 and index <= total_msgs)
    fill_in_can_rxtx_structure(msg_array[index - 1], ts, id, data)
end

function can:tx_append_msgs(ring_buffer_id, msg_array, start_index)
    local rb_name = 'TX'..ring_buffer_id
    local msgs_per_bd = self.tx_rb_msgs[rb_name]
    local total_msgs = ffi.sizeof(msg_array) / ffi.sizeof("can_rxtx_buffer_t")
    local msgs_queued = start_index or 0
    total_msgs = total_msgs - msgs_queued
    while total_msgs > 0 do
        local msgs_in_buffer
        if total_msgs > msgs_per_bd then
            msgs_in_buffer = msgs_per_bd
        else
            msgs_in_buffer = total_msgs
        end
        local buffer_size = msgs_in_buffer * ffi.sizeof("can_rxtx_buffer_t")
        local ptr = ffi.cast("can_rxtx_buffer_t *", msg_array) + msgs_queued
        if self:send_buffer(ptr, rb_name, buffer_size) then
            total_msgs = total_msgs - msgs_in_buffer
            msgs_queued = msgs_queued + msgs_in_buffer
        else
            return msgs_queued
        end
    end
    return msgs_queued
end

function can:tx_stop(ring_buffer_id)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_TX_STOP
    return self:send_command_buffer(cancmd)
end

function can:event_setup(arg)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_EVENT_SETUP
    if arg.bus_off_deassert then cancmd.prm.events.bus_off_deassert = 1 end
    if arg.bus_off then cancmd.prm.events.bus_off = 1 end
    return self:send_command_buffer(cancmd)
end

function can:event_start(start_time, heartbeat_interval,
                         events_per_bd, num_bds)
    local pa = self:add_ring_buffer('EV',
                                    "can_ev_buffer_t["..events_per_bd.."]",
                                    "can_ev_buffer_t", num_bds)
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_EVENT_START
    cancmd.prm.ev_start.phys_address = pa
    cancmd.prm.ev_start.start_time = start_time
    self:calculate_scg_parameters(cancmd.prm.ev_start.heartbeat_interval,
                                  heartbeat_interval, 1, self.master_freq)
    return self:send_command_buffer(cancmd)
end

function can:event_stop()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_EVENT_STOP
    return self:send_command_buffer(cancmd)
end

function can:enable_self_rx()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_ENABLE_SELF_RX
    return self:send_command_buffer(cancmd)
end

function can:disable_self_rx()
    local cancmd = ffi.new("can_cmd_s")
    cancmd.cmd = ffi.C.CAN_DISABLE_SELF_RX
    return self:send_command_buffer(cancmd)
end

return can
