-- edaqctrl package of helper functions for the edaqctrl driver
--
-- API Functions:
--  is_controller_in_reset(devname)
--  controller_reset(devname)
--  release_controller_reset(devname)
--  get_dma_mapping(devname, size)
--  release_dma_mapping(dma_map)
--  get_dma_mapping_physical_address(dma_map)
--  get_dma_mapping_virtual_address(dma_map)
--  close_device(devname)
--  test_for_interrupt(devname)
--  wait_for_interrupt(devname)
--


local u = require 'unix'
local bit = require 'bit'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'

ffi.cdef[[
    int ioctl(int, int, ...);
]]

local lja = {
    map_flush = function(addr, linesize, size) end
}

local edaqctrl = {}

-- Store file descriptor indexed by device name
local open_reg_devices = {}
local open_bar_devices = {}

-- Store map, map and size indexed by file desc
local open_bar_mappings = {}
local open_reg_mappings = {}

-- Store physical address, size, fd, and cache line size indexed by map
local open_dma_mappings = {}

function edaqctrl.dump_internal_tables()
    print("Open Devices")
    print("File", "Device")
    for k,v in pairs(open_devices) do print(v, k) end
    print("Open BAR Mappings")
    print("File", "Map", "", "", "Size")
    for k,v in pairs(open_bar_mappings) do
        print(k, v.map, v.sz)
    end
    print("Open DMA Mappings")
    print("File", "Map", "", "", "Phys Addr", "Size", "Cache Line Size")
    for k,v in pairs(open_dma_mappings) do
        for i, vv in ipairs(v) do
            print(v.fd, k, i,
                string.format("0x%08X", vv.pa), vv.sz, vv.cachelinesize)
        end
    end
end

function edaqctrl.dump_trace_buffer(devname)
end

function edaqctrl.check_for_exception(devname)
    if edaqctrl.is_controller_halted(devname) then
        local info = string.format("Controller %s has halted.\r", devname)
--        info = info .. edaqctrl.dump_exception_info_s(devname)
        edaqctrl.controller_reset(devname)
        error(info)
    end
end

local function get_fd(devname)
    local fd = open_bar_devices[devname]
    if not fd then
        fd = assert(u.open(devname,
            u.O_RDWR + (devname == '/dev/mem' and u.O_SYNC or 0)))
        open_bar_devices[devname] = fd
    end
    return fd
end

function edaqctrl.controller_reset(devname)
    local fd = get_fd(devname)
    if fd then
        local v = ffi.new('uint32_t[1]', {1})
        C.ioctl(fd, 3, v)
    end
end

function edaqctrl.set_response_bd_ptr(devname, pa)
    local fd = get_fd(devname)
    if fd then
        local v = ffi.new('unsigned long[1]', {pa})
        C.ioctl(fd, 4, v);
    end
end

function edaqctrl.release_controller_reset(devname)
    local fd = get_fd(devname)
    if fd then
        local v = ffi.new('uint32_t[1]', {0})
        C.ioctl(fd, 3, v)
    end
end

function edaqctrl.get_dma_mapping(devname, size, offset)
    local fd = get_fd('/dev/mem')
    if not fd then
        log:log(log.ERR, "Uaable to open /dev/mem")
        return
    end
    local map, pa, va
    if not offset or offset == 0 then
        offset = 0
        local _fd = get_fd(devname)
        local p = ffi.new('unsigned long[1]')
        if C.ioctl(_fd, 5, p) < 0 then
            C.close(_fd)
            log:log(log.ERR, "Error retrieving physical address")
            return
        end
        pa = p[0]
        map, va = assert(u.mmap(4096 * 1024, fd, pa))
        open_dma_mappings[devname] = {{pa=pa, sz=size, fd=fd, va=va,
                                    cachelinesize=64}}
    elseif open_dma_mappings[devname] then
        pa = open_dma_mappings[devname][1].pa
        va = open_dma_mappings[devname][1].va
        for _, v in ipairs(open_dma_mappings[devname]) do
            pa = pa + v.sz
            va = va + v.sz
        end
        open_dma_mappings[devname][#open_dma_mappings[devname] + 1] = {
            pa = pa,
            sz = size,
            fd = fd,
            va = va,
            cachelinesize = open_dma_mappings[devname][1].cachelinesize
        }
        map = open_dma_mappings[devname][#open_dma_mappings[devname]]

        -- Sanity check to make sure we haven't gone past the end of the
        -- mapped region
        local tot_size = 0
        local entries = 0
        for k, v in pairs(open_dma_mappings[devname]) do
            tot_size = tot_size + v.sz
            entries = entries + 1
        end
        if entries ~= #open_dma_mappings[devname] then
            log:log(log.ERR, string.format(
                "DMA mapping table size inconsistancy"))
            return
        end
        if tot_size >= (4096 * 1024) then
            log:log(log.ERR, string.format(
                "Total allocated size is larger than region size"))
            return
        end
    else
        log:log(log.ERR, string.format("No open DMA mapping for %s", devname))
        return
    end

    -- Error if any of the items are not available
    if not map or not pa or not va then
        log:log(log.ERR, string.format("Unable to map %d bytes at offset %d",
            size, offset))
    else
        ffi.fill(ffi.cast('void *', va), size, 0)
    end

    return map, pa, va
end

function edaqctrl.flush_dma_region_by_offset(map, length, offset)
end

function edaqctrl.flush_dma_region_by_ptr(map, ptr, length)
end

function edaqctrl.flush_entire_dma_region(map)
end

function edaqctrl.release_dma_mapping(dma_map)
    for k, v in pairs(open_dma_mappings) do
        for i, vv in ipairs(v) do
            if vv == dma_map then
                table.remove(open_dma_mappings[k], i)
                return
            end
        end
    end
end

function edaqctrl.wait_for_interrupt(devname)
    local fd = get_fd(devname)

    u.read(fd, 1)
end

function edaqctrl.clear_interrupts(devname)
    local fd = assert(u.open(devname, u.O_RDWR + u.O_NONBLOCK))

    while true do
        local buf, es, ec = u.read(fd, 1)

        if ec == 11 then
            u.close(fd)
            return
        end

        if not buf then
            error(es)
        end
    end
end

return edaqctrl
