require('lj_include'){
    'unistd',
    'linux/time',
    'sys/time',
}

local armada = require 'armada'
local ffi = require 'ffi'
local s = require 'struct'
local ts = require 'timestamps'
local u = require 'unix'
local ev = require 'ev'
local config = require 'system_config'
local status_lib = require 'status_lib'
local timesetter = require "timesetter"

local cjson = require 'cjson.safe'
local sl = status_lib:new('gps')

local control_status = {state = 'Wait'}

local C = ffi.C

local sample_clock = require 'sample_clock'
local gps = require 'gps'
local gpschar = require 'gpschar'
local membuf = require 'membuf'
local nmea_parser = require 'nmea_parser'

local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'

local dbus = require 'dbus'
local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

local __ = require '__'
local log = require 'log'


-- Function used to retrieve preferences for regulation of Si514 by GPS' PPS
-- Any failure returns values that cause the Si514 to be regulated by the PPS
-- signal of the GPS and the system/PTP time is NOT synchronized to GPS time
local function get_oscillator_control()
    if config.gps_timing_regulation then
        if config.gps_timing_regulation == 'none' then
            return false, true
        elseif config.gps_timing_regulation == 'regulate_only' then
            return false, false
        elseif config.gps_timing_regulation == 'sync_to_gps_time' then
            return true, false
        end
    end
    return false, false
end

local time_synced_to_gps, no_gps_regulation_of_oscillator =
    get_oscillator_control()

local function dbg(fmt, ...)
    log:log(log.DEBUG, fmt, ...)
end

----------------------------------------------------------------------

-- This function returns an array of C structures containing the
--     offsets and field lengths specifying the fields in the given
--     GPS input buffer.  The input fields are comma separated.
-- Parameter 'in_buf' is a membuf object containing the GPS input.
local function identify_fields(in_buf)
    local field_descrs = {}
    local offset = 0
    local str = ffi.string(in_buf.buf, #in_buf)
    for f in str:gfind('(.-)[,%*]') do
        field_descrs[#field_descrs + 1] = f
    end

    return field_descrs
end

-- This function returns an index into the input array for the
--     start of the suffix substring ".0", or nil if such is
--     not found.
-- Parameter utc_date_time is a membuf object
local function find_trailing_0(utc_date_time)
    if utc_date_time then
        local s, e = utc_date_time:find('%.0')
        if s then s = s - 1 end
        return s
    end
end

local function fix_date(str)
    return '20' .. str:sub(5, 6) .. '-' .. str:sub(3, 4) .. '-' .. str:sub(1, 2)
end

local function fix_time(str)
    return str:sub(1, 2) .. ':' .. str:sub(3, 4) .. ':' .. str:sub(5)
end

-- Returns string containing date followed by time
local function fix_date_time(fdd, fdt)
    local d, t = fix_date(fdd), fix_time(fdt)
    if d and t then return d .. 'T' .. t end
end

-- Returns latitude degrees.  Minutes are represented as decimal
-- fraction part of returned Lua number.  Degrees are negative if
-- parameter 'ns' equals "S".
local function fix_lat(str, ns_str)
    local degrees, minutes = str:match("^(%d%d)(.+)")
    degrees = tonumber(degrees) or 0
    minutes = tonumber(minutes) or 0

    return (degrees + (minutes / 60.0)) * (ns_str == 'S' and -1.0 or 1.0)
end

-- Returns longitude degrees.  Minutes are represented as decimal
-- fraction part of returned Lua number.  Degrees are negative if
-- parameter 'ew' equals "W".
local function fix_lon(str, ew_str)
    local degrees, minutes = str:match("^(%d%d%d)(.+)")
    degrees = tonumber(degrees) or 0
    minutes = tonumber(minutes) or 0

    return (degrees + (minutes / 60.0)) * (ew_str == 'W' and -1.0 or 1.0)
end

local function fix_mv(str, ew)
    if not tonumber(str) then return 0 end
    return tonumber(str) * (ew == 'E' and -1.0 or 1.0)
end

local channel_specs
local gps_proxy = class('gps_proxy', channel_proxy)

function gps_proxy:set_channel_specs()
    local t = self:get_supported_channels()
    channel_specs = {}
    if not self.is_egps then
        for k, v in pairs(t.models["Garmin GPS"].channels) do
            if not v.name then v.name = k end
            channel_specs[k] = v
            channel_specs[k].gps_type = "garmin"
        end
    else
        for k, v in pairs(t.models["EGPS-200-Plus"].channels) do
            if not v.name then v.name = k end
            channel_specs[k] = v
            channel_specs[k].gps_type = "egps200"
        end
    end
end

function gps_proxy:initialize(name, devname)

    self.proxy_type = 'gps'
    self.clk_frequency = sample_clock.master_rate

    cca_handler.new(devname, self, gps, 'GPS_0.bin', self.clk_frequency)
    self:cmd_block('txdr_power_off')

    self.nmea = nmea_parser:new()
    self.channels = {}
    self.external_names = {}
    self.internal_names = {}
    self.cur_sentence_ts = nil
    self.cur_sentence = ''
    self.heartbeat_text = string.format("#hb%c%c", 13, 10)
    self.pps_text = string.format("#pps%c%c", 13, 10)
    self.servo_ts = 0
    self.pps_ts = 0
    self.pps_ts_array = {}
    self.pps_ts_array2 = config.gps_pps_channels and {}
    self.oscillator_control_state = 'none'
    self.state = 'gps_pps_init'
    self.time_set = true

    -- Oscillator Control
    self.osc_collecting = false
    self.osc_pps_count = 0
    self.osc_servo = false
    self.osc_control = false
    self.osc_servo_ts = 0
    self.osc_pps_ts = 0
    self.osc_offset = 0

    self.sinks = {}

    -- GPS sync stuff
    self.pps_events = 0
    self.last_offset = 1e30
    self.dt = .2
    self.Kp, self.Ki, self.Kd = 37, 10, 4
    self.integral = 0

    self.pps_rate = math.ceil(1 / self.dt)
    self.counts_per_pps = math.floor(self.clk_frequency / self.pps_rate)

    self.pps_interval = self.clk_frequency / self.pps_rate
    self.gprmc_channels_good = false -- KLUDGE for Garmin

    if config.debug_files then
        self.max_regulations = 1000
        self.num_regulations = 0
        self.regulation_skip_count = 0
        self.handle = assert(io.open('/tmp/gps_regulation.txt', 'w+'))
        self.handle:write('count\toffset\toutput\n')
    end

    self.filters = {}

    channel_proxy.initialize(self, name)

    if config.device_type ~= 'edaqxr' then
        local cs = self:cmd_block('get_cable_detect_state')
        if cs.channel_specific[0] ~= 0 then
            -- If a cable is connected make the default set of
            -- channels available and turn on 5v power.
            self:cmd_block('txdr_power_on_5v')
        end
    else
        -- There is no cable detect on the eDAQXR
        self:cmd_block('txdr_power_on_5v')
    end
    self:_available_channels()

    -- An event loop is required to complete the stuff in this function so we
    -- defer these items until the event loop is up and going.
    self.init_completion = ev.Idle.new(function (loop, idle, revents)
            -- At this point we need to start up the PPS events
            -- so we can get GPS sync going (if we have GPS lock).
            idle:stop(loop)
            if config.device_type ~= 'edaqxr' then
                self:cmd('event_setup', {
                    cable_unplugged = true,
                    cable_plugged = true,
                    start_of_break = true,
                    end_of_break = true,
                    pps_falling = true,
                    pps_rising = true,
                    parity_error = true,
                    frame_error = true,
                })

                local now_40m = ts.ts_to_stamp(ts.get_ts())
                self:cmd('uart_rx_start',
                    now_40m + self.clk_frequency, 2, 16, 512, '8N1', 38400)
                self:cmd('get_cable_detect_state')
                self:cmd('event_start',
                    now_40m + self.clk_frequency, 2, 1, 128)
            else
                self:cmd('event_setup', {
                    start_of_break = true,
                    end_of_break = true,
                    pps_falling = true,
                    pps_rising = true,
                    parity_error = true,
                    frame_error = true,
                })
                self:cmd('uart_rx_start', 0, 2, 16, 512, '8N1', 38400)
                self:cmd('event_start', 0, 2, 1, 128)
            end
        end)
    self.init_completion:start(ev.Loop.default)

    self.watch_loop = ev.Timer.new(function(l,t,e)
            -- FIXME:  eDAQXR need to communicate PPS rate to driver!
            if self.is_egps and not self.is_egps_set then
                -- Switching to EGPS-200

                self:cmd('uart_rx_stop')
                self:cmd('event_stop')
                self:cmd('txdr_power_on_12v')
                local now_40m = ts.ts_to_stamp(ts.get_ts())

                -- Event stop turns these off so they need to be turned back on
                self:cmd('event_setup', {
                    start_of_break = true,
                    end_of_break = true,
                    pps_falling = true,
                    pps_rising = true,
                    parity_error = true,
                    frame_error = true,
                })

                self:cmd('event_start',
                    now_40m + self.clk_frequency, 2, 1, 128)
                self:cmd('uart_rx_start',
                    now_40m + self.clk_frequency, 10, 8, 512, '8N1', 460800)
                self.is_egps_set = true
                self.pps_rate = 1
                self.gprmc_channels_good = true
                self:_available_channels()
                self.dt = math.ceil(1 / self.pps_rate)
                self.counts_per_pps =
                    math.floor(self.clk_frequency / self.pps_rate)
            elseif not self.is_egps and self.is_egps_set then
                -- Switching back to Garmin
                self:cmd('uart_rx_stop')
                self:cmd('event_stop')
                self:cmd('txdr_power_on_5v')

                -- Event stop turns these off so they need to be turned back on
                self:cmd('event_setup', {
                    cable_unplugged = true,
                    cable_plugged = true,
                    start_of_break = true,
                    end_of_break = true,
                    pps_falling = true,
                    pps_rising = true,
                    parity_error = true,
                    frame_error = true,
                })

                local now_40m = ts.ts_to_stamp(ts.get_ts())
                self:cmd('event_start',
                    now_40m + self.clk_frequency, 2, 1, 128)
                self:cmd('uart_rx_start',
                    now_40m + self.clk_frequency, 2, 16, 512, '8N1', 38400)
                self.is_egps_set = nil
                self.pps_rate = 5
                self.gprmc_channels_good = false
                self:_available_channels()
                self.dt = math.ceil(1 / self.pps_rate)
                self.counts_per_pps =
                    math.floor(self.clk_frequency / self.pps_rate)
            end
        end, .5, .5)
    self.watch_loop:start(ev.Loop.default)

    -- We need the servo() calls to not happen within the cca reader context
    -- because DBus (via SignalHeartbeat()) will screw us over horribly by
    -- nesting calls to gps_proxy:buffer(). So, we use a pipe as a queue for
    -- the offsets, thereby avoiding any libev recursion.
    self.servo_offset_buffer = ffi.new('double[1]')
    self.servo_pipe = ffi.new('int[2]')
    assert(C.pipe(self.servo_pipe) == 0)
    self.servo_w = ev.IO.new(self.servo_watcher_fn, self.servo_pipe[0], ev.READ)
    self.servo_w.proxy = self
    self.servo_w:start(ev.Loop.default)
end

function gps_proxy:data_unavailable()
    if self.md_sink then
        self.md_sink{
            error_message =
                "GPS proxy has stopped receiving data from hardware"
        }
    end
    os.exit(-13)
end

function gps_proxy:get_supported_channels(params, cb)
    local jsonBlob = {
        models = {
            ["Garmin GPS"] = {
                channels = {
                    utc_date_time={
                        description="UTC date and time of position fix",
                        data_type = "message_raw",
                        category = "source/hbm_" .. config.device_type ..
                            "/msg_gps",
                        message_format = "text/plain"},
                    geoidal_height={
                        description="Height of geoid above WGS84",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand="Position",
                        electrical_units="m",
                        range_min=-1000,
                        range_max=1000,
                        invalid_data_output_value=1001},
                    fix_quality={
                        description="GPS positional fix quality ($GPGGA)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        --  0=>"Fix not available"
                        --  1=>"Non-differential GPS fix available"
                        --  2=>"Differential GPS (WAAS) fix available"
                        --  6=>"Estimated"
                        -- NOTE:  considered valid unless == 0
                        range_min = 0,
                        range_max = 32,
                        invalid_data_output_value = 255},
                    fix_type = {
                        description = "GPS positional fix type ($GPGSA)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        --  1 => "not available"
                        --  2 => "2D"
                        --  3 => "3D"
                        range_min = 0,
                        range_max = 3,
                        invalid_data_output_value = 255},
                    ground_heading = {
                        description = "Course over ground",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = 0,
                        range_max = 360,
                        invalid_data_output_value = 361},
                      ground_speed_knots = {
                        description = "Speed over ground in knots",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Velocity",
                        electrical_units = "knots",
                        range_min = 0,
                        range_max = 1000,
                        invalid_data_output_value = 1001},
                    ground_speed_mph = {
                        description = "Speed over ground in mph",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Velocity",
                        electrical_units = "mph",
                        range_min = 0,
                        range_max = 1152,
                        invalid_data_output_value = 1153},
                    altitude = {
                        description = "Altitude relative to mean sea level",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Position",
                        electrical_units = "m",
                        range_min = -10000,
                        range_max = 100000,
                        invalid_data_output_value = 100001},
                    horizontal_dilution_of_precision = {
                        description = "Horizontal dilution of precision",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Precision",
                        range_min = 0,
                        range_max = 100,
                        invalid_data_output_value = 101},
                    latitude = {
                        description = "Latitude",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    longitude = {
                        description = "Longitude",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -180,
                        range_max = 180,
                        invalid_data_output_value = 181},
                    magnetic_variation = {
                        description = "Magnetic variation",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        electrical_units = "degrees",
                        range_min = -180,
                        range_max = 180,
                        invalid_data_output_value = 181},
                    number_of_satellites_in_use = {
                        description = "Number of satellites in use",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        electrical_units = "satellites",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 255},
                    position_dilution_of_precision = {
                        description = "Position dilution of precision",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Precision",
                        range_min = 0,
                        range_max = 100,
                        invalid_data_output_value = 101},
                    utc_seconds = {
                        description = "UTC seconds (fractional)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "seconds",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_minute = {
                        description = "UTC minute",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "minute",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_hour = {
                        description = "UTC hour",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "hour",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 25},
                    utc_day = {
                        description = "UTC day",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "day",
                        range_min = 0,
                        range_max = 31,
                        invalid_data_output_value = 32},
                    utc_month = {
                        description = "UTC month",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "month",
                        range_min = 0,
                        range_max = 12,
                        invalid_data_output_value = 13},
                    utc_year = {
                        description = "UTC year",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Time",
                        electrical_units = "year",
                        range_min = 2000,
                        range_max = 2100,
                        invalid_data_output_value = 2101},
                    vertical_dilution_of_precision = {
                        description = "Vertical dilution of precision",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "Precision",
                        range_min = 0,
                        range_max = 100,
                        invalid_data_output_value = 101},
                    pps_sync_active = {
                        description="PPS time stamp recalculation active",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 5,
                            hbm_legacy = 5,
                            binary = 6.4},
                        measurand = "",
                        range_min = 0,
                        range_max = 1,
                        invalid_data_output_value = 2},
                },
            },
            ["EGPS-200-Plus"] = {
                channels = {
                    utc_date_time={
                        description="UTC date and time of position fix",
                        data_type = "message_raw",
                        category = "source/hbm_" .. config.device_type ..
                            "/msg_gps",
                        message_format = "text/plain"},
                    fix_quality={
                        description="Fix Quality (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        --  0=>"invalid"
                        --  1=>"GPS fix (SPS)"
                        measurand="Quality",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 32,
                        invalid_data_output_value = 255},
                    heading = {
                        description = "Heading (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = 0,
                        range_max = 360,
                        invalid_data_output_value = 361},
                    speed_3d = {
                        description = "3D Speed (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    altitude = {
                        description = "Altitude relative to mean sea level (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "m",
                        range_min = -10000,
                        range_max = 100000,
                        invalid_data_output_value = 100001},
                    latitude = {
                        description = "Latitude (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    longitude = {
                        description = "Longitude (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -180,
                        range_max = 180,
                        invalid_data_output_value = 181},
                    number_of_satellites_in_use = {
                        description = "Number of satellites in use (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        electrical_units = "satellites",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 255},
                    utc_seconds = {
                        description = "UTC seconds",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "seconds",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_minute = {
                        description = "UTC minute",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "minute",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_hour = {
                        description = "UTC hour",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "hour",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 25},
                    utc_day = {
                        description = "UTC day",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "day",
                        range_min = 0,
                        range_max = 31,
                        invalid_data_output_value = 32},
                    utc_month = {
                        description = "UTC month",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "month",
                        range_min = 0,
                        range_max = 12,
                        invalid_data_output_value = 13},
                    utc_year = {
                        description = "UTC year",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "year",
                        range_min = 2000,
                        range_max = 2100,
                        invalid_data_output_value = 2101},
                    speed_raw3d = {
                        description = "3D Raw Speed (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    speed_raw2d = {
                        description = "2D Raw Speed (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    accel_x = {
                        description = "X Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    accel_y = {
                        description = "Y Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    accel_z = {
                        description = "Z Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    yaw_rate = {
                        description = "Yaw Rate (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Rotation rate",
                        electrical_units = "degrees/s",
                        range_min = -300,
                        range_max = 300,
                        invalid_data_output_value = 301},
                    pitch_rate = {
                        description = "Pitch Rate (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Rotation rate",
                        electrical_units = "degrees/s",
                        range_min = -300,
                        range_max = 300,
                        invalid_data_output_value = 301},
                    roll_rate = {
                        description = "Roll Rate (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Rotation rate",
                        electrical_units = "degrees/s",
                        range_min = -300,
                        range_max = 300,
                        invalid_data_output_value = 301},
                    rtk_yaw = {
                        description = "RTK Yaw (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = 0,
                        range_max = 360,
                        invalid_data_output_value = 361},
                    rtk_pitch = {
                        description = "RTK Pitch/Roll (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    rtk_baseline = {
                        description = "RTK Baseline (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Length",
                        electrical_units = "mm",
                        range_min = 0,
                        range_max = 1600,
                        invalid_data_output_value = 1601},
                    rtk_accuracy = {
                        description = "RTK Accuracy (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Length",
                        electrical_units = "mm",
                        range_min = 0,
                        range_max = 1000,
                        invalid_data_output_value = 1001},
                    faa_mode = {
                        description = "FAA Mode (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Quality",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 127,
                        invalid_data_output_value = 255},
                    gradient = {
                        description = "Gradient (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    trigger = {
                        description = "Trigger (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Logical",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 1,
                        invalid_data_output_value = 2},
                    pps_sync_active = {
                        description="PPS time stamp recalculation active",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "",
                        range_min = 0,
                        range_max = 1,
                        invalid_data_output_value = 2},
                }
            },
            ["EGPS-200"] = {
                channels = {
                    utc_date_time={
                        description="UTC date and time of position fix",
                        data_type = "message_raw",
                        category = "source/hbm_" .. config.device_type ..
                            "/msg_gps",
                        message_format = "text/plain"},
                    fix_quality={
                        description="Fix Quality (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        --  0=>"invalid"
                        --  1=>"GPS fix (SPS)"
                        measurand="Quality",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 32,
                        invalid_data_output_value = 255},
                    heading = {
                        description = "Heading (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = 0,
                        range_max = 360,
                        invalid_data_output_value = 361},
                    speed_3d = {
                        description = "3D Speed (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    altitude = {
                        description = "Altitude relative to mean sea level (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "m",
                        range_min = -10000,
                        range_max = 100000,
                        invalid_data_output_value = 100001},
                    latitude = {
                        description = "Latitude (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    longitude = {
                        description = "Longitude (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Position",
                        electrical_units = "degrees",
                        range_min = -180,
                        range_max = 180,
                        invalid_data_output_value = 181},
                    number_of_satellites_in_use = {
                        description = "Number of satellites in use (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        electrical_units = "satellites",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 255},
                    utc_seconds = {
                        description = "UTC seconds",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "seconds",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_minute = {
                        description = "UTC minute",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "minute",
                        range_min = 0,
                        range_max = 60,
                        invalid_data_output_value = 61},
                    utc_hour = {
                        description = "UTC hour",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Time",
                        electrical_units = "hour",
                        range_min = 0,
                        range_max = 24,
                        invalid_data_output_value = 25},
                    utc_day = {
                        description = "UTC day",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "day",
                        range_min = 0,
                        range_max = 31,
                        invalid_data_output_value = 32},
                    utc_month = {
                        description = "UTC month",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "month",
                        range_min = 0,
                        range_max = 12,
                        invalid_data_output_value = 13},
                    utc_year = {
                        description = "UTC year",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Time",
                        electrical_units = "year",
                        range_min = 2000,
                        range_max = 2100,
                        invalid_data_output_value = 2101},
                    speed_raw3d = {
                        description = "3D Raw Speed (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    speed_raw2d = {
                        description = "2D Raw Speed (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Velocity",
                        electrical_units = "kph",
                        range_min = 0,
                        range_max = 1852,
                        invalid_data_output_value = 2000},
                    accel_x = {
                        description = "X Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    accel_y = {
                        description = "Y Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    accel_z = {
                        description = "Z Acceleration (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Acceleration",
                        electrical_units = "g",
                        range_min = -6,
                        range_max = 6,
                        invalid_data_output_value = 10},
                    faa_mode = {
                        description = "FAA Mode (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Quality",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 127,
                        invalid_data_output_value = 255},
                    gradient = {
                        description = "Gradient (20Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "Orientation",
                        electrical_units = "degrees",
                        range_min = -90,
                        range_max = 90,
                        invalid_data_output_value = 91},
                    trigger = {
                        description = "Trigger (200Hz)",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 200,
                            hbm_legacy = 300,
                            binary = 204.8},
                        measurand = "Logical",
                        electrical_units = "",
                        range_min = 0,
                        range_max = 1,
                        invalid_data_output_value = 2},
                    pps_sync_active = {
                        description="PPS time stamp recalculation active",
                        data_type = "message_float64",
                        default_sample_rate = {
                            decimal = 20,
                            hbm_legacy = 25,
                            binary = 25.6},
                        measurand = "",
                        range_min = 0,
                        range_max = 1,
                        invalid_data_output_value = 2},
                }
            }
        }
    }

    -- Lua has trouble digesting the keys "ground_speed_km/h" and
    -- "ground_speed_m/s" when used in the type of initialization
    -- used above for the rest of jsonBlob.  Hence the following:
    jsonBlob.models["Garmin GPS"].channels["ground_speed_km/h"] = {
        description = "Speed over ground in km/h",
        data_type = "message_float64",
        default_sample_rate = {
            decimal = 5,
            hbm_legacy = 5,
            binary = 6.4},
        measurand = "Velocity",
        electrical_units = "km/h",
        range_min = 0,
        range_max = 1852,
        invalid_data_output_value = 1853}

    jsonBlob.models["Garmin GPS"].channels["ground_speed_m/s"] = {
        description = "Speed over ground in m/s",
        data_type = "message_float64",
        default_sample_rate = {
            decimal = 5,
            hbm_legacy = 5,
            binary = 6.4},
        measurand = "Velocity",
        electrical_units = "m/s",
        range_min = 0,
        range_max = 515,
        invalid_data_output_value = 516}

    if not cb then return jsonBlob end
    cb {
        cjson.encode(jsonBlob)
    }
end

function gps_proxy:queue_servo(offset, override_state)
    if self.state == 'gps_pps_locked' or override_state then
        control_status.offset = offset
        sl:set_status(control_status)
        self.servo_offset_buffer[0] = offset
        local len = ffi.sizeof('double')
        assert(C.write(
            self.servo_pipe[1], self.servo_offset_buffer, len) == len)
    end
end

function gps_proxy.servo_watcher_fn(loop, w, revents)
    local self = w.proxy
    local len = ffi.sizeof('double')
    assert(C.read(self.servo_pipe[0], self.servo_offset_buffer, len) == len)
    self:servo(self.servo_offset_buffer[0])
end

function gps_proxy:debug_stop_check()
    if self.handle then
        if self.num_regulations == self.max_regulations then
            self.handle:close()
            self.handle = nil
        end
    end
end

function gps_proxy:write_debug(offset, output)
    if self.handle then
        self.num_regulations = self.num_regulations + 1
        self.handle:write(string.format("%s\t%s\t%s\n", self.num_regulations,
            offset, output))
    end
end

function gps_proxy:do_skip(offset)
    if self.handle then
        if self.regulation_skip_count > 0 then
            self.regulation_skip_count = self.regulation_skip_count - 1
            self.handle:write(string.format("%s\t%s\t0\n",
                -self.regulation_skip_count, offset))
            return true
        end
    end
end

function gps_proxy:time_synced_servo(offset)
    -- Always send the heartbeat if we have GPS lock.  This indicates to others
    -- that we have GPS sync regardless of whether or not the clock is being
    -- adjusted.
--dbg(string.format("Offset: %f (ns)\n", offset * (1000000000 / self.clk_frequency)))
    if math.abs(offset) > 10000 then
        armada.timing:AdjustClockRate(0, 'GPS')
        local ns = offset * (1000000000 / self.clk_frequency)
        local s, ns = math.floor(ns / 1000000000), math.floor(ns % 1000000000)
        armada.timing:AdjustClock(s, ns, 'GPS')
        self.integral = 0
    else
        if self:do_skip(offset) then return end
        if self.last_offset > 1e29 then
            self.last_offset = offset
        end
        self:debug_stop_check()
        local previous_error = self.last_offset
        local error = offset
        self.integral = self.integral + (error * self.dt)
        local derivative = (error - previous_error) / self.dt
        local output = (self.Kp * error) + (self.Ki * self.integral) +
                       (self.Kd * derivative)
        armada.timing:AdjustClockRate(output, 'GPS')
        self.last_offset = offset
        self:write_debug(offset, output)
    end
end

function gps_proxy:time_not_synced_servo(offset)
    -- Always send the heartbeat if we have GPS lock.  This indicates to others
    -- that we have GPS sync regardless of whether or not the clock is being
    -- adjusted.

    if self:do_skip(offset) then return end

    if self.last_offset > 1e29 then
        self.last_offset = offset
    end
    self:debug_stop_check()
    local previous_error = self.last_offset
    local error = offset
    self.integral = self.integral + (error * self.dt)
    local derivative = (error - previous_error) / self.dt
    local output = (self.Kp * error) + (self.Ki * self.integral) +
                   (self.Kd * derivative)
    armada.timing:AdjustClockRate(output, 'GPS')
    self.last_offset = offset
    self:write_debug(offset, output)
end

local function set_gps_regulation()
    if not no_gps_regulation_of_oscillator then
        if time_synced_to_gps then
            gps_proxy.servo = gps_proxy.time_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled with Linux Time Sync")
        else
            gps_proxy.servo = gps_proxy.time_not_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled")
        end
    else
        -- If no regulation of the oscillator via the GPS is desired make the
        -- servo do nothing
        gps_proxy.servo = function(self, offset) self.last_offset = offset end
        log:notice(__"GPS: Si514 PPS Servo Disabled")
    end
end

set_gps_regulation()

local start_stamp

-- This function is intended to synthesis data for the PPS offset channel when
-- the PPS signal drops out.  That can happen at any time when using an EGPS200
-- unit.  Note that the pps_offset and pps_timestamps channels are debugging
-- only channels I use to check oscillator regulation (both GPS and PTP).
-- This function was added simply to supply "fake" data for the dropout periods
-- so tests can be stopped without a "forced" stop occurring if the aggregate
-- amount of time the dropouts occurred exceeds the "forced" stop timeout.
function gps_proxy:fixup_pps_array2()
    -- Only do this if the PPS channels are enabled
    if self.pps_ts_array2 then
        -- Remove the current PPS TS from the end of the array
        table.remove(self.pps_ts_array2)

        -- If there is data left in the array
        if #self.pps_ts_array2 > 0 then
            local range = tonumber(
                self.pps_ts - self.pps_ts_array2[#self.pps_ts_array2])
            local secs = math.ceil(range / 40000000.0)
            local ts = tonumber(self.pps_ts_array2[#self.pps_ts_array2])
            if secs == 2 then
                local addend = math.floor((tonumber(self.pps_ts) - ts) / 2)
                ts = ts + addend
                table.insert(self.pps_ts_array2, ts)
            elseif secs > 0 then
                local addend = math.floor(range / secs)
                ts = ts + addend
                while ts < (self.pps_ts - 20000000) do
                    table.insert(self.pps_ts_array2, ts)
                    ts = ts + addend
                end
            end
            table.insert(self.pps_ts_array2, self.pps_ts)
        end
    end
end

-- FIXME:  The function below MAY need changes to work properly using an
-- EGPS200.
function gps_proxy:buffer(msg)
    if self.data_monitor:is_active() then
        self.data_monitor:again(ev.Loop.default)
    end
    local element_count = msg.element_count - 1
    for j = 0, element_count do
        -- OK so the heartbeat indicator is on the buffer descriptor not on
        -- individual elements within the buffer.  If a heartbeat is indicated
        -- it will ALWAYS be the last element within the buffer.
        if msg.heartbeat and j == element_count then
            local heartbeat = tonumber(msg.bufdata[j].ts)
            local k = self.external_names['']
            for _, sink in pairs(self.sinks[k] or {}) do
                sink(heartbeat, k, self.heartbeat_text, #self.heartbeat_text)
            end
        else
            if msg.cmd == C.GPS_UART_RX_START then
                self.cur_sentence = ffi.string(msg.bufdata[j].rxdata,
                    msg.bufdata[j].count)
                self.cur_sentence_ts = msg.bufdata[j].ts
                self:sentence()
            elseif msg.cmd == C.GPS_EVENT_START then
                local buffer = msg.bufdata[j]
                local events = buffer.events
                local pps_event = false

                -- The eDAQXR sends a "time zero" event that provides us with
                -- the timestamp of when the oscillator started running.
                -- The "time zero" event is piggy backed on the "cable plugged"
                -- field since the eDAQXR doesn't support cable plug and
                -- unplug events.
                if config.device_type == 'edaqxr' and not start_stamp and
                    events.cable_plugged == 1 then
                    start_stamp = tonumber(buffer.ts) +
                        (self.clk_frequency *
                            (self.srd == 'binary' and 1.25 or 1.0))
                end

                if events.pps_rising == 1 and not self.is_egps then
                    pps_event = true
                elseif events.pps_falling == 1 and self.is_egps then
                    pps_event = true
                end

                if pps_event then
                    -- put a timestamped msg in raw data output
                    local pps_ts = tonumber(buffer.ts)
                    local k = self.external_names['']
                    for _, sink in pairs(self.sinks[k] or {}) do
                        sink(pps_ts, k, self.pps_text, #self.pps_text)
                    end
                end

                if pps_event and not self.osc_control then
                    -- If we got a PPS rising event then mark the GPS as locked
                    -- and start a monitoring timer to mark it unlocked if we
                    -- lose PPS rising events for more than 3 seconds.
                    if not self.gps_lock_monitor then
                        self.gps_lock_monitor =
                            ev.Timer.new(function (loop, timer, revents)
                                    timer:stop(ev.Loop.default)
                                    self.gps_lock_monitor = nil
                                    self.pps_events = 0
                                end, 3, 0)
                        self.gps_lock_monitor:start(ev.Loop.default)
                    end
                    self.gps_lock_monitor:again(ev.Loop.default, 3)
                    self.pps_ts = buffer.ts

                    self:do_pps_state()


                    -- If we are NOT synchronizing time to GPS time then just
                    -- regulate the Si514 based on raw count.  FIXME:  This is
                    -- only for GARMIN.  Will need modification for Racetech
                    if not time_synced_to_gps then
                        self.pps_ts_array[#self.pps_ts_array + 1] = self.pps_ts
                        if self.pps_ts_array2 then
                            self.pps_ts_array2[#self.pps_ts_array2 + 1] =
                                self.pps_ts
                        end
                        if #self.pps_ts_array >= 2 then
                            local delta = self.counts_per_pps -
                                tonumber(self.pps_ts_array[#self.pps_ts_array] -
                                    self.pps_ts_array[#self.pps_ts_array - 1])

                            -- This deals with PPS dropouts with the EGPS200
                            if self.is_egps and math.abs(delta) > 2000 then
                                log:log(log.NOTICE, "PPS dropout detected")

                                -- Probably PPS drop out, remove first entry
                                table.remove(self.pps_ts_array, 1)

                                self:fixup_pps_array2()

                                -- done don't try to regulate oscillator
                                return
                            end

                            --log:notice(__"Delta:\t" ..  tostring(delta))
                            control_status.delta = delta
                            -- When the GPS is physically plugged in we seem to
                            -- get some false events that look like they are
                            -- switch bounce types of things.
                            local ppm = delta / (self.counts_per_pps / 1000000)
                            if math.abs(ppm) < 100 then
                                self:queue_servo(ppm)
                                table.remove(self.pps_ts_array, 1)
                            else
                                -- Do this because it is only done in queue
                                sl:set_status(control_status)
                                self.pps_ts_array = {}
                            end
                        end
                    end
                    self.pps_events = self.pps_events + 1
                elseif pps_event and self.osc_control then
                    -- Oscillator Control Operations
                    if self.osc_pps_ts ~= 0 then
                        self.pps_events = self.pps_events + 1
                        self.osc_offset = self.pps_interval -
                            tonumber(buffer.ts - self.osc_pps_ts)
                        -- Normalize offset to PPM of clock frequency
                        -- (self.clk_frequency / 1000000) = 40
                        local ppm = self.osc_offset / 40
                        if math.abs(ppm) < 10000 then
                            if self.osc_servo then
                                if buffer.ts ~= self.osc_servo_ts then
                                    self:queue_servo(ppm, true)
                                end
                                self.osc_servo_ts =
                                    buffer.ts + self.pps_interval
                            end
                            if self.osc_collecting then
                                self.osc_pps_data[#self.osc_pps_data + 1] = {
                                    ts = tonumber(buffer.ts),
                                    offset = tonumber(
                                        self.osc_offset * self.pps_rate),
                                    interval = tonumber(
                                        buffer.ts - self.osc_pps_ts),
                                }
                                self.osc_pps_count = self.osc_pps_count + 1
                            end
                        end
                    end
                    self.osc_pps_ts = buffer.ts
                end
                if config.device_type ~= 'edaqxr' then
                    if events.cable_unplugged == 1 then
                        if not self.plug_lock then
                            self.plug_lock = true
                            self:cmd_cb('txdr_power_off', function(msg) end)
                            self.is_egps = nil
                            self:_available_channels(true)
                            ev.Timer.new(function(l, t, e)
                                    self.plug_lock = nil
                                end, 5, 0):start(ev.Loop.default)
                        end
                    end
                    if events.cable_plugged == 1 then
                        if not self.plug_lock then
                            self.plug_lock = true
                            local cb = function(cs)
                                if cs.channel_specific[0] ~= 0 then
                                    self:cmd_cb('txdr_power_on_5v',
                                        function(msg) end)
                                    self.is_egps = nil
                                    self:_available_channels()
                                end
                                self.plug_lock = nil
                            end
                            ev.Timer.new(function(l, t, e)
                                    self:cmd_cb('get_cable_detect_state', cb)
                                end, 5, 0):start(ev.Loop.default)
                        end
                    end
                end
--[[
                -- These are more of a debugging thing and aren't really needed
                if events.start_of_break == 1 then
                    log:log(log.NOTICE, "GPS start of break")
                end
                if events.end_of_break == 1 then
                    log:log(log.NOTICE, "GPS end of break")
                end
                if events.parity_error == 1 then
                    log:log(log.NOTICE, "GPS parity error")
                end
                if events.frame_error == 1 then
                    log:log(log.NOTICE, "GPS framing error")
                end
]]
            end
        end
    end
end

function gps_proxy:start_comet_stream(w, channels, sink)
-- required channel_proxy callback function
end


-- Function used to set parameters used to control if the Si514 is regulated by
-- the GPS and whether or not PTP/System time is synchronized to GPS time.
function gps_proxy:oscillator_control(params, cb)
    local control = (params and params[1]) or 'none_specified'
    control = control:lower()
    if control == 'none' then
        time_synced_to_gps = false
        no_gps_regulation_of_oscillator = true
    elseif control == 'regulate_only' then
        time_synced_to_gps = false
        no_gps_regulation_of_oscillator = false
    elseif control == 'sync_to_gps_time' then
        time_synced_to_gps = true
        no_gps_regulation_of_oscillator = false
    elseif control == 'get' then
        return {
            result = 'success',
            msg = 'OK',
            current_state = self.oscillator_control_state,
        }
    else
        return {
            result = 'error',
            msg = string.format("Unknown oscillator_control: %s", control),
        }
    end
    self.oscillator_control_state = control

    -- Make sure the servo get switched to the correct one if the user changes
    -- the regulation mode via the API.
    set_gps_regulation()

    return {
        result = 'success',
        msg = 'OK',
        current_state = self.oscillator_control_state,
    }
end

gps_proxy.start_cb = {}

gps_proxy.start_cb['cx23r'] = function(self, start_time)
    self.start_callback = nil
    start_stamp = ts.ts_to_stamp(ts.iso_to_ts(start_time))
    if self.data_monitor and not self.data_monitor:is_active() then
        self.data_monitor:start(ev.Loop.default)
    end
end

gps_proxy.start_cb['edaqxr'] = function(self, start_time)

    -- Make sure the cable_plugged (time_zero) event is enabled
    self:cmd('event_setup', {
        cable_plugged = true,
        start_of_break = true,
        end_of_break = true,
        pps_falling = true,
        pps_rising = true,
        parity_error = true,
        frame_error = true,
    })

    self.start_callback = nil
    start_stamp = nil
    armada.timing:SetStartTime(start_time, self.srd == 'binary' and 1.25 or 1.0)
    if self.data_monitor and not self.data_monitor:is_active() then
        self.data_monitor:start(ev.Loop.default)
    end
end


function gps_proxy:start_channel_stream(w, spec, md_sink, sink)
    local channel_sink
    local ch_name = self.internal_names[spec.connector]

    if config.device_type == 'edaqxr' then
        self.srd = spec.sample_rate_domain
    end

    self.start_callback = self.start_cb[config.device_type] or
        function(self, start_time)
            log:log(log.ERR, string.format(
                "No start callback function for device type %s",
                config.device_type))
        end

    spec._timescale = string.format("1/%.0f", self.clk_frequency)
    local proxy = self
    if channel_specs[ch_name] and channel_specs[ch_name].range_max then
        channel_sink = function(timestamp, name, value)
            if not start_stamp then return end
            sink(timestamp - start_stamp, s.pack('>d', tonumber(value)))
        end
    elseif spec.connector:find('pps_offset') or
        spec.connector:find('pps_timestamps') then
        local ts = ffi.new('int64_t')
        local dd = ffi.new('double[1]')
        channel_sink = function(value)
            if not start_stamp then return end
            dd[0] = value
            sink(ts, dd, ffi.sizeof('double'))
            ts = ts + 1
        end
    else
        channel_sink = function(timestamp, name, value)
            if not start_stamp then return end
            sink(timestamp - start_stamp, tostring(value))
        end
    end

    local ch = spec.connector
    self.sinks[ch] = self.sinks[ch] or {}
    self.sinks[ch][w.socket] = channel_sink
    self.md_sink = md_sink
    w.cleanup = function()
        if self.sinks then
            if self.sinks[ch] then
                self.sinks[ch][w.socket] = nil
            end
            self.sinks[ch] = nil
        end
        start_stamp = nil
        self.md_sink = nil
        if self.data_monitor and self.data_monitor:is_active() then
            self.data_monitor:stop(ev.Loop.default)
        end
        if config.device_type == 'edaqxr' then
            armada.timing:StopClock()

            -- Disable the "cable_plugged" (time_zero) event
            self:cmd('event_setup', {
                start_of_break = true,
                end_of_break = true,
                pps_falling = true,
                pps_rising = true,
                parity_error = true,
                frame_error = true,
            })
        end
    end
    self:announce_ready(spec.connector, true)
end

local filter_set = {}

filter_set.garmin =  {
    {
        provides = {
            "utc_seconds",
            "latitude",
            "longitude",
            "fix_quality",
            "number_of_satellites_in_use",
            "horizontal_dilution_of_precision",
            "altitude",
            "geoidal_height",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGGA"},
        },
    },
    {
        provides = {
            "fix_type",
            "position_dilution_of_precision",
            "horizontal_dilution_of_precision",
            "vertical_dilution_of_precision",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGSA"},
        },
    },
    {
        provides = {
            "utc_date_time",
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "utc_day",
            "utc_month",
            "utc_year",
            "latitude",
            "longitude",
            "ground_speed_knots",
            "ground_speed_mph",
            "ground_speed_km/h",
            "ground_speed_m/s",
            "ground_heading",
            "magnetic_variation",
--[==[
            "ground_speed_mph",
            "ground_speed_km/h",
            "ground_speed_m/s",
--]==]
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPRMC"},
        },
    }
}

filter_set.egps = {
    {
        provides = {},
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PSMTID"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "latitude",
            "longitude",
            "fix_quality",
            "number_of_satellites_in_use",
            "altitude",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGGA"},
        },
    },{
        provides = {
            "utc_date_time",
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "utc_day",
            "utc_month",
            "utc_year",
            "faa_mode",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPRMC"},
        },
    },{
--[=[   provides = {
            "utc_seconds",
            "latitude",
            "longitude",
            "accelerometer_only",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGLL"},
        },
    },{
--]=]
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "speed_3d"
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLV"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "accel_x",
            "accel_y",
            "accel_z",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLA"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "speed_raw2d",
            "speed_raw3d",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLS"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "heading",
            "gradient",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLH"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "yaw_rate",
            "pitch_rate",
            "roll_rate",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLG"},
        },
    },{
        provides = {
            "utc_seconds",
            "utc_minute",
            "utc_hour",
            "rtk_yaw",
            "rtk_pitch",
            "rtk_baseline",
            "rtk_accuracy",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLR"},
        },
    },{
        provides = {
            "trigger",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$PRTLT"},
        },
    }
}

function gps_proxy:available_channels(hostname)
    self._hostname = hostname
    self.external_names = {}
    self.internal_names = {}
    self.channels = nil
    self.filters = {}

    -- These channels are always available
    local o = {
        {
            name = 'control',
            connector = 'control.status@' .. hostname,
            description = "GPS Control Channel",
            data_type = 'message_raw',
            category = 'source/hbm_' .. config.device_type .. '/gps',
            valid_controls = {
                get_supported_channels = true,
            },
        }
    }

    -- Add the "raw" channel
    o[#o + 1] = {
        name = 'raw_' .. hostname:gsub("%W", "_"),
        connector = '@' .. hostname,
        description = "Raw NMEA sentences",
        data_type = 'message_raw',
        category = 'source/hbm_' .. config.device_type .. '/msg_gps',
        valid_controls = {
            oscillator_control = true,
        },
        message_format = 'text/nmea',
    }
    self.external_names[''] = o[#o].connector -- KLUDGE for raw NMEA
    self.internal_names[o[#o].connector] = '' -- KLUDGE for raw NMEA

    self.channels = o

    if self.pps_ts_array2 then
        -- DEBUGGING CHANNELS
        o[#o + 1] = {
            name = 'pps_offset',
            connector = 'pps_offset@' .. hostname,
            sample_rate = 5,
            category = 'source/hbm_' .. config.device_type .. '/gps',
            data_type = 'sequential_float64',
            description = 'Clock ticks between PPS rising edges',
            measurand = 'Ticks',
        }
        self.external_names.pps_offset = o[#o].connector
        self.internal_names[o[#o].connector] = 'pps_offset'
        local cons = {o[#o].connector}

        o[#o + 1] = {
            name = 'pps_timestamps',
            connector = 'pps_timestamps@' .. hostname,
            sample_rate = 5,
            category = 'source/hbm_' .. config.device_type .. '/gps',
            data_type = 'sequential_float64',
            description = 'PPS timestamps',
            measurand = 'Ticks',
        }
        self.external_names.pps_offset = o[#o].connector
        self.internal_names[o[#o].connector] = 'pps_timestamps'
        cons[#cons + 1] = o[#o].connector

        local first_pps_ts

        self.pps_timer = ev.Timer.new(function(l, t, e)
                if not self.sinks then return end
                if #self.pps_ts_array2 == 0 then
                    for _, con in ipairs(cons) do
                        for _, sink in pairs(self.sinks[con] or {}) do
                            if con:find('timestamps') then
                                sink(-1)
                            else
                                sink(0)
                            end
                        end
                    end
                elseif #self.pps_ts_array2 > 1 then
                    if not first_pps_ts then
                        first_pps_ts = self.pps_ts_array2[1]
                    end
                    for _, con in ipairs(cons) do
                        for _, sink in pairs(self.sinks[con] or {}) do
                            if con:find('timestamps') then
                                for i = 1, #self.pps_ts_array2 - 1 do
                                    sink(self.pps_ts_array2[i] - first_pps_ts)
                                end
                            else
                                for i = 1, #self.pps_ts_array2 - 1 do
                                    sink(self.pps_ts_array2[i + 1] -
                                        self.pps_ts_array2[i])
                                end
                            end
                        end
                    end
                    local ts = self.pps_ts_array2[#self.pps_ts_array2]
                    self.pps_ts_array2 = {ts}
                end
            end, .2, .2)
            self.pps_timer:start(ev.Loop.default)
    end
    return self.channels
end

function gps_proxy:_available_channels(no_gps_channels)
    self:set_channel_specs()
    local hostname = self._hostname
    local o = self:available_channels(hostname)
    if no_gps_channels then
        if not self.no_channels then
            self.channels = o
            self:replace_channels(o)
            self.no_channels = true
        end
        return
    end
    self.no_channels = nil

    -- Set the filters based on GPS type
    if not self.is_egps then
        for _, f in ipairs(filter_set.garmin) do
            self.filters[#self.filters + 1] = f
        end
    else
        for _, f in ipairs(filter_set.egps) do
            self.filters[#self.filters + 1] = f
        end
    end

    local seen = {}
    for _, f in ipairs(self.filters) do
        for _, ch_name in ipairs(f.provides) do
            if not seen[ch_name] then
                seen[ch_name] = 1
                --log(log.NOTICE, 'ch_name == {0}\n', ch_name)
                assert(channel_specs[ch_name], "no value for " .. tostring(ch_name))
                local ch = {}
                if ch_name == 'utc_date_time' then
                    -- special case for utc_date_time
                    if self.is_egps then
                        ch.message_rate = 200 -- KLUDGE EGPS? specific
                    else
                        ch.message_rate = 5 -- KLUDGE Garmin-5Hz specific
                    end
                    ch.category = 'source/hbm_' .. config.device_type ..
                        '/msg_gps'
                else
                    if self.is_egps then
                        ch.message_rate = 200 -- KLUDGE EGPS? specific
                    else
                        ch.message_rate = 5 -- KLUDGE Garmin-5Hz specific
                    end
                    ch.category = 'source/hbm_' .. config.device_type .. '/gps'
                end
                for k, v in pairs(channel_specs[ch_name]) do
                    ch[k] = v
                end
                ch.connector = ch_name .. "@" .. hostname
                ch.data_type = ch.range_max
                    and 'message_float64' or 'message_raw'
--                o[#o + 1] = ch
                self.external_names[ch_name] = ch.connector
                self.internal_names[ch.connector] = ch_name
            end
        end
    end

    self.channels = o
    self:replace_channels(o)
end

function gps_proxy:do_pps_state()
    if not self.last_state or self.last_state ~= self.state then
        local states = {
            gps_pps_init = 'Wait',
            gps_pps_wait = 'Initializing',
            gps_pps_locked = 'Locked',
        }
        --log:notice(string.format("PPS State:  %s", states[self.state]))
        self.last_state = self.state
        control_status.state = states[self.state]
        sl:set_status(control_status)
    end
    if self.state == 'gps_pps_init' then
        self.state = 'gps_pps_wait'
        self.valid_pps_count = 1
        self.pps_check_timer = ev.Timer.new(function(l, t, e)
                if self.state == 'gps_pps_wait' and
                    self.valid_pps_count > (self.pps_rate == 5 and 20 or 4) then
                    self.state = 'gps_pps_locked'
                    if armada.timing.GPSGrandmaster then
                        armada.timing:GPSGrandmaster()
                    end
                    self.lock_monitor_timer =
                        ev.Timer.new(function(l, t, e)
                                if armada.timing.NormalGrandmaster then
                                    armada.timing:NormalGrandmaster()
                                end
                                self.state = 'gps_pps_init'
                                t:stop(l)
                                self.lock_monitor_timer = nil
                            end, 3, 3)
                    self.lock_monitor_timer:start(ev.Loop.default)
                else
                    self.state = 'gps_pps_init'
                end
            end, 6, 0):start(ev.Loop.default)
    elseif self.state == 'gps_pps_wait' then
        self.valid_pps_count = self.valid_pps_count + 1
    end
    if self.state == 'gps_pps_locked' and
        self.lock_monitor_timer then
        self.lock_monitor_timer:again(ev.Loop.default)
    end
end

function gps_proxy:sentence()
    -- this call does some validation on the input
    if not self.nmea:nmea_filter(self.cur_sentence) then return end

    if self.nmea_log_file then
        self.nmea_log_file:write(self.cur_sentence)
        self.nmea_log_file:flush()
    end

    if self.cur_sentence:find('$PSMTID', 1, true) then
        self.is_egps = true
        return
    end

    if not self.time_set then
        -- set box's time from the next GPS time seen
        local sentence_id, time =
            self.cur_sentence:match('^%$(%a+),([%d%.]+),.*')
        if time and time:find('%.0+$') and sentence_id and
           sentence_id == "GPRMC" then
            local date_index = 0
            -- find "date" field
            for fieldnbr = 1, 9 do
                date_index =
                    self.cur_sentence:find(',', date_index + 1, true)
                if not date_index then
                    break
                end
            end
            if date_index then
                local date =
                    self.cur_sentence:match('([%d%.]+),', date_index + 1, true)
                if date then
                    local dt_input =
                        {year = '20' .. date:sub(5, 6),
                         month = date:sub(3, 4),
                         day =   date:sub(1, 2),
                         hour =  time:sub(1, 2),
                         min =   time:sub(3, 4),
                         sec =   time:sub(5, 6)}
                    local ok =
                        timesetter.set_datetime(dt_input, 'APISERVER')
                    if ok then
                        self.time_set = true
                    end
                end
            end
        end
    end

    -- send raw sentence to proxy node
    local timestamp = tonumber(self.cur_sentence_ts)
    local k = self.external_names['']
    for _, sink in pairs(self.sinks[k] or {}) do
        sink(timestamp, k, self.cur_sentence, #self.cur_sentence)
    end
end

function gps_proxy:control(loop, w, http, spec, cb)

    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if spec.request == 'start_nmea_log' then
        self.nmea_log_file = io.open('/tmp/nmea_log.txt', 'w')
        if cb then cb({dumping = (self.nmea_log_file and true or false)}) end
        return
    end
    if spec.request == 'stop_nmea_log' then
        if self.nmea_log_file then self.nmea_log_file:close() end
        self.nmea_log_file = nil
        if cb then cb({dumping = false}) end
        return
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params, cb)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

-- Set the system time from the GPS.  Note, the response is deferred until
-- either the time is set or a timeout expires.  If the time is set, a success
-- message is returned, otherwise an error response is returned.
function gps_proxy:set_time_from_gps(params, cb)
    self.time_set = false
    local retry = 4
    ev.Timer.new(function(l, t, e)
            retry = retry - 1
            if retry == 0 then
                cb({
                    msg = 'Unable to set system time set to GPS time.',
                    result = 'error',
                })
                t:stop(l)
            elseif self.time_set then
                cb({
                    msg = 'System time set to GPS time',
                    result = 'success',
                })
                t:stop(l)
            end
        end, 1, 1):start(ev.Loop.default)
end

function gps_proxy:osc_mode(params)
    if not params then
        return {
            msg = 'osc_mode request not specified',
            result = 'error',
        }
    end
    local msg
    if params[1] == 'start' then
        if not self.osc_control then
            -- Setup Oscillator Calibration Environment
            armada.timing:AdjustClockRate(0, 'osc_mode_on')
            log:notice(__"GPS: Entering Si514 Calibration Mode on {0}",
                       self.proxy_hostname)
            if self.gps_lock_monitor then
                self.gps_lock_monitor:stop(ev.Loop.default)
                self.gps_lock_monitor = nil
            end
            self.osc_control = true
            msg = string.format("osc_mode %s succeeded on %s",
                                params[1], self.proxy_hostname)
        else
            msg = string.format("osc_mode %s already executed on %s",
                                params[1], self.proxy_hostname)
        end
        return {
            msg = msg,
            result = 'success',
        }
    elseif params[1] == 'stop' then
        if self.osc_control then
            -- Restore Default Oscillator Environment
            armada.timing:AdjustClockRate(0, 'osc_mode_off')
            log:notice(__"GPS: Leaving Si514 Calibration Mode on {0}",
                       self.proxy_hostname)
            self.osc_collecting = false
            self.osc_servo = false
            self.last_offset = 1e30
            self.integral = 0
            self.pps_events = 0
            if self.is_egps then
                self.pps_rate = 1
            else
                self.pps_rate = 5
            end
            self.dt = math.ceil(1 / self.pps_rate)
            self.pps_interval = self.clk_frequency / self.pps_rate
            self.Kp, self.Ki, self.Kd = 37, 10, 4
            self.osc_control = false
            set_gps_regulation()
            msg = string.format("osc_mode %s succeeded on %s",
                                params[1], self.proxy_hostname)
        else
            msg = string.format("osc_mode %s already executed on %s",
                                params[1], self.proxy_hostname)
        end
        return {
            msg = msg,
            result = 'success',
        }
    else
        return {
            msg = string.format("osc_mode %s failed on %s",
                                params[1], self.proxy_hostname),
            result = 'error',
        }
    end
end

function gps_proxy:osc_read(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_read not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    return {
        m_divider = armada.timing:OscillatorRead(),
        result = 'success',
    }
end

function gps_proxy:osc_write(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_write not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not params[1] then
        return {
            msg = string.format("osc_write value not specified on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local ppm = tonumber(params[1])
    return {
        m_divider = armada.timing:OscillatorWrite(ppm),
        result = 'success',
    }
end

function gps_proxy:osc_collect(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_collect not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not params[1] then
        return {
            msg = string.format("osc_collect count not specified on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local count = tonumber(params[1])
    self.osc_collecting = false
    self.osc_servo = false
    self.osc_offset = 0
    self.osc_pps_data = {}
    self.osc_pps_count = 0
    self.osc_pps_ts = 0
    self.last_offset = 1e30
    self.integral = 0
    self.pps_events = 0
    if params[2] then
        if params[2] == "on" then
            self.osc_servo = true
            if params[3] and params[4] and params[5] then
                self.Kp = params[3]
                self.Ki = params[4]
                self.Kd = params[5]
            end
            if params[6] then
                -- params[6] is the rate at which PPS rising edges
                -- will arrive at the GPS PPS connector pin.
                self.pps_rate = params[6]
                self.pps_interval = math.ceil(self.clk_frequency / params[6])
                self.dt = 1 / params[6]
            else
                return {
                    msg = string.format(
                                "osc_collect PPS interval not specified on %s",
                                self.proxy_hostname),
                    result = 'error',
                }
            end
            gps_proxy.servo = gps_proxy.time_not_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled")
        elseif params[2] == "off" then
            self.osc_servo = false
            gps_proxy.servo = function(self, offset)
                                  self.last_offset = offset
                              end
            log:notice(__"GPS: Si514 PPS Servo Disabled")
            if params[3] then
                -- params[3] is the rate at which PPS rising edges
                -- will arrive at the GPS PPS connector pin.
                self.pps_rate = params[3]
                self.pps_interval = math.ceil(self.clk_frequency / params[3])
                self.dt = 1 / params[3]
            else
                return {
                    msg = string.format(
                                "osc_collect PPS interval not specified on %s",
                                self.proxy_hostname),
                    result = 'error',
                }
            end
        else
            return {
                msg = string.format(
                                "osc_collect servo state not specified on %s",
                                self.proxy_hostname),
                result = 'error',
            }
        end
    end
    self.osc_collecting = true
    repeat
        ev.Loop.default:loop('once')
    until self.osc_pps_count == count
    return {
        data = self.osc_pps_data,
        msg = string.format("osc_collect %s succeeded on %s",
                            params[1], self.proxy_hostname),
        result = 'success',
    }
end

function gps_proxy:osc_store(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_store not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local offset = 0
    if params and params[1] then
        offset = armada.timing:OscillatorOffset(tonumber(params[1]))
    end
    return {
        msg = string.format("osc_store %d succeeded on %s",
                            offset, self.proxy_hostname),
        result = 'success',
    }
end

function gps_proxy:transducer_power(params)
    local msg
    local result
    if params[1] == '12v' then
        self:cmd_block('txdr_power_on_12v')
        msg = 'Successfully set GPS power to 12v'
        result = 'success'
    elseif params[1] == '5v' then
        self:cmd_block('txdr_power_on_5v')
        msg = 'Successfully set GPS power to 5v'
        result = 'success'
    elseif params[1] == 'off' then
        self:cmd_block('txdr_power_off')
        msg = 'Successfully turned off GPS power'
        result = 'success'
    else
        msg = 'Unknown parameter passed to GPS transducer power control API'
        result = 'error'
    end
    return {
        msg, result
    }
end

return gps_proxy
