local armada = require 'armada'
local ts = require 'timestamps'
local aux = require 'aux'
local listener = require 'listener'
local dbus = require 'dbus'
local sample_clock = require 'sample_clock'
local bus = dbus.bus_get('system')
local timing = assert(bus:auto_proxy('com.somat.armada.Timing',
                                     '/com/somat/armada/Timing'))
local ffi = require 'ffi'

local clkfreq = sample_clock.master_rate

local cmd_timeout = 0.01

local function cable_detect(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    aux_channel:get_cable_detect_state()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local plug = {[0]="UNPLUGGED", [1]="PLUGGED"}
    print("AUX cable detect with " .. devname .. " is ",
          plug[msg.channel_specific[0]])
end

local function interactive_power_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    io.write("Ready to turn on transducer power") io.read()
    aux_channel:txdr_power_on()
    assert(listener({aux_channel}, cmd_timeout))
    io.write("Ready to turn off") io.read()
    aux_channel:txdr_power_off()
    assert(listener({aux_channel}, cmd_timeout))
    io.write("Off now") io.read()
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_ts_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get current timestamp") io.read()
        aux_channel:get_current_ts()
        local msg = assert(listener({aux_channel}, cmd_timeout))
        print(msg.ts)
    end
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_sync_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get latched timestamp") io.read()
        aux_channel:get_sync_ts()
        local msg = assert(listener({aux_channel}, cmd_timeout))
        print(msg.ts)
    end
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function synchronization_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    aux_channel:setup_sync_ts()
    assert(listener({aux_channel}, cmd_timeout))
    local sync_time = timing:SendSyncPulse()
    aux_channel:get_sync_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local offset = sync_time - msg.ts + 1
    aux_channel:ts_offset(offset)
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:setup_sync_ts()
    assert(listener({aux_channel}, cmd_timeout))
    sync_time = timing:SendSyncPulse()
    aux_channel:get_sync_ts()
    msg = assert(listener({aux_channel}, cmd_timeout))
    offset = sync_time - msg.ts
    if offset == 0 then
        print("Passed")
    else
        print("Failed, offset is", offset)
    end
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function display_heartbeat_messages(channel_obj, name, ts)
    io.write("HB ") io.flush()
    channel_obj.heartbeat_count = channel_obj.heartbeat_count + 1
end

local function heartbeat_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    aux_channel.heartbeat_func = display_heartbeat_messages
    aux_channel.heartbeat_count = 0
    aux_channel:get_current_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    aux_channel:event_start(start_time - 3, 4, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:rx_start(start_time + 10000, 4, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))
    listener({aux_channel}, 10.0)
    print("\nCleaning up")
    aux_channel:event_stop()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:rx_stop()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:halt()
    print("Heartbeats detected = "..aux_channel.heartbeat_count)
    print("Done")
    os.exit(0)
end

local function display_plug_messages(channel_obj, name, event)
    if event.events.cable_unplugged == 1 then
        print("CBL_UNPLUGGED")
    end
    if event.events.cable_plugged == 1 then
        print("CBL_PLUGGED")
    end
    if event.events.vbm_plugged == 1 then
        print("VBM_PLUGGED")
    end
    if event.events.vbm_unplugged == 1 then
        print("VBM_UNPLUGGED")
    end
end

local function detection_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    aux_channel.element_action_func = display_plug_messages
    aux_channel:txdr_power_on()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:setup_aux()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:event_setup({cable_unplugged = true, cable_plugged = true,
                             vbm_plugged = true, vbm_unplugged = true})
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:get_current_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    aux_channel:event_start(start_time - 3, 2, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:rx_start(start_time - 3, 1, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))
    print("Running for 30 seconds, plug and unplug a VBM")
    listener({aux_channel}, 30.0)
    print("Cleaning up")
    aux_channel:event_stop()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:rx_stop()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:txdr_power_off()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function edaq_sync_test(devname)
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    aux_channel:get_current_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local aux_start_time = msg.ts + clkfreq
    aux_channel:setup_sync(2, aux_start_time, aux_start_time, clkfreq / 2)
    assert(listener({aux_channel}, cmd_timeout))
    io.write("Sync packets transmitting...") io.read()
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function display_vbm_message(channel_obj, name, pktstruct)
    if pktstruct.is_control_frame == 1 then
        print(ffi.string(pktstruct.msg, pktstruct.msg_length))
    else
        local s = ffi.string(pktstruct.msg, pktstruct.msg_length)
        local bytes = {s:byte(1,-1)}
        for i=1,#bytes do io.write(string.format("%02X ", bytes[i])) end
        print("TS "..tonumber(pktstruct.ts))
    end
end

local function aux_monitor_test(arg)
    local devname = arg[1]
    local mode = arg[2]
    local power = arg[3]
    local drd = {}
    if mode and mode:lower() == 'sync' then
        drd = arg[3]
        power = nil
    end
    local aux_channel = aux:new(devname, "aux.bin", clkfreq)
    if power and power:lower() == 'power' then
        aux_channel:txdr_power_on()
        assert(listener({aux_channel}, cmd_timeout))
    end
    local start_time = 0
    local clock_rate = 0
    aux_channel:get_current_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    if mode == 'vbm' then
        print("VBM Mode")
        aux_channel.element_action_func = display_vbm_message
        aux_channel:setup_vbm()
        assert(listener({aux_channel}, cmd_timeout))
        local aux_start_time = msg.ts + clkfreq
        aux_channel:rx_start(aux_start_time - 3, 1, 16, 128)
        assert(listener({aux_channel}, cmd_timeout))
        aux_channel:tx_start(0, 16, 128)
        assert(listener({aux_channel}, cmd_timeout))
        aux_channel:tx_cmd_msg(0, {"HIGH_SPEED=YES",
                                   "TERMINATION_ENABLE=YES",
                                   "CAN_MODE=INITIALIZE",
                                   "SYNC_JUMP_WIDTH=3",
                                   "BAUD_RATE_PRESCALER=4",
                                   "TIME_SEGMENT_1=11",
                                   "TIME_SEGMENT_2=4",
                                   "SAMPLES_PER_BIT=3",
                                   "LISTEN_ONLY_MODE=NO",
                                   "CAN_MODE=NORMAL",
                                   "APPSTATE?",
                                   "RUN=YES"
                                   })
        assert(listener({aux_channel}, 3.0))
        -- Send messages
        local msgs = aux_channel:allocate_msg_array(2)
        local vbm_msg_1 = string.char(0x49, 0x93, 0x42, 0x66, 0x65)..'Hello'
        local vbm_msg_2 = string.char(0x89, 0x93, 0x42, 0x66, 0x67)..'World'
        aux_channel:insert_msg(msgs, 1, 0, vbm_msg_1)
        aux_channel:insert_msg(msgs, 2, aux_start_time + (6 * clkfreq), vbm_msg_2)
        aux_channel:tx_append_msgs(0, msgs)
        -- Run for another 2.0 seconds
        assert(listener({aux_channel}, 2.0))
        print("Cleaning up")
        aux_channel:tx_stop(0)
        assert(listener({aux_channel}, cmd_timeout))
        aux_channel:rx_stop()
        assert(listener({aux_channel}, cmd_timeout))
    elseif mode == 'sync' then
        print("SYNC Mode")
        if drd == 'decimal' then
            start_time = msg.ts + clkfreq
            clock_rate = 100000 * 2
        else
            start_time = msg.ts + clkfreq + (clkfreq / 4)
            clock_rate = 98304 * 2
        end
        aux_channel:setup_sync(clock_rate, start_time)
        assert(listener({aux_channel}, cmd_timeout))
        -- Run for another 60.0 seconds
        listener({aux_channel}, 60.0)
        print("Cleaning up")
    elseif mode == 'ecm' then
        print("ECM Mode")
        if drd == 'decimal' then
            start_time = msg.ts + clkfreq
            clock_rate = 100000 * 2
        else
            start_time = msg.ts + clkfreq + (clkfreq / 4)
            clock_rate = 98304 * 2
        end
        aux_channel:setup_ecm(clock_rate, start_time, start_time, clkfreq)
        assert(listener({aux_channel}, cmd_timeout))
        print("Cleaning up")
    end
    if power then
        aux_channel:txdr_power_off()
        assert(listener({aux_channel}, cmd_timeout))
    end
    aux_channel:halt()
    print("Done")
    os.exit(0)
end

local function usage ()
    print("Usage: aux_test <device name> [POWER_TEST|DETECTION_TEST|GET_TS|"..
          "GET_SYNC|SYNC_TEST|HEARTBEAT_TEST|CABLE_DETECT]")
    print("       aux_test <device name> [vbm] <power>")
    print("       aux_test <device name> [sync|ecm] <decimal|legacy>")
end

if #arg < 1 then
    usage()
    os.exit(1)
end

if #arg > 0 then
    if not arg[1] or not arg[2] then
        usage()
        os.exit(1)
    end
    if arg[2] == "POWER_TEST" then interactive_power_test(arg[1]) end
    if arg[2] == "DETECTION_TEST" then detection_test(arg[1]) end
    if arg[2] == "GET_TS" then get_ts_test(arg[1]) end
    if arg[2] == "GET_SYNC" then get_sync_test(arg[1]) end
    if arg[2] == "SYNC_TEST" then synchronization_test(arg[1]) end
    if arg[2] == "HEARTBEAT_TEST" then heartbeat_test(arg[1]) end
    if arg[2] == "EDAQ_SYNC_TEST" then edaq_sync_test(arg[1]) end
    if arg[2] == "CABLE_DETECT" then cable_detect(arg[1]) end
    if arg[2] == 'vbm' or arg[2] == 'sync' or arg[2] == 'ecm' then
        aux_monitor_test(arg)
        os.exit(0)
    end
end
