-- Test code for channel package

local ch = require 'channel'
local cc = require 'connctrl'
local ffi = require 'ffi'
local rb = require 'ring_buffers'


ffi.cdef[[
typedef struct { uint8_t testbuf[1200];
} testbuf_t;
]]

local function test_ring_initialization(devname)
    local bd_ring = rb.make_new_bd_ring(devname, "testbuf_t", "testbuf_t",  5)
    local bd_size = ffi.sizeof("bd_t")

    bd_ring:dump("Initialization Test")

    for i=0,bd_size*5,bd_size do
        io.write(string.format("%04X: ", i))
        for j=0,bd_size-4,4 do
            io.write(string.format("%08X ", bd_ring.dma_map:r32(i + j)))
        end
        print("")
    end

    os.exit(0);
end

ffi.cdef[[
typedef struct { uint8_t cmdbuf[64];
} cmdbuf_t;
]]

ffi.cdef[[
typedef struct { uint8_t intbuf[160];
} intbuf_t;
]]

local function test_ring_buffers(devname)
    local sendsize

    chan = ch:new_channel(devname, "test4.bin", "cmdbuf_t", "intbuf_t")

    for i=1,1000 do
        if cc.is_controller_halted(devname) then
            print("Failed, exception taken")
            cc.dump_exception_info(devname)
            cc.controller_reset(devname)
            os.exit(1)
        end

        cmd = ffi.new("cmdbuf_t")

        sendsize = ((264643 + i) % 17) + ((56333 + i) % 11) + 2

        for j=1,sendsize do cmd.cmdbuf[j - 1] = j end

        if not chan.command_rb:put_buffer(cmd, sendsize) then
            print("Failed send_buffer()")
            os.exit(1)
        end

        local interrupt_taken = false
        for k=1,100000 do
            if cc.is_controller_halted(devname) then
                print("Failed, exception taken")
                cc.dump_exception_info(devname)
                cc.controller_reset(devname)
                os.exit(1)
            end

            if cc.test_for_interrupt(devname) then
                interrupt_taken = true
                break
            end
        end

        if not interrupt_taken then
            print("Failed, interrupt not received")
            cc.controller_reset(devname)
            os.exit(1)
        end

        local intr, bytes_written = chan.response_rb:get_buffer()
        if not bytes_written then
            print("Failed, no buffer returned")
            cc.controller_reset(devname)
            os.exit(1)
        end

        if bytes_written ~= sendsize + 1 then
            print("Failed, bad byte count returned", bytes_written, sendsize)
            chan:dump_ring_buffers()
            cc.controller_reset(devname)
            os.exit(1)
        end

        for j=1,sendsize do
            if cmd.cmdbuf[j - 1] + 1 ~= intr.intbuf[j - 1] then
                print("Failed, buffer mismatch")
                chan:dump_ring_buffers()
                cc.controller_reset(devname)
                os.exit(1)
            end
        end

        if intr.intbuf[sendsize] ~= bytes_written then
            print("Failed, buffer length mismatch",
                  intr.intbuf[sendsize], bytes_written)
            chan:dump_ring_buffers()
            cc.controller_reset(devname)
            os.exit(1)
        end
    end

    cc.controller_reset(devname)

    print("Passed")
end


if #arg < 1 then
    print("Usage channel_test <device name> [<loop count>|RING_INIT]")
    os.exit(1)
end

local loop_count = 1
if #arg > 1 then
    if arg[2] == "RING_INIT" then test_ring_initialization(arg[1]) end
    loop_count = tonumber(arg[2])
end

for i=1,loop_count do
    test_ring_buffers(arg[1])
end

