--------------------------------------------------------------------------------
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--------------------------------------------------------------------------------

require 'middleclass'

require('lj_include') {
    'fcntl',
    'stdlib',
    'string',
    'unistd',
    'netinet/in',
    'sys/types',
    'sys/socket',
}

local bit   = require 'bit'
local ev    = require 'ev'
local ffi   = require 'ffi'

local C = ffi.C

--------------------------------------------------------------------------------

local tcp_server = class('tcp_server')

--------------------------------------------------------------------------------

function tcp_server:initialize(loop, port)

    self.loop = loop
    self.port = port

end

--------------------------------------------------------------------------------

function tcp_server:on_new_client(handler) self.on_new_client = handler end
function tcp_server:on_error(handler) self.on_error = handler end
function tcp_server:on_shutdown(handler) self.on_shutdown = handler end

--------------------------------------------------------------------------------

function tcp_server:start()

    if self.running then return end

    local function err(message, errno, nothrow)
        if errno then message = message .. string.format(' (errno: %d)', errno) end
        if nothrow and self.on_error then self.on_error(self, message) end
        self:_shutdown(not nothrow)
        if not nothrow then error(message) end
    end

    -- Create a socket for the TCP server.
    self.socket = C.socket(C.AF_INET, C.SOCK_STREAM, C.IPPROTO_TCP)
    if self.socket == -1 then err('Failed to create server socket', ffi.errno()) end

    -- Allow address reuse (ignore lingering sockets in TIME_WAIT). Note that
    -- failure to set this option is not currently handled as a fatal error;
    -- we will try to continue anyway.
    local reuse = ffi.new('int[1]', { 1 })
    local reuse_len = ffi.new('socklen_t', ffi.sizeof('int'))
    C.setsockopt(self.socket, C.SOL_SOCKET, C.SO_REUSEADDR, reuse, reuse_len)

    -- Create and populate the local address structure.
    local sin_size = ffi.sizeof('struct sockaddr_in')
    local sin = ffi.cast('struct sockaddr_in *', C.malloc(sin_size))
    if not sin then err('Failed to allocate server socket address struct', ffi.errno()) end
    C.memset(sin, 0, sin_size)
    sin.sin_family = C.AF_INET
    sin.sin_port = C.htons(self.port)

    -- Bind the server socket to the local address.
    local rv = C.bind(self.socket, ffi.cast('const struct sockaddr *', sin), sin_size)
    if rv ~= 0 then
        local errno = ffi.errno()
        C.free(sin)
        err('Failed to bind server socket to local address', errno)
    end
    C.free(sin)

    -- Put the socket into nonblocking mode.
    local flags = C.fcntl(self.socket, C.F_GETFL, 0)
    local rv = C.fcntl(self.socket, C.F_SETFL, bit.bor(flags, C.O_NONBLOCK))
    if rv ~= 0 then err('Failed to put server socket into nonblocking mode', ffi.errno()) end

    -- Begin listening for incoming connections.
    local rv = C.listen(self.socket, 5)
    if rv ~= 0 then err('Failed to put server socket in listening mode', ffi.errno()) end

    -- Set up an event loop listener to trigger on incoming connection.
    self.watcher = ev.IO.new(
        function(loop, watcher, e)

            -- Create the remove address structure.
            local sin = ffi.cast('struct sockaddr_in *', C.malloc(ffi.sizeof('struct sockaddr_in')))
            if not sin then return err('Failed to allocate client socket address struct', nil, true) end
            local addr_len = ffi.new('int[1]', { ffi.sizeof('struct sockaddr_in') })

            -- Accept the incoming connection.
            local client_socket = C.accept(self.socket, ffi.cast('struct sockaddr *', sin), addr_len)
            if client_socket == -1 then
                return err('Failed to accept incoming client connection', ffi.errno(), true)
            end

            -- Invoke the client handler (or if there is none, close the socket).
            if self.on_new_client then
                local ip_bytes = ffi.cast('unsigned char *', sin.sin_addr)
                local ip = string.format('%d.%d.%d.%d',
                    ip_bytes[0], ip_bytes[1], ip_bytes[2], ip_bytes[3])
                local port = C.ntohs(sin.sin_port)
                self.on_new_client(self, client_socket, ip, port)
            else
                C.close(client_socket)
            end

        end, self.socket, ev.READ)

    self.watcher:start(self.loop)
    self.running = true

end

--------------------------------------------------------------------------------

function tcp_server:stop()

    self:_shutdown()

end

--------------------------------------------------------------------------------

function tcp_server:_shutdown(noraise)

    if self.watcher then self.watcher:stop(self.loop) end
    if self.socket then C.close(self.socket) end

    self.socket = nil
    self.watcher = nil
    self.running = false

    if not noraise and self.on_shutdown then self.on_shutdown(self) end

end

--------------------------------------------------------------------------------

return tcp_server
