--! \file

require'middleclass'

require('lj_include'){
    'stdlib',
    'unistd',
    'errno',
    'sys/types',
    'sys/socket',
    'ifaddrs',
    'netdb',
    'net/if',
    'linux/sockios',
    'linux/net_tstamp',
}

local ev = require'ev'
local struct = require'struct'
local membuf = require'membuf'
local ffi = require'ffi'
local config = require 'system_config'
if config.device_type ~= 'edaqxr' then
    require 'ptpv2-ffi/cdata'
end

local C = ffi.C

ffi.cdef[[
struct scm_timestamping {
    struct timespec     systime;
    struct timespec     hwtimetrans;
    struct timespec     hwtimeraw;
};

/* Pinger stuff */
struct iphdr {
    uint8_t       ihl_version;
    uint8_t       tos;
    uint16_t      tot_len;
    uint16_t      id;
    uint16_t      frag_off;
    uint8_t       ttl;
    uint8_t       protocol;
    uint16_t      check;
    uint32_t      saddr;
    uint32_t      daddr;
} __attribute__((packed))__;

struct icmphdr {
    uint8_t       type;
    uint8_t       code;
    uint16_t      checksum;
    union {
        struct {
            uint16_t      id;
            uint16_t      sequence;
        } echo;
        uint32_t   gateway;
        struct {
            uint16_t      __unused;
            uint16_t      mtu;
        } frag;
    } un;
    uint8_t       data[56];
} __attribute__((packed))__;

enum {
    ICMP_ECHO = 8,
};
]]

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

--! Various network utilities.
--? namespace net
local net = { }

function net.shutdown(sd)
    C.shutdown(sd, C.SHUT_RDWR)
    C.close(sd)
end

function net.packmsg(msg)
    return struct.pack('>I4c0', msg:len(), msg)
end

--! Normalize an IPv6 address.
function net.normalize_ipv6(addr)
dbg('normalize_ipv6("%s") A\n', addr)
    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_flags = C.AI_NUMERICSERV + C.AI_NUMERICHOST
    hints[0].ai_socktype = C.SOCK_STREAM
    hints[0].ai_family = C.AF_INET6
    local ret = ffi.new('struct addrinfo *[1]')
    if C.getaddrinfo(addr, "0", hints, ret) ~= 0 then
dbg('normalize_ipv6("%s") B\n', addr)
        return nil
    end

dbg('normalize_ipv6("%s") C\n', addr)
    local ai = ret[0]
    local host = ffi.new('char[?]', C.NI_MAXHOST)
    local ip
    if C.getnameinfo(ai.ai_addr, ai.ai_addrlen, host, C.NI_MAXHOST, nil, 0,
            C.NI_NUMERICHOST + C.NI_NUMERICSERV) == 0 then
        ip = ffi.string(host)
dbg('normalize_ipv6("%s") D, ip = %s\n', addr, ip)
    end
    C.freeaddrinfo(ai)
dbg('normalize_ipv6("%s") E, ip = %s\n', addr, ip)
    return ip
end

function net.getifaddrs(interface)
    local ifap = ffi.new('struct ifaddrs *[1]')
    local addrs = {}
    if C.getifaddrs(ifap) == 0 then
        local host = ffi.new('char[?]', C.NI_MAXHOST)
        local ifa = ifap[0]
        while ifa ~= nil do
            if ffi.string(ifa.ifa_name) == interface then
                local size
                if ifa.ifa_addr.sa_family == C.AF_INET then
                    size = ffi.sizeof('struct sockaddr_in')
                elseif ifa.ifa_addr.sa_family == C.AF_INET6 then
                    size = ffi.sizeof('struct sockaddr_in6')
                end
                if size then
                    if C.getnameinfo(ifa.ifa_addr, size, host,
                        C.NI_MAXHOST, nil, 0, C.NI_NUMERICHOST) == 0 then
                        addrs[#addrs + 1] = ffi.string(host)
                    end
                end
            end
            ifa = ifa.ifa_next
        end
        C.freeifaddrs(ifap[0])
    end
    return addrs
end

--! An ARP class
--?
net.ARP = class('net.ARP')

function net.ARP:initialize(callback)
    self.callback = callback
end

--function net.ARP:whohas(iface, my_ip, target_ip, timeout, loop)
function net.ARP:whohas(iface, my_ip, target_ip, timeout, loop)
    loop = loop or ev.Loop.default

    -- Allocate memory for various arrays.
    self.src_mac = ffi.new('unsigned char[6]')
    self.dst_mac = ffi.new('unsigned char[6]')
    self.ether_frame = ffi.new('unsigned char[?]', C.IP_MAXPACKET)
    self.interface = ffi.new('char[40]')
    self.target = ffi.new('char[40]')
    self.src_ip = ffi.new('char[?]', C.INET_ADDRSTRLEN)


    -- Interface to send packet through.
    ffi.copy(self.interface, iface)

    -- Submit request for a socket descriptor to look up interface.
    self.sd = C.socket(C.AF_INET, C.SOCK_RAW, C.IPPROTO_RAW)
    if self.sd < 0 then
        error("Unable to create socket")
    end

    -- Use ioctl() to look up interface name and get its MAC address.
    local ifr = ffi.new('struct ifreq[1]')
    ffi.copy(ifr[0].ifr_ifrn.ifrn_name, self.interface, iface:len())
    local status = C.ioctl(self.sd, C.SIOCGIFHWADDR, ifr)
    if status < 0 then
        C.close(self.sd)
        error("ioctl() failed to get source MAC address")
    end
    C.close(self.sd)

    -- Copy source MAC address.
    ffi.copy(self.src_mac, ifr[0].ifr_ifru.ifru_hwaddr.sa_data, 6)

    -- Find interface index from interface name and store index in
    -- struct sockaddr_ll device, which will be used as an argument of sendto().
    local device = ffi.new('struct sockaddr_ll[1]')
    device[0].sll_ifindex = C.if_nametoindex(iface)
    if device[0].sll_ifindex == 0 then
        error("if_nametoindex() failed to obtain interface index ")
    end

    -- Set destination MAC address: broadcast address
    ffi.fill(self.dst_mac, ffi.sizeof(self.dst_mac), 0xff)

    -- Source IPv4 address:  you need to fill this out
    ffi.copy(self.src_ip, my_ip, my_ip:len())

    -- IPv4 address we want to find MAC address of
    ffi.copy(self.target, target_ip, target_ip:len())

    -- Fill out hints for getaddrinfo().
    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_family = C.AF_INET
    hints[0].ai_socktype = C.SOCK_STREAM
    hints[0].ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV

    -- Resolve source using getaddrinfo().
    local res = ffi.new('struct addrinfo *[1]')
    local status = C.getaddrinfo(self.src_ip, nil, hints, res)
    if status ~= 0 then
        error(string.format("getaddrinfo() failed: %s",
            ffi.string(C.gai_strerror(status))))
    end

    local ipv4 = ffi.cast('struct sockaddr_in *', res[0].ai_addr)

    local arphdr = ffi.new('arp_hdr[1]')
    ffi.copy(arphdr[0].sender_ip, ipv4.sin_addr, 4)
    C.freeaddrinfo(res[0])

    -- Resolve target using getaddrinfo().
    local status = C.getaddrinfo(self.target, nil, hints, res)
    if status ~= 0 then
        error(string.format("getaddrinfo() failed: %s",
            ffi.string(C.gai_strerror(status))))
    end

    local ipv4 = ffi.cast('struct sockaddr_in *', res[0].ai_addr)
    ffi.copy(arphdr[0].target_ip, ipv4.sin_addr, 4)
    self.look_for_ip = ffi.new('char[4]')
    ffi.copy(self.look_for_ip, ipv4.sin_addr, 4)
    C.freeaddrinfo(res[0])

    -- Fill out sockaddr_ll.
    device[0].sll_family = C.AF_PACKET
    ffi.copy(device[0].sll_addr, self.src_mac, 6)
    device[0].sll_halen = C.htons(6)

    -- ARP header

    -- Hardware type (16 bits): 1 for ethernet
    arphdr[0].htype = C.htons(1)

    -- Protocol type (16 bits): 2048 for IP
    arphdr[0].ptype = C.htons(C.ETH_P_IP)

    -- Hardware address length (8 bits): 6 bytes for MAC address
    arphdr[0].hlen = 6

    -- Protocol address length (8 bits): 4 bytes for IPv4 address
    arphdr[0].plen = 4

    -- OpCode: 1 for ARP request
    arphdr[0].opcode = C.htons(C.ARPOP_REQUEST)

    -- Sender hardware address (48 bits): MAC address
    ffi.copy(arphdr[0].sender_mac, self.src_mac, 6)

    -- Sender protocol address (32 bits)
    -- See getaddrinfo() resolution of src_ip.

    -- Target hardware address (48 bits): zero, since we don't know it yet.
    ffi.fill(arphdr[0].target_mac, 6)

    -- Target protocol address (32 bits)
    -- See getaddrinfo() resolution of target.

    -- Fill out ethernet frame header.

    -- Ethernet frame length = ethernet header (MAC + MAC + ethernet type) +
    -- ethernet data (ARP header) + 18 bytes of padding
    local frame_length = 6 + 6 + 2 + C.ARP_HDRLEN + 18

    -- Destination and Source MAC addresses
    ffi.copy(self.ether_frame, self.dst_mac, 6)
    ffi.copy(self.ether_frame + 6, self.src_mac, 6)

    -- Next is ethernet type code (ETH_P_ARP for ARP).
    -- http://www.iana.org/assignments/ethernet-numbers
    self.ether_frame[12] = C.ETH_P_ARP / 256
    self.ether_frame[13] = C.ETH_P_ARP % 256

    -- Next is ethernet frame data (ARP header).

    -- ARP header
    ffi.copy(self.ether_frame + C.ETH_HDRLEN, arphdr, C.ARP_HDRLEN)

    -- Submit request for a raw socket descriptor.
    self.sd = C.socket(C.AF_PACKET, C.SOCK_RAW, C.htons(C.ETH_P_ALL))
    if self.sd < 0 then
        error("socket() failed")
    end

    local read_watcher = ev.IO.new(function(l, i, e)
            local addrlen =
                ffi.new('socklen_t[1]', {ffi.sizeof('struct sockaddr')})
            local bytes =
                C.recvfrom(self.sd, self.ether_frame, C.IP_MAXPACKET,
                           C.MSG_DONTWAIT,
                           ffi.cast('struct sockaddr *', device), addrlen)

            if bytes <= 0 then
                if bytes < 0 and ffi.errno() ~= C.EAGAIN then
                    -- RX error, done
                    i:stop(l)
                    return
                end
                -- No response
            else
                local ahdr = ffi.cast('struct _arp_hdr *',
                    self.ether_frame + C.ETH_HDRLEN)
                if C.memcmp(self.ether_frame, self.src_mac, 6) == 0 and
                    self.ether_frame[12] == math.floor(C.ETH_P_ARP / 256) and
                    self.ether_frame[0] == self.src_mac[0] and
                    self.ether_frame[1] == self.src_mac[1] and
                    self.ether_frame[2] == self.src_mac[2] and
                    self.ether_frame[3] == self.src_mac[3] and
                    self.ether_frame[4] == self.src_mac[4] and
                    self.ether_frame[5] == self.src_mac[5] and
                    self.ether_frame[13] == C.ETH_P_ARP % 256 and
                    ahdr.sender_ip[0] == self.look_for_ip[0] and
                    ahdr.sender_ip[1] == self.look_for_ip[1] and
                    ahdr.sender_ip[2] == self.look_for_ip[2] and
                    ahdr.sender_ip[3] == self.look_for_ip[3] and
                    ahdr.opcode == C.ARPOP_RESPONSE then

                    i:stop(l)
                    local cb = self.callback
                    self.callback = nil
                    if cb then
                        local mac = ''
                        for i = 0, 5 do
                            mac = mac .. string.format("%02X%s",
                                bit.band(ahdr.sender_mac[i], 0xff),
                                (i < 5) and ':' or '')
                        end
                        local ipd = ffi.new('char[64]')
                        if C.inet_ntop(C.AF_INET, ffi.cast('const void *',
                            self.ether_frame + C.ETH_HDRLEN +
                            ffi.offsetof('struct _arp_hdr', 'sender_ip')),
                            ipd, ffi.sizeof(ipd)) == nil then

                            cb(target_ip)
                            return
                        end
                        cb(ffi.string(ipd), mac)
                    end
                end
            end
        end, self.sd, ev.READ)

    ev.Timer.new(function(l, t, e)
            t:stop(l)
            read_watcher:stop(l)
            C.close(self.sd)
            if self.callback then self.callback(target_ip) end
        end, timeout, 0):start(loop)

    read_watcher:start(loop)

    -- Send ethernet frame to socket.
    local bytes = C.sendto(self.sd, self.ether_frame, frame_length, 0,
        ffi.cast('struct sockaddr *', device), ffi.sizeof(device))
    if bytes <= 0 then
        error("sendto() failed")
    end

    return true
end

--! A ping class
--?
net.Pinger = { }
net.Pinger.__index = net.Pinger

local function in_cksum(addr, len)
    local sum = ffi.new('unsigned short', 0)
    local answer = ffi.new('unsigned short', 0)
    local w = addr
    local nleft = len

    while nleft > 1 do
        sum = sum + w[0]
        w = w + 1
        nleft = nleft - 2
    end
    assert(nleft == 0)
    if nleft == 1 then
    end

    sum = bit.rshift(tonumber(sum), 16) + bit.band(tonumber(sum), 0xffff)
    sum = sum + bit.rshift(tonumber(sum), 16)
    answer = ffi.new('unsigned short',
        bit.band(bit.bxor(tonumber(sum), 0xffff), 0xffff))
    return answer
end

-- This really could be done via ljsyscall netlink BUT there are way too many
-- collisions with that stuff to deal with at the moment.
function net.Pinger:get_addrs(interface)
    if not self.ipv4_addrs and not self.ipv6_addrs then
        local fh = io.popen('ip addr show dev ' .. interface)
        for line in fh:lines() do
            local addr = line:match('inet (.+)/(%d+) brd')
            if addr then
                self.ipv4_addrs = self.ipv4_addrs or {}
                self.ipv4_addrs[#self.ipv4_addrs + 1] = addr
            else
                local addr = line:match('inet6 (.+)/(%d+) scope')
                if addr then
                    self.ipv6_addrs = self.ipv6_addrs or {}
                    self.ipv6_addrs[#self.ipv6_addrs + 1] = addr
                end
            end
        end
        fh:close()
    end
end

function net.Pinger:ping(host, timeout, interface, loop, saddr)
    loop = loop or ev.Loop.default
    interface = interface or 'ethint'

    if not saddr then
        self:get_addrs(interface)
        for _, addr in ipairs(self.ipv4_addrs) do
            if not addr:find('^169%.254%.') then
                saddr = addr
                break
            end
        end
        if not saddr then
            return
        end
    end

    math.randomseed(os.time())
    local rate = .1
    local sockfd = C.socket(C.AF_INET, C.SOCK_RAW, C.IPPROTO_ICMP)
    if sockfd < 0 then
        return
    end
    local optval = ffi.new('int[1]', {1})
    C.setsockopt(sockfd, C.IPPROTO_IP, C.IP_HDRINCL, optval, ffi.sizeof(optval))

    -- Bind to the specified device
    local status = C.setsockopt(sockfd, C.SOL_SOCKET, C.SO_BINDTODEVICE,
        interface, ffi.new('int', interface:len()))
    if status ~= 0 then
        C.close(sockfd)
        return
    end

    local tot_len = ffi.sizeof('struct iphdr') + ffi.sizeof('struct icmphdr')

    local packet = ffi.cast('char *', C.malloc(tot_len))
    if packet == nil then
        C.close(sockfd)
        return
    end
    local buffer = ffi.cast('char *', C.malloc(tot_len))
    if buffer == nil then
        C.free(packet)
        C.close(sockfd)
        return
    end
    local ip = ffi.cast('struct iphdr *', packet)
    local icmp = ffi.cast('struct icmphdr *', ip + 1)
    local connection = ffi.cast('struct sockaddr_in *',
        C.malloc(ffi.sizeof('struct sockaddr_in')))
    if connection == nil then
        C.free(buffer)
        C.free(packet)
        C.close(sockfd)
        return
    end
    local addrlen = ffi.cast('socklen_t *', C.malloc(ffi.sizeof('socklen_t')))
    if addrlen == nil then
        C.free(connection)
        C.free(buffer)
        C.free(packet)
        C.close(sockfd)
        return
    end

    local got_response
    local seq = 1
    local watcher = ev.Timer.new(function(l, t, e)

        -- Set up the IP header
        ip.ihl_version = 0x45;
        ip.tos = 0;
        ip.tot_len = tot_len
        ip.id = bit.band(math.random(0, 0xffff), 0xffff)
        ip.frag_off = 0
        ip.ttl = 255
        ip.protocol = C.IPPROTO_ICMP;
        ip.saddr = C.inet_addr(saddr)
        ip.daddr = C.inet_addr(host)

        -- Set up the ICMP header
        icmp.type = C.ICMP_ECHO
        icmp.code = 0
        icmp.un.echo.id = 0
        icmp.un.echo.sequence = C.ntohs(seq)
        icmp.checksum = 0
        icmp.checksum = in_cksum(ffi.cast('unsigned short *', icmp),
            ffi.sizeof('struct icmphdr'))
        ip.check = in_cksum(ffi.cast('unsigned short *', ip),
            ffi.sizeof('struct iphdr'))
        seq = seq + 1

--[[
local str = 'Packet: '
for i = 0, tot_len - 1 do
    str = str .. string.format(" %02X", bit.band(packet[i], 0xff))
end
print(str)
]]

        -- Set up the send to address
        connection.sin_family = C.AF_INET
        connection.sin_addr.s_addr = C.inet_addr(host)

        -- Send the ICMP ECHO request
        local sent = C.sendto(sockfd, packet, ip.tot_len, 0,
            ffi.cast('struct sockaddr *', connection),
            ffi.sizeof('struct sockaddr'))
        if sent < 0 then
            timeout = 0
            t:stop(l)
            return
        end

        if not self.callback then
            -- Check for a response
            addrlen[0] = ffi.sizeof('struct sockaddr_in')
            local rcvd =
                C.recvfrom(sockfd, buffer, tot_len, C.MSG_DONTWAIT,
                ffi.cast('struct sockaddr *', connection),
                addrlen)
            if rcvd <= 0 then
                if ffi.errno() ~= C.EAGAIN then
                    -- RX error, done
                    timeout = 0
                    t:stop(l)
                    return
                end
                -- No response, go try again
            else
                if connection.sin_addr.s_addr == ip.daddr then
                    -- Got response, done
                    t:stop(l)
                    got_response = true
                end
            end

            timeout = timeout - rate

            -- Timed out?
            if timeout <= 0 then
                -- YES
                t:stop(l)
            end
        end
    end, rate, rate)

    -- Start pinging
    watcher:start(loop)

    -- Wait for timeout or response, whichever arrives first
    while timeout > 0 and not got_response and self.callback == nil do
        loop:loop('once')
    end

    -- If a callback function has been attached, use it rather than blocking
    -- until complete.
    if self.callback then
        local read_watcher = ev.IO.new(function(l, i, e)
                -- Check for a response
                addrlen[0] = ffi.sizeof('struct sockaddr_in')
                local rcvd =
                    C.recvfrom(sockfd, buffer, tot_len, C.MSG_DONTWAIT,
                    ffi.cast('struct sockaddr *', connection),
                    addrlen)
                if rcvd < 0 then
                    if ffi.errno() ~= C.EAGAIN then
                        -- RX error, done
                        watcher:stop(loop)
                        i.tw:stop(loop)
                        i:stop(l)
                        C.close(sockfd)
                        C.free(connection)
                        C.free(buffer)
                        C.free(packet)
                        self.callback()
                        return
                    end
                else
                    if connection.sin_addr.s_addr == ip.daddr then
                        watcher:stop(loop)
                        i.tw:stop(loop)
                        i:stop(l)
                        C.close(sockfd)
                        C.free(connection)
                        C.free(buffer)
                        C.free(packet)
                        self.callback(ffi.string(C.inet_ntoa(ip.daddr)))
                    end
                end
            end, sockfd, ev.READ)
        local tw = ev.Timer.new(function(l, t, e)
                timeout = timeout - rate
                if timeout <= 0 then
                    t:stop(loop)
                    read_watcher:stop(loop)
                    watcher:stop(loop)
                    C.close(sockfd)
                    C.free(connection)
                    C.free(buffer)
                    C.free(packet)
                    self.callback()
                end
            end, rate, rate)
        read_watcher.tw = tw
        read_watcher:start(loop)
        tw:start(loop)
        return
    end

    -- Clean up resources
    C.close(sockfd)
    C.free(connection)
    C.free(buffer)
    C.free(packet)

    -- Return status
    return got_response
end


--! A very low-level socket class
--? class net.Socket
net.Socket = { }
net.Socket.__index = net.Socket

function net.socket(opts)
    opts = opts or { }
    return setmetatable(opts, net.Socket)
end

-- I don't want this to ever go away
if config.device_type ~= 'edaqxr' then
    net.Socket.ts = ffi.cast('internal_time_t *',
        C.calloc(1, ffi.sizeof('internal_time_t')))
end

function net.Socket:gai(passive, dont_need_ts)
    if self.addrs then return end
    local flags = passive and C.AI_PASSIVE or 0
    local host = self.host
    local service = tostring(self.service)
    local family = self.family
    local type = self.type or C.SOCK_STREAM
    if self.ip then
        host = self.ip
        flags = flags + C.AI_NUMERICHOST
    end
    if host and not family then family = C.AF_UNSPEC end
    if not service and not self.port then self.port = 0 end
    if self.port then
        service = self.port
        flags = flags + C.AI_NUMERICSERV
    end
    service = tostring(service)
    if not family then family = C.AF_INET6 end
    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_flags = flags
    hints[0].ai_family = family
    hints[0].ai_socktype = type
    local ret = ffi.new('struct addrinfo *[1]')
    local status = C.getaddrinfo(host, service, hints, ret)
    assert(status == 0, string.format('getaddrinfo failure: %s',
        ffi.string(C.gai_strerror(status))))
    self.addrs = ret[0]

    if not dont_need_ts then
        -- timestamping stuff
        self.rx_ts = {}
        local bsize = 8192
        -- Don't use ffi.new for this stuff so it doesn't get garbage collected!
        self.rx_ts.vec = ffi.cast('struct iovec *',
            C.calloc(1, ffi.sizeof('struct iovec')))
        self.rx_ts.vec[0].iov_base = C.calloc(bsize, ffi.sizeof('char'))
        self.rx_ts.vec[0].iov_len = bsize
        self.rx_ts.msg = ffi.cast('struct msghdr *',
            C.calloc(1, ffi.sizeof('struct msghdr')))
        self.rx_ts.data_len = 8192
        self.rx_ts.data = ffi.cast('char *',
            C.calloc(self.rx_ts.data_len, ffi.sizeof('char')))
        self.rx_ts.name_len = 128
        self.rx_ts.name = ffi.cast('char *',
            C.calloc(self.rx_ts.name_len, ffi.sizeof('char')))
        self.rx_ts.host_len = 256
        self.rx_ts.host = ffi.cast('char *',
            C.calloc(self.rx_ts.host_len, ffi.sizeof('char')))
        self.rx_ts.service_len = 256
        self.rx_ts.service = ffi.cast('char *',
            C.calloc(self.rx_ts.service_len, ffi.sizeof('char')))
            ffi.gc(self.rx_ts.service, function(p)
                if self.rx_ts.vec ~= nil and
                    self.rx_ts.vec[0].iov_base ~= nil then
                    C.free(self.rx_ts.vec[0].iov_base)
                    self.rx_ts.vec[0].iov_base = nil
                end
                for _, k in ipairs{'host','name','data','msg','vec'} do
                    if self.rx_ts[k] ~= nil then
                        C.free(self.rx_ts[k])
                        self.rx_ts[k] = nil
                    end
                end
            end)
    end

    ffi.gc(ret[0], function(p)
            C.freeaddrinfo(p)
        end)
end

function net.Socket:recvfrom(wantts)
    local msg = self.rx_ts.msg[0]
    local flags = C.MSG_DONTWAIT + self.extra_recv_flags

    if wantts then
        msg.msg_control = self.rx_ts.data
        msg.msg_controllen = self.rx_ts.data_len
    else
        msg.msg_control = nil
        msg.msg_controllen = 0
    end
    msg.msg_name = self.rx_ts.name
    msg.msg_namelen = self.rx_ts.name_len
    msg.msg_iov = self.rx_ts.vec
    msg.msg_iovlen = 1
    msg.msg_flags = 0

    local ret = C.recvmsg(self.sock, self.rx_ts.msg, flags)
    if ret < 0 then
        local errno = ffi.errno()
        if errno == C.EAGAIN or errno == C.EINTR then
            if self.error_cb then
                self.error_cb(errno)
            end
            return
        elseif errno == C.ENOMSG then
            self.extra_recv_flags = 0
            if self.error_cb then
                self.error_cb(errno)
            end
            return
        end
        if self.error_cb then
            self.error_cb(errno)
        end
        return
    end

    assert(bit.band(msg.msg_flags, C.MSG_TRUNC) == 0,
        "Received truncated message")

    local rawcount
    if wantts and bit.band(msg.msg_flags, C.MSG_CTRUNC) == 0 then
        local cmsg = ffi.cast('struct cmsghdr *', self.rx_ts.data)
        if msg.msg_controllen >= ffi.sizeof('struct cmsghdr') then
            if cmsg.cmsg_level == C.SOL_SOCKET then
                local d = cmsg + 1
                if cmsg.cmsg_type == C.SO_TIMESTAMP then
                    local tv = ffi.cast('struct timeval *', d)
                    self.ts.seconds = tv.tv_sec
                    self.ts.nanoseconds = (tv.tv_usec * 1000)
                elseif cmsg.cmsg_type == C.SO_TIMESTAMPING then
                    local scm_timestamping =
                        ffi.cast('struct scm_timestamping *', d)
                    -- NOTE - this timestamp must be converted
                    -- to system time
                    rawcount =
                        (scm_timestamping.hwtimeraw.tv_sec * 1000000000) +
                        scm_timestamping.hwtimeraw.tv_nsec
                    self.ts.seconds = scm_timestamping.hwtimeraw.tv_sec
                    self.ts.nanoseconds = scm_timestamping.hwtimeraw.tv_nsec
                end
                if bit.band(self.extra_recv_flags,
                    msg.msg_flags, C.MSG_ERRQUEUE) == C.MSG_ERRQUEUE then
                    self.extra_recv_flags = 0
                    if self.tx_ts_cb then
                        self.tx_ts_cb(self.ts, rawcount)
                    end
                    return
                end
            end
        end
    end

    if C.getnameinfo(ffi.cast('struct sockaddr *', msg.msg_name),
        msg.msg_namelen, self.rx_ts.host, self.rx_ts.host_len,
        self.rx_ts.service, self.rx_ts.service_len,
        C.NI_NUMERICHOST + C.NI_NUMERICSERV) ~= 0 then
        error(string.format("Unable to get name info: %s",
            ffi.string(C.strerror(ffi.errno()))))
    end

    return self.rx_ts.vec[0].iov_base, ret, ffi.string(self.rx_ts.host),
        ffi.string(self.rx_ts.service), self.ts, rawcount
end

function net.Socket:connect()
    self:gai(false, true)
    -- KLUDGE handle optional non-blocking connect
    local a = self.addrs
    while a ~= nil do
        local sock = C.socket(a.ai_family, a.ai_socktype, a.ai_protocol)
        assert(sock > 0, "socket failure")
        local status = 0
        if a.ai_socktype == C.SOCK_STREAM then
            status = C.connect(sock, a.ai_addr, a.ai_addrlen)
            -- print('new connect', sock)
        end
        if status == 0 then
            self.addr = a
            self.sock = sock
            return self
        end
        -- Don't leak the socket away if the connect fails
        C.close(sock)
        a = a.ai_next
    end
    error("Couldn't connect")
end

function net.Socket:connect_nb(loop, callback)
    self:gai(false, true)
    -- KLUDGE handle optional non-blocking connect
    local a = self.addrs
    while a ~= nil do
        local sock = C.socket(a.ai_family, a.ai_socktype + C.SOCK_NONBLOCK,
            a.ai_protocol)
        assert(sock > 0, "socket failure")
        local status = 0
        if a.ai_socktype == C.SOCK_STREAM then
            self.addr = a
            self.sock = sock
            local connect_complete_watcher = ev.IO.new(function(l, t, e)
                    t:stop(l)
                    local e = ffi.new('int[1]')
                    local sz = ffi.new('int[1]')
                    sz[0] = ffi.sizeof(e)
                    C.getsockopt(sock, C.SOL_SOCKET, C.SO_ERROR, e, sz)
                    if sz[0] == ffi.sizeof(e) and e[0] == 0 then
                        callback(self)
                    else
                        C.close(sock)
                        callback()
                    end
                end, sock, ev.WRITE)
            status = C.connect(sock, a.ai_addr, a.ai_addrlen)
            if status == 0 then
                callback(self)
            elseif ffi.errno() == C.EINPROGRESS then
                connect_complete_watcher:start(loop)
            else
                C.close(sock)
                callback()
            end
            return
        end
        a = a.ai_next
    end
end

function net.Socket:bind(opts)
    opts = opts or {}
    if opts.interface then self.interface = opts.interface end
    self:gai(true)
    self.addr = self.addrs[0]
    local a = self.addr
    local sock = C.socket(a.ai_family, a.ai_socktype, a.ai_protocol)
    assert(sock > 0, string.format("Unable to create socket: %s",
        ffi.string(C.strerror(ffi.errno()))))

    -- Bind to a specific interface
    if self.interface then
        local status = C.setsockopt(sock, C.SOL_SOCKET, C.SO_BINDTODEVICE,
            self.interface, ffi.new('int', self.interface:len()))
        assert(status == 0, string.format("Unable to bind to interface %s: %s",
            self.interface, ffi.string(C.strerror(ffi.errno()))))
    end

    -- IPV6 only if specified
    if opts.ipv6_only then
        local on = ffi.new('int[1]', {1})
        local status = C.setsockopt(sock, C.IPPROTO_IPV6, C.IPV6_V6ONLY, on,
            ffi.sizeof(on))
        assert(status == 0, string.format("Unable able to set V6ONLY mode: %s",
            ffi.string(C.strerror(ffi.errno()))))
    end

    local status = C.bind(sock, a.ai_addr, a.ai_addrlen)
    assert(status == 0, string.format("Unable to bind socket: %s",
        ffi.string(C.strerror(ffi.errno()))))

    local v = ffi.new('int[1]', {1})
    local v_len = ffi.new('socklen_t', ffi.sizeof(v))
    status = C.setsockopt(sock, C.SOL_SOCKET, C.SO_REUSEADDR, v, v_len)
    assert(status == 0, string.format("Unable to set SO_REUSEADDR: %s",
        ffi.string(C.strerror(ffi.errno()))))

    self.sock = sock

    local host = ffi.new('char[1024]')
    local serv = ffi.new('char[1024]')
    local name = ffi.new('char[1024]')
    local name_len = ffi.new('socklen_t[1]', {ffi.sizeof(name)})
    status = C.getsockname(sock, ffi.cast('struct sockaddr *', name), name_len)
    assert(status == 0, string.format("Unable to get socket name: %s",
        ffi.string(C.strerror(ffi.errno()))))
    status = C.getnameinfo(ffi.cast('const struct sockaddr *', name),
        name_len[0], host, ffi.sizeof(host), serv, ffi.sizeof(serv),
        C.NI_NUMERICHOST + C.NI_NUMERICSERV)
    assert(status == 0, string.format("Unable to get socket name info: %s",
        ffi.string(C.strerror(ffi.errno()))))
    self.ip = ffi.string(host)
    self.port = ffi.string(serv)
    return self
end

function net.Socket:listen(opts)
    opts = opts or { }
    if not self.sock then self:bind(opts) end
    local status = C.listen(self.sock, opts.backlog or 16)
    assert(status == 0,
        string.format("Listen failure %s", ffi.string(C.strerror(ffi.errno()))))
    self.acceptor = opts.acceptor
    local function handler(loop, w)
        local addr = ffi.new('char[256]')
        local addr_len = ffi.new('int[1]', {ffi.sizeof(addr)})
        local s = C.accept(self.sock, ffi.cast('struct sockaddr *', addr),
            addr_len)
        assert(s > 0, string.format("Accept failure: %s",
            ffi.string(C.strerror(ffi.errno()))))
        self:acceptor(loop, s, addr)
    end
    self.listen_watcher = ev.IO.new(handler, self.sock, ev.READ)
    if self.loop then
        sock.listen_watcher:start(self.loop)
    end
    return self
end

function net.Socket:getpeername(sock)
    local addr = ffi.cast('struct sockaddr *', ffi.new('char[256]'))
    local addrlen = ffi.new('int[1]', {ffi.sizeof(addr)})
    if C.getpeername(sock or self.sock, addr, addrlen) ~= 0 then
        print(string.format("getpeername failed: %s",
            ffi.string(C.strerror(ffi.errno()))))
        return
    end
    local host = ffi.new('char[256]')
    local port = ffi.new('char[256]')
    if C.getnameinfo(addr, addrlen[0], host, ffi.sizeof(host), port,
        ffi.sizeof(port), C.NI_NUMERICHOST + C.NI_NUMERICSERV) ~= 0 then
        print(string.format("getnameinfo failed: %s",
            ffi.string(C.strerror(ffi.errno()))))
        return
    end
    return ffi.string(host), ffi.string(port)
end

function net.Socket:close()
    if self.sock then
        C.close(self.sock)
    end

    -- This causes all of the gc stuff to happen
    self.addrs = nil
    if self.rx_ts then
        self.rx_ts.service = nil
    end
    collectgarbage()
end

function net.Socket:start(loop)
    self.loop = loop
    if self.listen_watcher then
        self.listen_watcher:start(loop)
    end
end

function net.Socket:stop(loop)
    assert(loop == self.loop)
    if self.listen_watcher then
        self.listen_watcher:stop(loop)
        self:close()
    end
    self.loop = nil
end

function net.Socket:is_hw_timestamped()
    return self.hw_rx_timestamps or self.hw_tx_timestamps
end

function net.Socket:is_timestamped()
    return self.timestamping_enabled
end

function net.Socket:timestamping(rx_enable, tx_enable)
    self.hw_tx_timestamps = false
    self.hw_rx_timestamps = false
    self.timestamping_enabled = false
    if self.type ~= C.SOCK_DGRAM then return end
    assert(self.interface,
        string.format("Timestamping requires binding to a specific interface"))

    local ifreq = ffi.new('struct ifreq[1]')
    local hwtstamp_config = ffi.new('struct hwtstamp_config[1]')
    ffi.copy(ifreq[0].ifr_ifrn.ifrn_name, self.interface)
    ifreq[0].ifr_ifru.ifru_data = ffi.cast('void *', hwtstamp_config)
    if tx_enable then
        hwtstamp_config[0].tx_type = C.HWTSTAMP_TX_ON
    else
        hwtstamp_config[0].tx_type = C.HWTSTAMP_TX_OFF
    end
    hwtstamp_config[0].rx_filter = C.HWTSTAMP_FILTER_ALL

    local status = C.ioctl(self.sock, C.SIOCSHWTSTAMP, ifreq)
    if status < 0 then
        local e = ffi.errno()
        assert(e ~= C.EINVAL, string.format(
            "Failed to enable hw timestamping: %s", ffi.string(C.strerror(e))))
        io.stderr:write(string.format(
            "Failed to enable HW timestamping reverting to SW timestamping"))
        local enable = ffi.new('int[1]', {1})
        status =
            C.setsockopt(self.sock, C.SOL_SOCKET, C.SO_TIMESTAMP, enable,
            ffi.new('socklen_t', ffi.sizeof(enable)))
            assert(status == 0, string.format(
                "Unable to revert to sw tx timestamping: %s",
                ffi.string(C.strerror(ffi.errno()))))
    end

    local hwts_opts = ffi.new('int[1]')
    if tx_enable then
        hwts_opts[0] = C.SOF_TIMESTAMPING_TX_HARDWARE
    end
    hwts_opts[0] = hwts_opts[0] + C.SOF_TIMESTAMPING_RX_HARDWARE +
            C.SOF_TIMESTAMPING_RAW_HARDWARE + C.SOF_TIMESTAMPING_SOFTWARE
    status = C.setsockopt(self.sock, C.SOL_SOCKET, C.SO_TIMESTAMPING,
        hwts_opts, ffi.new('socklen_t', ffi.sizeof(hwts_opts)))

    if status < 0 then
        local enable = ffi.new('int[1]', {1})
        status = C.setsockopt(self.sock, C.SOL_SOCKET, C.SO_TIMESTAMP, enable,
            ffi.new('socklen_t', ffi.sizeof(enable)))
            assert(status == 0, string.format(
                "Unable to revert to sw rx timestamping: %s",
                ffi.string(C.strerror(ffi.errno()))))
    else
        self.hw_rx_timestamps = true
        if tx_enable then
            self.hw_tx_timestamps = true
        end
    end
    self.timestamping_enabled = true
    self.extra_recv_flags = 0
end

function net.Socket:join_multicast_group(group)
    local idx
    if self.interface then
        idx = ffi.new('int[1]', {C.if_nametoindex(self.interface)})
        assert(C.setsockopt(self.sock, C.IPPROTO_IPV6, C.IPV6_MULTICAST_IF,
            idx, ffi.sizeof(idx)) == 0,
            string.format("Unable to get interface index: %s",
            ffi.string(C.strerror(ffi.errno()))))
    end

    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_flags = C.AI_NUMERICSERV + C.AI_NUMERICHOST
    hints[0].ai_socktype = C.SOCK_DGRAM
    hints[0].ai_family = C.AF_INET6
    local ret = ffi.new('struct addrinfo *[1]')
    local status = C.getaddrinfo(group, "0", hints, ret)
    assert(status == 0, string.format("Unable to get addr info: %s",
        ffi.string(C.gai_strerror(status))))

    local mr = ffi.new('struct ipv6_mreq[1]')
    mr[0].ipv6mr_interface = idx[0]
    local a = ffi.cast('struct sockaddr_in6 *', ret[0].ai_addr)
    ffi.copy(mr[0].ipv6mr_multiaddr.__in6_u.__u6_addr8, a.sin6_addr,
        ffi.sizeof(mr[0].ipv6mr_multiaddr.__in6_u.__u6_addr8))
    C.freeaddrinfo(ret[0])
    status = C.setsockopt(self.sock, C.IPPROTO_IPV6, C.IPV6_JOIN_GROUP, mr,
        ffi.sizeof(mr))
    assert(status == 0, string.format("Can not join multicast group %s: %s",
        group, ffi.string(C.strerror(ffi.errno()))))

    local ttl = ffi.new('int[1]', {self.ttl})
    status = C.setsockopt(self.sock, C.IPPROTO_IPV6, C.IPV6_MULTICAST_HOPS,
        ttl, ffi.sizeof(ttl))
    assert(status == 0, string.format("Can not set multicast hop limit: %s",
        ffi.string(C.strerror(ffi.errno()))))

    if self.timestamping_enabled then
        local loopback = ffi.new('int[1]', {1})
        local rcverr = ffi.new('int[1]', {0})
        if self.hw_tx_timestamps then
            loopback[0] = 0
            rcverr[0] = 1
        end
        status = C.setsockopt(self.sock, C.IPPROTO_IPV6,
            C.IPV6_MULTICAST_LOOP, loopback, ffi.sizeof(loopback))
        assert(status == 0,
            string.format("Can not set multicast loopback: %s",
            ffi.string(C.strerror(ffi.errno()))))

        status = C.setsockopt(self.sock, C.IPPROTO_IPV6, C.IPV6_RECVERR,
            rcverr, ffi.sizeof(rcverr))
        assert(status == 0,
            string.format("Can not enable error receipt: %s",
            ffi.string(C.strerror(ffi.errno()))))
    end
end

function net.Socket:set_sendto_addr(host, port)
    local sa = ffi.new('char[8192]')
    local salen = ffi.new('socklen_t[1]', {ffi.sizeof(sa)})

    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_flags = C.AI_NUMERICSERV + C.AI_NUMERICHOST
    hints[0].ai_socktype = C.SOCK_DGRAM
    hints[0].ai_family = C.AF_UNSPEC
    local ret = ffi.new('struct addrinfo *[1]')
    port = tostring(port)
    local status = C.getaddrinfo(host, port, hints, ret)

    assert(status == 0, string.format("Unable to get addr info: %s",
        ffi.string(C.gai_strerror(status))))

    if self.sendto_addr then
        C.free(self.sendto_addr)
    end
    self.sendto_addr = ffi.cast('struct sockaddr *',
        C.calloc(ret[0].ai_addrlen, ffi.sizeof('char')))
    ffi.copy(self.sendto_addr, ret[0].ai_addr, ret[0].ai_addrlen)
    self.sendto_addrlen = ret[0].ai_addrlen
    C.freeaddrinfo(ret[0])
end

function net.Socket:sendto(buf, blen, ts_cb)
    if ts_cb and self.hw_tx_timestamps then
        self.extra_recv_flags = C.MSG_ERRQUEUE
        self.tx_ts_cb = ts_cb
    end

    local sent = C.sendto(self.sock, buf, blen, 0, self.sendto_addr,
        self.sendto_addrlen)

    if sent < 0 then
        local e = ffi.errno()
        if self.error_cb then
            self.error_cb(e)
        else
            error(string.format("sendto error: %s", ffi.string(C.strerror(e))))
        end
    end
end

--! A buffer abstraction for stream sockets
--? class net.StreamBuffer
net.StreamBuffer = class('net.StreamBuffer')

function net.streambuf(socket)
    return net.StreamBuffer:new(socket)
end

function net.StreamBuffer:initialize(socket, blocksize, reader_time)
    if type(socket) == 'table' and socket.sock then
        self.socket = socket.sock
        self.sockobj = socket
    else
        self.socket = socket
    end
    self.blocksize = blocksize or 8192
    self.readbuf = membuf:new(self.blocksize)

    self.read_closing = false
    self.write_closing = false
    local function reader(loop, w)
        self.readbuf:assure_available()
        local got = C.recv(self.socket, self.readbuf:tail(),
            self.readbuf:available(), C.MSG_DONTWAIT)

        local read_will_close = false
        if got == 0 then
            self.reader:stop(loop)
            read_will_close = true
        end
        if got < 0 then
            local e = ffi.errno()
            if e == C.ECONNRESET then
                if self.close_cb then
                    self:close_cb(loop, 'read', 'remote')
                end
                self:close()
                return
            end
            if e ~= C.EAGAIN then
                print(string.format("reader: recv error: %d", e))
                self.readbuf:clear()
            end
            return
        end
        self.readbuf:append(nil, got)
        while self.read_cb and #self.readbuf do
            local consumed = self:read_cb(loop)
            if consumed == 0 then break end
            local e, m = pcall(self.readbuf.remove, self.readbuf, 0, consumed)
            if not e then
                self.readbuf:clear()
                break
            end
        end

        if read_will_close then
            self.read_closed = true
            if self.close_cb then
                self:close_cb(loop, 'read', 'remote')
            end
            if self.write_closed then
                self:close()
            end
            self.readbuf:clear()
        end
    end
    if reader_time then
        self.reader_time = reader_time
        self.reader = ev.Timer.new(reader, reader_time, reader_time)
    else
        self.reader = ev.IO.new(reader, self.socket, ev.READ)
    end

    self.write_closing = false
    self.write_closed = false
    self.writebuf = membuf:new(self.blocksize)

    local function writer(loop, w)
        local sent = C.send(self.socket, self.writebuf:head(),
            #self.writebuf, C.MSG_DONTWAIT)
        if sent <= 0 then
            local e = ffi.errno()
            if sent < 0 and e ~= C.EAGAIN then
                self.writebuf:clear()
                self.writer:stop(loop)
                self.write_closed = true
                if self.read_closed then
                    self:close()
                end
            end
            return
        elseif sent then
            self.writebuf:remove(0, sent)
        end
        if self.write_closing then
            if #self.writebuf == 0 then
                self.writer:stop(loop)
                C.shutdown(self.socket, C.SHUT_WR)
                self.write_closed = true
                if self.read_closed then
                    self:close()
                end
                return
            end
        else
            if self.write_cb then
                self:write_cb(loop, sent)
            end
        end
        if #self.writebuf == 0 then
            self.writer:stop(loop)
        end
    end
    self.writer = ev.IO.new(writer, self.socket, ev.WRITE)
end

function net.StreamBuffer:read(size)
    local avail = #self.readbuf
    if not size or size > avail then
        size = avail
    end
    local buf = self.readbuf:head()
    return size, buf
end

function net.StreamBuffer:write(data, size)
    if self.write_closed then
        return
    end

    local osize = size or #data
    assert(osize)
    if osize == 0 then
        return 0
    end

    local sent = 0
    -- If the write buffer is empty, try just sending the data from the buffer
    -- it already is in.  If send returns EAGAIN, shove the data into the write
    -- buffer and start the write watcher.  If send returns some other error,
    -- drop the data on the floof and return.  If send returns 0 (remote
    -- disconnected), stop the write watcher and return error.  If less than
    -- the requested amount was sent, add the remainder to the write buffer and
    -- start the write watcher.  If the write buffer already has data in it
    -- when this function is entered, just add the new data to the write buffer
    -- and make sure the write watcher is running.
    if #self.writebuf == 0 then
        sent = C.send(self.socket, data, osize, C.MSG_DONTWAIT)
        if sent < 0 then
            local e = ffi.errno()
            if e ~= C.EAGAIN then
                -- Error, stop write watcher
                self.writer:stop(self.loop)
                if self.write_cb then
                    self.write_cb(self.loop, sent)
                end
                return
            end
            sent = 0
        elseif sent == 0 then
            self.writer:stop(self.loop)
            if self.write_cb then
                self.write_cb(self.loop, sent)
            end
            return
        elseif self.write_cb then
            self.write_cb(self.loop, sent)
        end
    end
    local s = osize - sent
    if s > 0 then
        if type(data) == 'string' then
            self.writebuf:append(data:sub(sent + 1), s)
        else
            self.writebuf:append(ffi.cast('uint8_t *', data) + sent, s)
        end
        self.writer:start(self.loop)
    end
    return osize
end

function net.StreamBuffer:writev(vec, count)
    if #self.writebuf == 0 then
        -- KLUDGE do real C.writev() call here, and handle short writes
        -- properly. Need to set O_NONBLOCK on fd for that to work.
    end
    local wrote = 0
    for i = 0, count - 1 do
        wrote = wrote + (self:write(vec[i].iov_base, vec[i].iov_len) or 0)
    end
    return wrote
end

function net.StreamBuffer:close()
    if self.sockobj then
        self.sockobj:close()
        self.socket = nil
    end
    if self.socket then
        C.close(self.socket)
    end
    if self.loop then self:stop(self.loop) end
    self.socket = nil
end

function net.StreamBuffer:shutdown_write()
    if not self.write_closed then
        self.write_closing = true
        if #self.writebuf == 0 then
            C.shutdown(self.socket, C.SHUT_WR)
            self.write_closed = true
            if self.loop then
                self.writer:stop(self.loop)
            end
        end
    end
    if self.write_closed and self.read_closed then
        self:close()
    end
end

function net.StreamBuffer:shutdown()
    self:shutdown_write()
end

function net.StreamBuffer:start(loop)
    self.loop = loop
    self.writebuf:clear()
    self.readbuf:clear()
    if not self.read_closed then
        self.reader:start(loop)
    end
    if #self.writebuf > 0 and not self.write_closed then
        self.writer:start(loop)
    end
end

function net.StreamBuffer:stop(loop)
    self.reader:stop(loop)
    self.writer:stop(loop)
    self.writebuf:clear()
    self.readbuf:clear()
    self.loop = nil
    if self.stop_cb then self.stop_cb() end
end

-- local sb = net.streambuf(s)
-- function sb:read_size()
--     local got = self:read(4)
--     self.size = struct.unpack('>I4', got)
--     self:read_watermark(self.size)
--     self:read_callback(self.read_data)
-- end
-- function sb:read_data()
--     local got = self:read(self.size)
--     -- process data
--     self:write(output)
--     self:read_watermark(4)
--     self:read_callback(self.read_size)
-- end
-- sb:read_callback(sb.read_size)
-- sb:read_watermark(4)
-- sb:start(loop)

-- local s = net.socket{host = 'www.google.com', port = 80}:connect()
-- local sbuf = net.streambuf(s.sock)
-- function sbuf:read_cb(loop)
--     local got = self:read()
--     if got then
--         print('got', got:len())
--     else
--         self:shutdown()
--     end
-- end
-- sbuf:write('GET / HTTP/1.0\r\n')
-- sbuf:write('Host: www.google.com\r\n')
-- sbuf:write('\r\n')
-- sbuf:start(ev.Loop.default)

-- local ls = net.socket{port = 12345}:listen()
-- function ls:acceptor(loop, sock)
--     local sbuf = net.streambuf(sock)
--     function sbuf:read_cb(loop)
--         local got = self:read()
--         if got then
--             self:write(got)
--         else
--             self:shutdown()
--         end
--     end
--     sbuf:start(loop)
-- end
-- ls:start(ev.Loop.default)

-- local ls = net.socket{port = 12346}:listen()
-- function ls:acceptor(loop, sock)
--     local sbuf = net.streambuf(sock)
--     sbuf:start(loop)
--     sbuf:shutdown_read()
--     for i=1,1000 do
--         sbuf:write(string.format('i = %f, i * i = %f\n', i, i * i))
--     end
--     sbuf:shutdown()
-- end
-- ls:start(ev.Loop.default)

-- ev.Loop.default:loop()

--[=[
local ev = require 'ev'

--local ls = net.socket{port = 16587, type = C.SOCK_STREAM, interface = 'eth0'}:listen()
local ls = net.socket{port = 7412}:listen()
function ls:acceptor(loop, sock)
    local sbuf = net.streambuf(sock)
    function sbuf:read_cb(loop)
        local s, b = self:read()
        if s > 0 then
            print(string.format("Got: %s", ffi.string(b, s)))
            return s
        elseif s == 0 then
            print("Remote disconnect")
            sbuf:shutdown()
        else
            print(string.format("Error: %s",
                ffi.string(C.strerror(ffi.errno()))))
            sbuf:shutdown()
        end
        return 0
    end
    function sbuf:close_cb(loop, dir, reason)
        print("Closed")
    end
    sbuf:start(loop)
end
ls:start(ev.Loop.default)
    
--    net.socket{port = 16587, type = C.SOCK_DGRAM, interface = 'eth0'}:bind()

-- Listen IPv4 any
--local ds = net.socket{port = 16587, ip = '*', type = C.SOCK_DGRAM}

-- Listen IPv4 specific interface
--local ds = net.socket{port = 16587, ip = '172.26.1323.106', type = C.SOCK_DGRAM}

-- Listen IPv6 link local
--local ds = net.socket{port = 16587, ip = 'fe80::2d0:56ff:feff:6%eth0', type = C.SOCK_DGRAM}

-- Listen IPv6 global
--local ds = net.socket{port = 16587, ip = 'fd13:497:ade4:0:2d0:56ff:feff:6', type = C.SOCK_DGRAM}

-- Listen IPv6 any
--local ds = net.socket{port = 16587, ip = '::', type = C.SOCK_DGRAM}
--local ds = net.socket{port = 340, type = C.SOCK_DGRAM}
local ds = net.socket{port = 340, type = C.SOCK_DGRAM}

if ds then
    ds:start(ev.Loop.default)
    ds:bind({interface = 'eth1'})
    ds:timestamping(true)
    ds.ttl = 1
    ds:join_multicast_group("ff02::181")
    local rc, tc = 0, 0
    local function read_func(loop, io)
        local buf, blen, host, serv, time = ds:recvfrom_with_ts()
        if not buf then return end
        rc = rc + 1
        time = time or 0.0
        print(string.format('Got "%s" from %s on port %s at %.0f %d',
            ffi.string(buf, blen), host, serv, time, rc))
    end
    ev.IO.new(read_func, ds.sock, ev.READ):start(ev.Loop.default)

    local tstr = 'This is a test string'
    ds:set_sendto_addr("ff02::181", 340)
    function ds:tx_ts_cb(ts)
        tc = tc + 1
        if ts then
            print(string.format('Sent "%s" at %.0f %d', tstr, ts, tc))
        end
    end
    ev.Timer.new(function(loop, t, e)
            ds:sendto_get_ts(tstr, tstr:len())
        end, .125, .125):start(ev.Loop.default)
end

ev.Loop.default:loop()
print("Done")
]=]

return net
