local ffi = require 'ffi'
local C = ffi.C

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

ffi.cdef[[
    struct timeval {
        unsigned long       tv_sec;
        long                tv_usec;
    };

    struct rusage {
        struct timeval ru_utime; /* user time used */
        struct timeval ru_stime; /* system time used */
        long   ru_maxrss;        /* maximum resident set size */
        long   ru_ixrss;         /* integral shared memory size */
        long   ru_idrss;         /* integral unshared data size */
        long   ru_isrss;         /* integral unshared stack size */
        long   ru_minflt;        /* page reclaims */
        long   ru_majflt;        /* page faults */
        long   ru_nswap;         /* swaps */
        long   ru_inblock;       /* block input operations */
        long   ru_oublock;       /* block output operations */
        long   ru_msgsnd;        /* messages sent */
        long   ru_msgrcv;        /* messages received */
        long   ru_nsignals;      /* signals received */
        long   ru_nvcsw;         /* voluntary context switches */
        long   ru_nivcsw;        /* involuntary context switches */
    };

    enum {
        RUSAGE_SELF = 0,
        RUSAGE_CHILDREN = -1,
        RUSAGE_THREAD = 1
    };

    int getrusage(int who, struct rusage *usage);

    struct timezone {
        int tz_minuteswest;     /* minutes west of Greenwich */
        int tz_dsttime;         /* type of DST correction */
    };

    int gettimeofday(struct timeval *tv, struct timezone *tz);
]]

local rusage_labels = {
    'user time',
    'system time',
    'max resident set size',
    'shared memory size',
    'unshared data size',
    'unshared stack size',
    'page reclaims',
    'page faults',
    'swaps',
    'block input ops',
    'block output ops',
    'messages sent',
    'messages received',
    'signals received',
    'voluntary context switches',
    'involuntary context switches',
}

local rusage_fields = {
    'ru_utime',
    'ru_stime',
    'ru_maxrss',
    'ru_ixrss',
    'ru_idrss',
    'ru_isrss',
    'ru_minflt',
    'ru_majflt',
    'ru_nswap',
    'ru_inblock',
    'ru_oublock',
    'ru_msgsnd',
    'ru_msgrcv',
    'ru_nsignals',
    'ru_nvcsw',
    'ru_nivcsw',
}

local mt = {}

local function timeval(t)
    return tonumber(t.tv_sec) + tonumber(t.tv_usec) * 1e-6
end

local function minus(a, b)
    return a - b
end

local function plus(a, b)
    return a + b
end

function mt:binop(op, arg, result)
    for i = 1, 2 do
        local f = rusage_fields[i]
        local a = timeval(self[f])
        local b = timeval(arg[f])
        local r = op(a, b)
        result[f].tv_sec = math.floor(r)
        result[f].tv_usec = 1e6 * (r - tonumber(result[f].tv_sec))
    end
    for i = 3, #rusage_fields do
        local f = rusage_fields[i]
        result[f] = op(self[f], arg[f])
    end
end

function mt:add(arg, result)
    return self:binop(plus, arg, result)
end

function mt:sub(arg, result)
    return self:binop(minus, arg, result)
end

function mt:to_table()
    local o = {}
    for i = 1, 2 do
        local f = rusage_fields[i]
        o[f] = timeval(self[f])
        o.total_time = (o.total_time or 0) + o[f]
    end
    for i = 3, #rusage_fields do
        local f = rusage_fields[i]
        o[f] = tonumber(self[f])
    end
    return o
end

function mt:dump()
    local o = {}
    for i = 1, 2 do
        o[#o + 1] = string.format('%s: %.6f\n',
            rusage_labels[i], timeval(self[rusage_fields[i]]))
    end
    for i = 3, #rusage_fields do
        o[#o + 1] = string.format('%s: %d\n',
            rusage_labels[i], tonumber(self[rusage_fields[i]]))
    end
    return table.concat(o)
end

mt.labels = {}
for i = 1, #rusage_fields do
    mt.labels[rusage_fields[i]] = rusage_labels[i]
end

function mt:dump_percentage(all, wall)
    local o = {}
    local ttime = 0
    local ttime_all = 0
    for i = 1, 2 do
        local f = rusage_fields[i]
        local v = timeval(self[f])
        ttime = ttime + v
        local va = timeval(all[f])
        ttime_all = ttime_all + va
        local pct = va == 0 and '-' or string.format('%.02f%%', 100 * v / va)
        local wpct = wall == 0 and '-' or
            string.format('%.02f%%', 100 * v / wall)
        o[#o + 1] = string.format('%s: %.6f (%s) (wall: %s)\n',
            rusage_labels[i], v, pct, wpct)
    end
    if ttime_all > 0 then
        o[#o + 1] = string.format(
            'total time:: %.6f (%.02f%%) (wall: %.02f%%)\n',
            ttime, 100 * ttime / ttime_all, 100 * ttime / wall)
    end
    for i = 3, #rusage_fields do
        local f = rusage_fields[i]
        local v = tonumber(self[f])
        local va = tonumber(all[f])
        local pct = va == 0 and '-' or string.format('%.02f%%', 100 * v / va)
        o[#o + 1] = string.format('%s: %d (%s)\n', rusage_labels[i], v, pct)
    end
    return table.concat(o)
end

local ctor = ffi.metatype('struct rusage', {__index = mt})

require 'middleclass'

local m = class('profuser')

function m:initialize(filename, who)
    self.timeval = ffi.new('struct timeval')
    self.start_time = self:gettime()
    self.tags = {}
    self.ofh = filename and assert(io.open(filename, 'w+')) or io.stderr
    who = who or C.RUSAGE_SELF
    self.rbufs = ffi.new('struct rusage[3]')
    self.current = 0
    self.who = who
    self.start_time = os.date('%Y-%m-%dT%H:%M:%S')
    self:getrusage()
end

function m:gettime()
    local tv = self.timeval
    assert(C.gettimeofday(tv, nil) == 0)
    return tonumber(tv.tv_sec) + tonumber(tv.tv_usec) * 1e-6
end

function m:getrusage()
    assert(C.getrusage(self.who, self.rbufs + self.current) == 0)
    self.current = 1 - self.current
end

function m:get_tag(tag)
    if not self.tags[tag] then
        local t = {}
        t.calls = 0
        t.enter = ffi.new('struct rusage')
        t.leave = ffi.new('struct rusage')
        t.diff = ffi.new('struct rusage')
        t.accum = ffi.new('struct rusage')
        self.tags[tag] = t
    end
    return self.tags[tag]
end

function m:enter(tag)
    local t = self:get_tag(tag)
    t.calls = t.calls + 1
    assert(C.getrusage(self.who, t.enter) == 0)
end

function m:leave(tag)
    local t = self:get_tag(tag)
    assert(C.getrusage(self.who, t.leave) == 0)
    t.leave:sub(t.enter, t.diff)
    t.accum:add(t.diff, t.accum)
end

function m:printf(fmt, ...)
    self.ofh:write(string.format(fmt, ...))
end

function m:do_unaccounted()
    local unacc = self:get_tag('__unaccounted')
    ffi.fill(unacc.accum, ffi.sizeof('struct rusage'))
    for k, t in pairs(self.tags) do
        if not k:match('^__') then
            unacc.accum:add(t.accum, unacc.accum)
        end
    end
end

function m:dump()
    local now = self:gettime()
    local elapsed = now - self.start_time
    self:leave('__main')
    self:do_unaccounted()

    local all = self:get_tag('__main')

    self:printf('BEGIN DUMP\n')
    local n = {}
    for k, v in pairs(self.tags) do
        n[#n + 1] = k
    end
    table.sort(n, function(a, b)
        local ta, tb = self.tags[a], self.tags[b]
        return timeval(ta.accum.ru_utime) < timeval(tb.accum.ru_utime)
    end)
    for i = 1, #n do
        local tag = n[i]
        local t = self:get_tag(tag)
        local v = all and t.accum:dump_percentage(all.accum, elapsed) or
            t.accum:dump()

        self:printf('######## %s (%d calls) elapsed=%.6f ########\n%s\n',
            tag, t.calls, elapsed, v)
    end
    self:printf('END DUMP\n')
    self.ofh:flush()
end

return m
