require 'strict'
local __ = require '__'
local log = require 'log'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'
local parser = ffi.load('libparser.so')
local rational = require 'rational'
local tablex = require 'tablex'

ffi.cdef[[
typedef int (*FXN1)(char*, int*);
extern int ParseInFixExpr(char *InFixExpr, int RequireInputChs, FXN1 func1, 
                          int *ExpressionType, char *ErrInfo,
                          int MaxCharsErrInfo, char *OutFixQuotedExpr);
extern int GetChanInfo(char *name, int *chan_type);
]]


require 'middleclass'

local M = class('setup_spec')

function M:initialize(setup)
    self.setup = setup
end

local function errorf(msg)
    local msg = "SXR file parser error: "..msg
    log:log(log.ERR, msg)
    error(msg)
end

function M:import(input)
    local ok, setup = pcall(cjson.decode, input)
    if not ok then
        errorf("Input is invalid JSON: "..setup)
    elseif not setup then
        errorf("Setup is nil")
    end
    self.setup = setup
    return self.setup
end

function M:validate_structure()
    local setup = self.setup
    assert(type(setup) == 'table')
    local max = 0
    for k, v in pairs(setup) do
        assert(type(k) == 'number' and k >= 1 and math.floor(k) == k)
        max = math.max(k, max)
        assert(type(v) == 'table')
        for ik, iv in pairs(v) do
            assert(type(ik) == 'string')
        end
    end
    assert(#setup == max)
    log:notice("SXR file parser: basic structure validation completed")
end

-- These strings, if found empty, are not deleted per the rule described below.
local keep_strings = {
    sample_rate = true,
    request_rate = true,
    bit_match_mask = true,  -- (probably not used) for 0xXXXX bit message
    bit_match_value = true, -- (probably not used) for 0xXXXX bit message
    name = true,
    db_msg_id = true,
    db_msg_mask = true,
    output_units = true,
    measurand = true,
    description = true,
}

local function input_channel_table_normalize(t)
    local nt = {}
    for _, v in ipairs(t) do
        if type(v) == 'string' then
            nt[#nt + 1] = v
        else
            nt[#n1 + 1] = '"' .. v .. '"'
        end
    end
    return nt
end

-- Preserve the sensor table as is
local function sensor_table_normalize(t)
    return t
end

-- This lists tables within the JSON which normalize must handle specially.
local normalize_tables = {
    input_channels = input_channel_table_normalize,
    sensor = sensor_table_normalize,
    teds_sensor = sensor_table_normalize,
    database_sensor = sensor_table_normalize,
}

-- Some tables are not processed, e.g. sensor tables, because certain fields
--      such as id will evaluate wrongly to number values.
-- Also, note that empty strings coming from the GUI are returned as nil
--      because apprently the GUI has no capacity to generate JSON nil.
--      Strings which must be kept as fields when empty are entered into the
--      keep_strings table above.
local function normalize(v, k)
    local t = type(v)
    if t == 'number' or t == 'boolean' then
        return v
    elseif t == 'string' then
        if keep_strings[k] then
            return v
        end
        local tv = tonumber(v)
        if tv then
            return tv
        elseif v == 'true' then
            return true
        elseif v == 'false' then
            return false
        elseif v == '' then
            return nil
        end
        return v
    elseif t == 'table' then
        -- If this table needs special processing, do so
        for k1, f in pairs(normalize_tables) do
            if k1 == k then return f(v) end
        end
        local nt = {}
        for key, val in pairs(v) do
            nt[key] = normalize(val, key)
        end
        return nt
    elseif v == cjson.null then
        return nil
    else
        return nil
    end
end

function M:normalize(setup)
    setup = setup or self.setup
    return normalize(setup)
end

-- Default separator handling is everything crammed together on one line.
-- The following are exceptions to that rule for the top two levels as
-- a reasonable compromise between compactness and readability.
local separators = {
    [1] = {
        [']'] = ']\n',
    },
    [2] = {
        ['{'] = '{\n',
        ['}'] = '\n}',
        [','] = ',\n',
        [':'] = ': ',
        [''] = ' ',
    },
}

function M:emit(s)
    self.out[#self.out + 1] = s
end

function M:emit_punct(s)
    if separators[self.level] then
        s = separators[self.level][s] or s
    end
    self:emit(s)
end

function M:export_item(item)
    if type(item) == 'table' then
        self.level = self.level + 1
        if item[1] ~= nil then
            self:emit_punct('[')
                for i = 1, #item do
                    if i > 1 then
                        self:emit_punct(',')
                    end
                    self:export_item(item[i])
                end
            self:emit_punct(']')
        else
            local keys = {}
            for k, v in pairs(item) do
                keys[#keys + 1] = k
            end
            table.sort(keys)
            self:emit_punct('{')
            for i = 1, #keys do
                self:emit_punct('')
                self:emit(cjson.encode(tostring(keys[i])))
                self:emit_punct(':')
                self:export_item(item[keys[i]])
                if i < #keys then
                    self:emit_punct(',')
                end
            end
            self:emit_punct('}')
        end
        self.level = self.level - 1
    else
        self:emit(cjson.encode(item))
    end
end

local function basic_item_check(item)
    local err = "Bad setup file format: channel '"
    if item == nil then
        errorf(err.."' returned nil")
    elseif not item.category then
        errorf(err.."' is missing field: category")
    elseif item.category ~= 'summary' and not item.name then
        errorf(err.."<type "..item.category..">".."' is missing field: name")
    elseif not item.id then
        errorf(err..(item.name or item.category).."' is missing field: id")
    elseif not item.sort_order then
        errorf(err..(item.name or item.id).."' is missing field: sort_order")
    elseif item.category == 'summary' and item.database then
        errorf("Bad setup file format: summary cannot contain field database")
    elseif item.category ~= 'summary' and not item.category:match("sink")
        and not item.data_type then
        errorf(err..item.name.."' is missing field: data type")
    elseif item.category ~= 'summary' and not item.category:match("sink")
        and item.data_type ~= 'sequential_float32'
        and item.data_type ~= 'sequential_float64'
        and item.data_type ~= 'message_float32'
        and item.data_type ~= 'message_float64'
        and item.data_type ~= 'message_float32_ntp'
        and item.data_type ~= 'message_float64_ntp'
        and item.data_type ~= 'message_can'
        and item.data_type ~= 'message_raw' then
        errorf(err..item.name.."' has incorrect field: data type")
    elseif item.category ~= 'summary' and item.data_type ~= 'message_raw'
        and item.data_type ~= 'message_can' and not item.category:match("sink")
        and not item.data_type:find('message_float')
        and not item.sample_rate then
        errorf(err..item.name.."' is missing field: sample rate")
    end

    if item.category == 'summary' and not item.pipe_rate then
        item.pipe_rate = 5
    end
    if item.category ~= 'summary' and item.data_type ~= 'message_raw' and
        item.data_type ~= 'message_can' and not item.category:match("sink") then
        if not item.output_units then
            item.output_units = ""
        end
        if not item.measurand then
            item.measurand = ""
        end
    end
    if item.category ~= 'summary' and not item.description then
        item.description = ""
    end
end

local function toboolean(b)
    if type(b) == 'string' then
        b = string.lower(b)
    end
    if b == true or b == "true" then
        return true
    elseif b == false or b == "false" then
        return false
    else
        return nil
    end
end

local function escape_magic(str)
    local special = { "%%", "%^", "%$", "%(", "%)", "%.",
                      "%[", "%]", "%*", "%+", "%-", "%?"}
    for i=1,#special do
        str = (str:gsub(special[i], (special[i]:gsub("%%","%%%%"))))
    end
    return str
end

function M:user_order(setup)
    setup = setup or self.setup
    local temp, unordered, conflict = {}, {}, {}
    for i = 1, #setup do
        if setup[i].category == "summary" then
            if temp[1] ~= nil then
                if temp[1].category == "summary" then
                    conflict[#conflict+1] = setup[i]
                else
                    conflict[#conflict+1] = temp[1]
                    temp[1] = setup[i]
                end
            else
                temp[1] = setup[i]
            end
        elseif tonumber(setup[i].sort_order) then
            if  temp[tonumber(setup[i].sort_order)+1] ~= nil then
                conflict[#conflict+1] = setup[i]
            else
                temp[tonumber(setup[i].sort_order)+1] = setup[i]
            end
        else
            unordered[#unordered+1] = setup[i]
        end
    end
    setup = tablex.array_concat(tablex.collect_sparse_array(temp), conflict)
    setup = tablex.array_concat(setup, unordered)
    for i = 1, #setup do
        setup[i].sort_order = i - 1
    end
    return setup
end

function M:test_order(setup)
    setup = setup or self.setup
    local category = { "summary",
                       "source/hbm_edaqxr/msg_",
                       "source/hbm_edaqxr/gps",
                       "source/hbm_edaqxr/can",
                       "source/hbm_edaqxr/brg",
                       "source/hbm_edaqxr/dio/input",
                       "source/hbm_edaqxr/dio/pulse",
                       "source/hbm_edaqxr/hls",
                       "source/hbm_edaqxr/itb",
                       "source/hbm_edaqxr/ntb",
                       "source/hbm_edaqxr/power",
                       "source/hbm_edaqxr/system",
                       "source/hbm_cx23r/msg_",
                       "source/hbm_cx23r/gps",
                       "source/hbm_cx23r/can",
                       "source/hbm_cx23r/vbm",
                       "source/hbm_cx23r/power",
                       "source/hbm_cx23r/system",
                       "source/hbm_cx23r/dio_input",
                       "source/hbm_edaqxr/digital_input",
                       "source/hbm_dx23r/dio_input",
                       "source/hbm_mx1615b",
                       "source/hbm_mx1601b",
                       "source/hbm_mx1609b",
                       "source/hbm_mx411b",
                       "source/hbm_mx460b",
                       "source/hbm_mx471b",
                       "source/hbm_mx840b",
                       "sink/hbm_mx878b",
                       "filter/",
                       "sink/hbm_cx23r/digital_output",
                       "sink/hbm_edaqxr/digital_output",
                       "sink/hbm_edaqxr/dio/output",
                       "sink/sie/" }
    local bin = {}
    for _, _ in ipairs(category) do bin[#bin + 1] = {} end
    local category_lookup = {}
    for k,v in pairs(category) do category_lookup[v] = tonumber(k) end

    local lookup = {}

    for i = 1, #setup do
        basic_item_check(setup[i])
        local categ = 0

        for j = 1, #category do
            if setup[i].category:match( category[j] ) then 
                categ = j
                break
            elseif j == #category then
                errorf("Bad setup file format, channel "..setup[i].name..
                       " unknown item category '"..setup[i].category.."'")
            end
        end

        if (setup[i].category:match("filter/") or
            setup[i].category:match("sink/")) and
            not setup[i].category:find("sink/hbm_mx878") then
            
            if setup[i].category:match("sink/sie/") and not
                tonumber(setup[i].sort_order) then
                errorf("Bad setup file format, channel "..setup[i].name..
                       " missing sort_order")
            end

            local refs = tablex.array_concat({}, setup[i].input_channels or {})
            refs[#refs+1] = type(setup[i].trigger_channel) == 'string'
                and #setup[i].trigger_channel > 1
                and setup[i].trigger_channel or nil
            setup[i].all_input_channels = tablex.array_concat({}, refs or {})
            local missing_definition = false

            for j = 1, #refs do
                if not lookup[refs[j]] then
                    bin[categ].queue = bin[categ].queue or {}
                    bin[categ].queue[#bin[categ].queue + 1] = setup[i]
                    missing_definition = true
                    break
                end
            end
            if not missing_definition then
                if category[categ] == "sink/sie/" then
                    bin[categ][tonumber(setup[i].sort_order)] = setup[i]
                else
                    bin[categ][#bin[categ] + 1] = setup[i]
                end
                lookup[tostring(setup[i].name)] = setup[i]
            end
        elseif category[categ] == "summary" then
            bin[categ][#bin[categ] + 1] = setup[i]
        else
            bin[categ][#bin[categ] + 1] = setup[i]
            lookup[tostring(setup[i].name)] = setup[i]
        end
    end

    local summary = category_lookup["summary"]
    local filter = category_lookup["filter/"]
    local sink_sie = category_lookup["sink/sie/"]

    for categ = filter, sink_sie do
        local queue = bin[categ].queue or {}
        local queuesize = #queue
        local iter = 0
        local missing_definition
        while #queue > 0 and iter <= queuesize do
        
            for i = #queue,1,-1 do
                missing_definition = false
                
                for j = 1, #queue[i].all_input_channels do
                    if not lookup[tostring(queue[i].all_input_channels[j])] then
                        missing_definition = "'"..queue[i].all_input_channels[j]
                                             .."' by channel "..queue[i].name
                        break
                    end
                end
                
                if not missing_definition then
                    lookup[tostring(queue[i].name)] = queue[i]
                    if queue[i].category:match("sink/sie/") then
                        bin[categ][tonumber(queue[i].sort_order)] =
                            table.remove(queue, i)
                    else
                        bin[categ][#bin[categ] + 1] = table.remove(queue, i)
                    end
                end
            end
            
            iter = iter + 1
        end
        
        if #queue ~= 0 then
            errorf("Bad setup file, undefined input channel "..
                    missing_definition)
        end
    end
    
    bin[summary][#bin[summary]].enable_sync_stops = true
    for i = 1, #bin[filter] do
        if bin[filter][i].category:match("filter/time_base_shifter") or
           bin[filter][i].category:match("filter/smooth_filter") then
            bin[summary][#bin[summary]].enable_sync_stops = false
        end
    end
    
    for i = 1, #bin[filter] do
        if bin[filter][i].category:match("filter/expression") then
            local name = bin[filter][i].name
            local input_expr   = ffi.new('char[4096]',bin[filter][i].expression)
            local output_expr  = ffi.new('char[4096+2*256]')
            local expr_type    = ffi.new("int[1]")
            local error_string = ffi.new('char[513]')
            local success = parser.ParseInFixExpr(input_expr, 0,
                parser.GetChanInfo, expr_type, error_string, 512, output_expr)
            if tonumber(success) == 0 then
                errorf("Invalid expression in channel "..name..", "
                       .."\n"..ffi.string(output_expr))
            else
                bin[filter][i].expression = ffi.string(output_expr)
            end
            
            local check = bin[filter][i].expression
            local c, _c
            local sample_rate
            local a, num_a, den_a, b, num_b, den_b
            for chan = 1, #bin[filter][i].input_channels do
                c  = tostring(bin[filter][i].input_channels[chan])
                _c = "'"..c.."'"
                if not lookup[c].sample_rate then
                    errorf("Error on channel "..lookup[c].name..
                           ": Sample rate undefined")
                end
                
                if not sample_rate then
                    sample_rate = lookup[c].sample_rate
                end
                a,num_a, den_a = rational:new(sample_rate):tonumbers()
                b,num_b, den_b = rational:new(lookup[c].sample_rate):tonumbers()
                
                if num_a ~= num_b or den_a ~= den_b then
                    errorf("Sample rate mismatch in expression '"..name.."' "..
                        "input channel '"..lookup[c].name.."' had sample rate"..
                        " of "..tostring(b).." when expecting "..tostring(a))
                end
                
                if not bin[filter][i].expression:match(escape_magic(_c)) then
                    log:notice("SXR file parser: expression "..bin[filter][i].expression)
                    errorf("Internal file format error: channel ".._c.." in "..
                        "input list not found in expression for channel "..name)
                end
                
                if toboolean(lookup[c].is_logical) == true then
                    check = (check:gsub( escape_magic(_c),'b') )
                else
                    check = (check:gsub( escape_magic(_c),'c') )
                end
            end
            
            input_expr = ffi.new('char[4096]',check)
            success = parser.ParseInFixExpr(input_expr, 0, parser.GetChanInfo,
                expr_type, error_string, 512, output_expr)
            if tonumber(success) == 0 then
                errorf("Invalid expression in channel "..name..", "
                       .."\n"..ffi.string(error_string))
            end

            if tonumber(expr_type[0]) == 2 then
                bin[filter][i].is_logical = false
            elseif tonumber(expr_type[0]) == 1 then
                bin[filter][i].is_logical = true
            else
                errorf("Invalid expression, cannot resolve is_logical for "..
                    " channel '"..name.."'\nexpr_type:"..tonumber(expr_type[0]))
            end
            
        end
    end
    
    bin[sink_sie].queue = nil
    bin[sink_sie] = tablex.collect_sparse_array(bin[sink_sie])
    
    local setup = {}
    for i = 1, #bin do
        setup = tablex.array_concat(setup, bin[i])
    end
    log:notice("SXR file parser: main verification completed")
    return normalize(setup)
end

function M:export(setup)
    setup = setup or self.setup
    setup = setup == "user_order" and self:user_order() or setup
    setup = setup == "test_order" and self:test_order() or setup
    self.out = {}
    self.level = 0
    self:export_item(setup)
    return table.concat(self.out)
end

function M:item_export(item)
    self.out = {}
    self.level = 1
    self:export_item(item)
    return table.concat(self.out)
end

return M
