#!/usr/bin/luajit

require 'strict'
local __ = require "__"
local cjson = require "cjson.safe"
local ffi = require "ffi"
local libsie = require 'libsie_ffi'
local log = require 'log'
local ss = require 'setup_spec'
local notify = require 'sys_notify'
local lfs = require 'lfs'

--Purpose: to extract XML tag data from an SIE file for API calls.

--USES
  -- Extract setup file from SIE
  -- Extract taglist/tag-sublist from SIE file
  -- Extract a particular file tag from SIE file
  
local function errorf(msg)
    local msg = "SIE parser error:  "..msg
    log:log(log.ERR, msg)
    error(msg)
end

require 'middleclass'

local M = class('sie_parser')

function M:initialize(filename)
    if not filename then
        return errorf("'New' called without reference to SIE file")
    end
    self.filename = filename
end

function M:_open(filename, itg)
    local filename = self.filename or filename
    local context = libsie.sie_context_new()
    if itg then libsie.sie_ignore_trailing_garbage(context, itg) end
    local file = libsie.sie_file_open(context, filename)
    if file == nil then
        local exception = libsie.sie_get_exception(context)
        return errorf(ffi.string(libsie.sie_verbose_report(exception))..'\n')
    end
    return context, file
end

function M:open(filename)
    self.struct = {}
    --self.filename = filename or self.filename
    local filename = self.filename or filename
    if not filename then
        return errorf("'Open' called without reference to SIE file")
    end
    local stat = lfs.attributes(filename)
    if not stat then
        return errorf("Unable to get SIE file info")
    end

    if stat.size < 1000000 then
        -- If the file is less than 1M just deal with it directly
        self.context, self.file = self:_open(filename)
    else
        -- If the file is more than 1M use only the beginning of the file
        local f = io.open(filename, 'r')
        filename = string.gsub(filename, '/storage/', '/tmp/.')
        local f2 = io.open(filename, 'w')
        -- Make sure we get at least to the test end XML tag
        local total = 0
        while true do
            local d = f:read(1000000)
            f2:write(d)
            total = total + 1000000
            if d:find('</test>') then
                local d = f:read(1000000)
                f2:write(d)
                total = total + 1000000
                f:close()
                f2:close()
                break
            end
        end
        self.context, self.file = self:_open(filename, total)
    end
end

function M:close()
    local context = self.context
    local file = self.file
    if libsie.sie_check_exception(context) ~= nil then
        local exception = libsie.sie_get_exception(context)
        errorf("Exception:  "..ffi.string(libsie.sie_verbose_report(exception)))
        libsie.sie_release(exception)
    end

    libsie.sie_release(file)

    local leaked_objects = libsie.sie_context_done(context)
    if leaked_objects ~= 0 then
        errorf("Warning:  Leaked "..leaked_objects.." SIE objects!")
    end
end

local function get_tag(tag, long_name)
    local tag_value = ffi.new("const char *[?]", 1)
    local tag_size =  ffi.new("size_t[?]", 1)
    local ok = libsie.sie_tag_get_value_b(tag, tag_value, tag_size)
    if ok ~= 0 then
        local name = ffi.string(libsie.sie_tag_get_id(tag))

        local value = (tag_size[0] <= 50 or long_name == name)
                      and ffi.string(tag_value[0], tag_size[0])
                      or "long tag of "..tonumber(tag_size[0]).." bytes."
                      
        if long_name == name and long_name == "sie:block_index" then
            value = ""
            for byte = 1, tag_size[0] do
                value = value .. bit.tohex( string.byte(ffi.string(tag_value[0],
                    tag_size[0]):sub(byte,byte)), 2) .. " "
                if byte%12 == 0 then 
                    value = value .. "\n"
                end
            end
        end
        
        libsie.sie_free(ffi.cast("void *", tag_value[0]))
        
        if long_name and long_name ~= name then
            return
        end
        
        return name, value
    end
end

function M:_get_file_tags(tag_to_get)
    local tag_iterator = libsie.sie_get_tags(self.file)
    local tag, name, value
    
    while true do
        tag = libsie.sie_iterator_next(tag_iterator)
        if tag == nil then
            break
        end
        name, value = get_tag(tag, tag_to_get)
        if tag_to_get == name then
            break
        end
        if name then
            self.struct[name] = value
            --To be helpful, show a tag level reference to the setup name w list
            if name == "somat:sxr_setup" and tag_to_get == nil then
                self.struct["somat:setup_name"] =
                    cjson.decode(table.remove({get_tag(tag,
                    "somat:sxr_setup")},2))[1].test_name
            end
        end
    end
    libsie.sie_release(tag_iterator)
    if tag_to_get then
        return value
    end
end

function M:_get_test_tags(selected_tag_type)
    self.struct.tests = self.struct.tests or {}
    local tag_to_get = selected_tag_type ~= "all" and
                       selected_tag_type ~= "test" and
                       selected_tag_type ~= "channel" and
                       selected_tag_type
    --It is possible but unimplemented to return all tag matches in a table
    local tag_value
    if tag_to_get then
--~         selected_tag_type = 'test'
        selected_tag_type = 'all'
        tag_value = {}
    end
    
    local test_iterator = libsie.sie_get_tests(self.file)
    local test, tid
    while true do
        test = libsie.sie_iterator_next(test_iterator)
        if test == nil then
            break
        end
        tid = #self.struct.tests + 1
        self.struct.tests[tid] = {}

        if selected_tag_type == "all" or selected_tag_type == "test" then
            local tag_iterator = libsie.sie_get_tags(test)
            local tag, name, value
            --self.struct.tests[tid].id = tostring(libsie.sie_get_id(test))
            while true do
                tag = libsie.sie_iterator_next(tag_iterator)
                if tag == nil then
                    break
                end
                name, value = get_tag(tag, tag_to_get)
                if tag_to_get == name then
                    tag_value[#tag_value+1] = value
--~                     tag_value = value
--~                     break
                end
                if name then
                    self.struct.tests[tid][name] = value
                end
            end
            libsie.sie_release(tag_iterator)
        end
        
--~         if tag_to_get then
--~             break
--~         end

        if selected_tag_type == "all" or selected_tag_type == "channel" then
            self.struct.tests[tid].channels = {}
            local channel_iterator = libsie.sie_get_channels(test)
            local channel, cid
            while true do
                channel = libsie.sie_iterator_next(channel_iterator)
                if channel == nil then
                    break
                end
                
                cid = #self.struct.tests[tid].channels + 1
                self.struct.tests[tid].channels[cid] = {}
                --self.struct.tests[tid].channels[cid].id = tostring(
                --    libsie.sie_get_id(channel))

                local tag_iterator = libsie.sie_get_tags(channel)
                local tag, name, value
                while true do
                    tag = libsie.sie_iterator_next(tag_iterator)
                    if tag == nil then
                        break
                    end
                    name, value = get_tag(tag, tag_to_get)
                    if tag_to_get == name then
                        tag_value[#tag_value+1] = value
                    end
                    if name then
                        self.struct.tests[tid].channels[cid][name] = value
                    end
                end
                libsie.sie_release(tag_iterator)
                
                self.struct.tests[tid].channels[cid].dimensions = {}
                local dimension_iterator = libsie.sie_get_dimensions(channel)
                local dimension, did
                while true do
                    dimension = libsie.sie_iterator_next(dimension_iterator)
                    if dimension == nil then
                        break
                    end

                    did = #self.struct.tests[tid].channels[cid].dimensions + 1
                    self.struct.tests[tid].channels[cid].dimensions[did] = {}
                    --self.struct.tests[tid].channels[cid].dimensions[did].id =
                    --    tostring(libsie.sie_get_index(dimension))
                    
                    local tag_iterator = libsie.sie_get_tags(dimension)
                    local tag, name, value
                    while true do
                        tag = libsie.sie_iterator_next(tag_iterator)
                        if tag == nil then
                            break
                        end
                        name, value = get_tag(tag, tag_to_get)
                        if tag_to_get == name then
                            tag_value[#tag_value+1] = value
                        end
                        if name then
                            self.struct.tests[tid].channels[cid].dimensions
                                [did][name] = value
                        end
                    end
                    libsie.sie_release(tag_iterator)

                end
                libsie.sie_release(dimension_iterator)

                --if selected_tag_type == "data"
                --    local spigot = libsie.sie_attach_spigot(channel)
                --    local output = libsie.sie_spigot_get(spigot)
                --    local outs = libsie.sie_output_get_struct(output)
                --    self.struct.tests[tid+1].channels[cid+1].output = outs
                --    libsie.sie_release(spigot)
                --end
            end
            libsie.sie_release(channel_iterator)
        end
    end
    libsie.sie_release(test_iterator)
    
    if tag_to_get then
        return table.concat(tag_value, '\n')
    end
end

--high level functions
function M:get_tags(filename)
    self:open(filename)
    self:_get_file_tags()
    self:_get_test_tags('all')
    self:close()
    return cjson.encode(self.struct)
end

function M:get_file_tags(filename)
    self:open(filename)
    self:_get_file_tags()
    self:close()
    return cjson.encode(self.struct)
end

function M:get_test_tags(filename)
    self:open(filename)
    self:_get_test_tags("test")
    self:close()
    return cjson.encode(self.struct.tests)
end

function M:get_channel_tags(filename)
    self:open(filename)
    self:_get_test_tags("channel")
    self:close()
    local ret = {}
    if #self.struct.tests > 1 then
        ret[#ret+1] = self.struct.tests[#ret+1].channels
    else
        ret = self.struct.tests[1].channels
    end
    return cjson.encode(ret)
end

function M:get_tag_by_name(tagname, filename)
    self:open(filename)
    local tag = self:_get_file_tags(tagname) or
                self:_get_test_tags(tagname)
    self:close()
    return tag
end

--Some specific tags
function M:get_system_version(filename)
    self:open(filename)
    local version = self:_get_file_tags("somat:system_version")
    self:close()
    return version
end

function M:get_setup(filename)
    self:open(filename)
    local setup = self:_get_file_tags("somat:sxr_setup")
    self:close()
    return setup
end

function M:get_source_setup(filename)
    self:open(filename)
    local setup = self:_get_file_tags("somat:source_sxr_setup")
    self:close()
    return setup
end

function M:get_summary(filename)
    self:open(filename)
    local summary = cjson.decode(self:_get_file_tags("somat:sxr_setup"))[1]
    self:close()
    return cjson.encode(summary)
end

function M:get_source_summary(filename)
    self:open(filename)
    local summary = cjson.decode(self:_get_file_tags("somat:source_sxr_setup"))[1]
    self:close()
    return cjson.encode(summary)
end

function M:get_summary_field(field, filename)
    self:open(filename)
    local summary = cjson.decode(self:_get_file_tags("somat:sxr_setup"))[1]
    self:close()
    return cjson.encode(summary[field])
end

function M:get_source_summary_field(field, filename)
    self:open(filename)
    local summary = cjson.decode(self:_get_file_tags("somat:source_sxr_setup"))[1]
    self:close()
    return cjson.encode(summary[field])
end

return M
