require 'strict'
local ffi = require 'ffi'
local C = ffi.C
local config = require 'system_config'
local libte = ffi.load(config.test_engine_library)

local ffi_sizeof = ffi.sizeof
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy

if config.test_engine_library == '/usr/lib/libte-debug.so' then
ffi.cdef[[
    typedef int64_t bh_offset_t;

    struct hbms_hdr {
        uint32_t        hdr;            // HBM stream header info
//        uint32_t        ts;             // timestamp of first sample in frame
//      FIXME:  Make the ts below a double sometime
        float           ts;             // timestamp of first sample in frame
    } __attribute__((packed));

    struct frame {
        struct frame    *next;          // Next frame in list
        struct frame    *prev;          // Previous frame in list
        struct frame    *root;          // Pointer to root frame (clones only)
        struct frame    **ff_list;      // Pointer to free frame list
        uint16_t        msg_frame;      // != 0 is a message frame
        int16_t         latch;          // reference count
        uint32_t        element_size;   // Size of elements in buffer
        uint32_t        maxsize;        // Maximum elements in frame
        uint32_t        count;          // Number of elements in frame
        bh_offset_t     offset;         // Trigger offset (by Burst History)
        int64_t         ts;             // Total sample count less this frame
        uint32_t        sequence_num;   // Frame sequence number for debug
        void            *buf;           // Buffer containing elements
        struct hbms_hdr *hdr;           // HBM stream header
    } __attribute__((packed));
]]
else
ffi.cdef[[
    typedef int64_t bh_offset_t;

    struct hbms_hdr {
        uint32_t        hdr;            // HBM stream header info
//        uint32_t        ts;             // timestamp of first sample in frame
//      FIXME:  Make the ts below a double sometime
        float           ts;             // timestamp of first sample in frame
    } __attribute__((packed));

    struct frame {
        struct frame    *next;          // Next frame in list
        struct frame    *prev;          // Previous frame in list
        struct frame    *root;          // Pointer to root frame (clones only)
        struct frame    **ff_list;      // Pointer to free frame list
        uint16_t        msg_frame;      // != 0 is a message frame
        int16_t         latch;          // reference count
        uint32_t        element_size;   // Size of elements in buffer
        uint32_t        maxsize;        // Maximum elements in frame
        uint32_t        count;          // Number of elements in frame
        bh_offset_t     offset;         // Trigger offset (by Burst History)
        int64_t         ts;             // Full precision timestamp
        void            *buf;           // Buffer containing elements
        struct hbms_hdr *hdr;           // HBM stream header
    } __attribute__((packed));
]]
end

ffi.cdef[[
    void *calloc(size_t nmemb, size_t size);
    void free(void *ptr);

    // Prototypes for functions in libte
    struct frame *frameCreate(int bufsize, int element_bytes, void *buf);
    void frameDelete(struct frame *frame);
    struct frame *frameClone(struct frame *root);
    int frameBufsize(double sample_rate, double pipe_frame_rate);
    int frameLength(struct frame *frame);
    int frameAvailable(struct frame *frame);
    char *frameTail(struct frame *frame);
]]


if config.use_libte then
    return {
        create = function(bufsize, element_bytes)
            return libte.frameCreate(bufsize, element_bytes, nil)
        end,
        delete = libte.frameDelete,
        clone = libte.frameClone,
        bufsize = libte.frameBufsize,
        length = libte.frameLength,
        available = libte.frameAvailable,
        tail = libte.frameTail,
    }
end

local C_calloc = C.calloc
local C_free = C.free

local m = {}

-- Create a normal frame
function m.create(bufsize, element_bytes)
    local f = ffi_cast('struct frame *',
        C_calloc(1, ffi_sizeof('struct frame') +
        ffi_sizeof('struct hbms_hdr') + (bufsize * element_bytes)))
    if f == nil then
        error("Can not allocate memory for frame")
        return
    end

    f.element_size = element_bytes
    f.latch = 1
    f.maxsize = bufsize
    f.hdr = ffi_cast('struct hbms_hdr *', f + 1)
    f.buf = ffi_cast('char *', f.hdr) + ffi_sizeof('struct hbms_hdr')

    return f
end

--[[
-- Create a normal frame
function m.create(bufsize, element_bytes, buf)
    local f
    if buf ~= nil then
        f = ffi_cast('struct frame *',
            C_calloc(1, ffi_sizeof('struct frame') +
            ffi_sizeof('struct hbms_hdr')))
    else
        f = ffi_cast('struct frame *',
            C_calloc(1, ffi_sizeof('struct frame') +
            ffi_sizeof('struct hbms_hdr') + (bufsize * element_bytes)))
    end
    if f == nil then
        error("Can not allocate memory for frame")
        return
    end

    f.element_size = element_bytes
    f.latch = 1
    f.maxsize = bufsize
    f.hdr = ffi_cast('struct hbms_hdr *', f + 1)
    f.buf = ffi_cast('char *', (buf or f.hdr)) + ffi_sizeof('struct hbms_hdr')

    return f
end

-- Create a message frame
function m.create_message(buf, bufsize)
    local f = m.create(bufsize, 1, buf)
    f.msg_frame = 1
    return f
end
]]

-- Delete a frame
function m.delete(frame)
    if not frame or frame == nil then
        error("Unable to delete frame:  argument required")
    end
    frame.latch = frame.latch - 1
    local root = frame.root
    if root == nil then
        if frame.latch == 0 then
            if frame.msg_frame ~= 0 then
                if frame.hdr ~= nil then C_free(frame.hdr) end
            end
            C_free(frame)
        end
    else
        m.delete(root)
        if frame.latch == 0 then
            C_free(frame)
        end
    end
end

-- Clone a root frame
function m.clone(root)
    if not root or root == nil then
        error("Unable to clone frame:  argument required")
    end

    -- If cloning from a clone, clone the root of that clone
    if root.root ~= nil then
        return m.clone(root.root)
    end

    local clone = ffi_cast('struct frame *',
        C_calloc(1, ffi_sizeof('struct frame')))
    if clone == nil then error("Can not allocate memory for clone") end
    ffi_copy(clone, root, ffi_sizeof('struct frame'))

    clone.latch = 1
    clone.root = root

    local curframe = root
    while curframe ~= nil do
        curframe.latch = curframe.latch + 1
        curframe = curframe.root
    end

    return clone
end

-- Determine frame buffer size (in samples) based on sample rate and pipe
-- frame rate
function m.bufsize(sample_rate, pipe_frame_rate)
    local size = math.floor(sample_rate / pipe_frame_rate)
    return (size > 1 and size) or 1
end

function m.length(frame)
    return frame.count
end

function m.available(frame)
    return frame.maxsize - frame.count
end

function m.tail(frame)
    return ffi_cast('char *', frame.buf) + (frame.element_size * frame.count)
end

return m
