require 'middleclass'
local ev = require'ev'
local http_server = require 'http_server'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'
local cjson = require 'cjson.safe'
local Frame = require 'test_engine/frame'
local Pipe = require 'test_engine/pipe'
local MinMax = require 'test_engine/min_max'
local tcp_server = require 'net/tcp_server'
local tcp_stream = require 'net/tcp_stream'
local ws_server = require 'net/ws_server'

local Pipe_attach_pipe = Pipe.attach_pipe
local Pipe_detach_pipe = Pipe.detach_pipe
local Pipe_get_frame = Pipe.get_frame
local Pipe_flush_frames = Pipe.flush_frames
local Frame_delete = Frame.delete
local Frame_create = Frame.create

local ffi_sizeof = ffi.sizeof
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_string = ffi.string
local ffi_new = ffi.new

local htonl = ffi.abi('le') and bit.bswap or bit.tobit

local min_max_data = ffi_new('struct min_max_data[1]')

ffi.cdef[[
typedef struct hbm_stream_packet {
    unsigned long       hdr;
    unsigned long       length;
    unsigned char       data[0];
} hbm_stream_packet_t;

typedef struct hbm_small_stream_packet {
    unsigned long       hdr;
    unsigned char       data[256];
} hbm_small_stream_packet_t;

enum {
    HDR_DATA = 0x10000000,
    HDR_METAINFO = 0x20000000
};

enum {
    METATYPE_JSON = 1,
};

enum {
    WS_BINARY = 2,
};

void *realloc(void *, int);
void free(void *);
]]

local C_realloc = C.realloc
local C_free = C.free

local ss = class('stream_server')

local jsonrpc_version = '2.0'

local api_version_json = cjson.encode({
    method = 'apiVersion',
    params = {'1.0'},
})

-- Function used to send short metadata packets on a websocket
function ss:_send_short_metadata(ws, pkt, id, data)
--print(id, data)
    local length = #data
    pkt[0].hdr = htonl(C.HDR_METAINFO + ((length + 4) * 1048576) + id)
    ffi_cast('unsigned int *', pkt[0].data)[0] = htonl(C.METATYPE_JSON)
    ffi_copy(pkt[0].data + 4, data, length)
    ws:send(pkt, length + 8)
end

-- Function used to send long metadata packets on a websocket
function ss:_send_long_metadata(ws, pkt, id, data)
    local length = #data
    pkt.hdr = htonl(C.HDR_METAINFO + id)
    pkt.length = htonl(length + 4)
    ffi_cast('unsigned int *', pkt.data)[0] = htonl(C.METATYPE_JSON)
    ffi_copy(pkt.data + 4, data, length)
    ws:send(pkt, length + 12)
end

-- Function used to send metadata packets on a websocket
function ss:send_metadata(ws_info, id, data)
    if #data <= 252 then
        return self:_send_short_metadata(ws_info.ws, ws_info.s_pkt, id, data)
    else
        local len = #data + ffi_sizeof('hbm_stream_packet_t') + 4
        if len > ws_info.pkt_size then
            ws_info.pkt_size = len
            ws_info.pkt = ffi_cast('hbm_stream_packet_t *',
                ffi_new('char[?]', len))
        end
        return self:_send_long_metadata(ws_info.ws, ws_info.pkt, id, data)
    end
end

-- Function used to send the initial set of metadata on a websocket when a
-- client first connects to the websocket
function ss:send_stream_metadata(ws_info)
    local pkt = ws_info.s_pkt
    local ws = ws_info.ws

    -- Send the API version metadata with stream id 0
    self:send_metadata(ws_info, 0, api_version_json)

    -- Send the "init" metadata with stream id 0
    self:send_metadata(ws_info, 0, cjson.encode({
            method = 'init',
            params = {
                streamId = ws_info.id,
                supported = {
                    alive = {timeout = 2},
--                    fill = true,
                },
                commandInterfaces = {
                    ['jsonrpc-http'] = self.http_server_info,
                },
            },
        }))

    -- Send the "available" channels metadata with stream id 0
    self:send_metadata(ws_info, 0, cjson.encode({
            method = 'available',
            params = self.channel_list,
        }))
end

-- Function called when a new websocket client connects
function ss:ws_connect(server, socket)
    local stream = tcp_stream:new(ev.Loop.default, socket)
    local ws = ws_server:new(stream)
    ws:on_close(function(...) self:ws_close(...) end)
    ws:on_handshake(function(...)

        -- Generate the "alive" metadata packet for this stream
        local alive_pkt = ffi_new('hbm_small_stream_packet_t[1]')
        local data = cjson.encode({
            method = 'alive',
        })
        local length = #data
        alive_pkt[0].hdr = htonl(C.HDR_METAINFO + ((length + 4) * 1048576))
        ffi_cast('unsigned int *', alive_pkt[0].data)[0] = htonl(C.METATYPE_JSON)
        ffi_copy(alive_pkt[0].data + 4, data, length)

        local id = string.format("WS_STREAM_%d", math.random(10000))
        local client_info = {
            ws = ws,
            subscribed_channels = {},
            pkt_size = ffi_sizeof('hbm_stream_packet_t'),
            pkt = ffi_cast('hbm_stream_packet_t *',
                ffi_new('char[?]', ffi_sizeof('hbm_stream_packet_t'))),
            s_pkt = ffi_new('hbm_small_stream_packet_t[1]'),
            id = id,
            fsts_sent = {},
        }
        self.ws_client_list[#self.ws_client_list + 1] = client_info

        -- Timer used to send "alive" metadata messages on the stream
        local hbt = ev.Timer.new(function(l, t, e)
                ws:send(alive_pkt, length + 8)
            end, 1, 1)
        hbt.info = self.ws_client_list[#self.ws_client_list]
        self.ws_client_list[#self.ws_client_list].hbt = hbt

        -- If everything is OK, send the "stream" metadata back on the stream
        self:send_stream_metadata(client_info)

        -- Start the heartbeat timer
        hbt:start(self.loop)
    end)
end

-- Function called when a websocket client disconnects
function ss:ws_close(ws)
    -- Remove the client information from the client info list
    for i, ci in ipairs(self.ws_client_list) do
        if ci.ws == ws then
            -- Make sure we "unsubscribe" from each of the channels subscribed
            -- on this connection
            for _, subscription in ipairs(ci.subscribed_channels) do
                local cinfo = self.channel_info[subscription.name]

                if cinfo then
                    cinfo.subscribers = cinfo.subscribers - 1
                    if cinfo.subscribers == 0 and cinfo.pipe ~= nil then
                        Pipe_detach_pipe(cinfo.source_pipe_list, cinfo.pipe)
                        cinfo.pipe = nil
                    end
                end
            end

            -- Stop the heartbeats for this client
            if ci.hbt then ci.hbt:stop(self.loop) end
            table.remove(self.ws_client_list, i)
            return
        end
    end
end

-- Function used to get a client list entry by stream id
function ss:get_ws_client(stream_id)
    for _, c in ipairs(self.ws_client_list) do
        if c.id == stream_id then
            return c
        end
    end
end

-- Function used to extract the stream id from the methods used to subscribe or
-- unsubscribe channels on a stream
function ss:get_stream_id(method, req_id)
    local stream_id = method:match('^(.+)%.unsubscribe$') or
        method:match('^(.+)%.subscribe$') or
        method:match('^(.+)%.want_min_max$') or
        method:match('^(.+)%.no_min_max$')
    if stream_id then
        local client = self:get_ws_client(stream_id)
        if not client then
            return nil, cjson.encode({
                jsonrpc = jsonrpc_version,
                error = {
                    message = 'Stream ID not found',
                    code = -32601,
                },
                id = req_id,
            })
        end
        return stream_id
    end
    return nil, cjson.encode({
        jsonrpc = jsonrpc_version,
        error = {
            message = 'Unable to parse stream id from method',
            code = -32603,
        },
        method = method,
        id = req_id,
    })
end

-- Mapping of CX23 test engine datatypes to HBM streaming protocol datatypes
local value_types = {
    sequential_float32 = 'real32',
    sequential_float64 = 'real64',
    message_float32 = 'real32',
    message_float64 = 'real64',

    -- NOTE:  The entries below are here for completeness but no native data
    -- source on a CX23 supplies data in any of these formats.
    int32 = 's32',
    uint32 = 'u32',
    int64 = 's64',
    uint64 = 'u64',
}

-- Somat protocol extension: Stream control function called by a client to
-- request that the test engine perform min/max value tracking of a "low
-- speed" channel and send metadata messages containing the min/max values.
-- This extension is intended for "low speed" channels ONLY but this is not
-- enforced by the server (i.e., we support it even for full-speed channels).
-- This feature is consumed by the HBM Champaign stream library and plotter.
function ss:want_min_max(json)
    return self:set_min_max(json, true)
end
function ss:no_min_max(json)
    return self:set_min_max(json, false)
end
function ss:set_min_max(json, value)
    local req_id = json.id
    local stream_id, err = self:get_stream_id(json.method, req_id)
    if not stream_id then
        return err
    end
    local failed_list
    local client = self:get_ws_client(stream_id)
    local subscribed_channels = client.subscribed_channels
    for _, sig_name in ipairs(json.params) do
        local cinfo = nil
        for _, sc in ipairs(subscribed_channels) do
            if sc.name == sig_name then
                cinfo = sc
                break
            end
        end
        if cinfo then
            cinfo.want_min_max = value
        else
            failed_list = failed_list or {}
            failed_list[#failed_list + 1] = sig_name
        end
    end
    if failed_list then
        return cjson.encode({
            jsonrpc = jsonrpc_version,
            error = {
                message = 'Invalid params',
                code = -32602,
                data = failed_list,
            },
            id = req_id,
        })
    end
    return cjson.encode({
        jsonrpc = jsonrpc_version,
        result = {},
        id = req_id,
    })
end

-- Function called to process an incoming subscribe request for a stream
-- FIXME:  Does subscribing to a channel that has already been subscribed to
-- result in an error response or success?  And, if success, is a second
-- instance provided in the stream or is the existing instance used?
function ss:subscribe(json)
    local req_id = json.id
    local stream_id, err = self:get_stream_id(json.method, req_id)
    if not stream_id then
        return err
    end
    local failed_list
    local client = self:get_ws_client(stream_id)
    local ws = client.ws
    local ws_send = ws.send
    local subscribed_channels = client.subscribed_channels
    local pkt = client.s_pkt
    for _, sig_name in ipairs(json.params) do
        -- Check to see if the channel has already been subscribe to
        local cinfo
        for s, ci in pairs(self.channel_info) do
            if s == sig_name then
                if not ci.subscribers or ci.subscribers == 0 then
                    ci.pipe = Pipe_attach_pipe(ci.source_pipe_list)
                    ci.subscribers = ci.subscribers or 0
                end
                ci.subscribers = ci.subscribers + 1
                cinfo = ci
                break
            end
        end

        if cinfo then
            -- Send the subscribe metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'subscribe',
                    params = {sig_name},
                }))

            -- Send the signal rate metadata for this channel on the websocket
            -- client.
            -- FIXME:  What should be done here for "asynchronous" channels?
            if cinfo.md.sample_rate and cinfo.md.sample_rate > 0 then
--[[
                local seconds, fractions =
                    math.modf(1.0/(cinfo.md.sample_rate / 5))
                fractions = fractions * (2^32)
                self:send_metadata(client, cinfo.slot, cjson.encode({
                        method = 'signalRate',
                        params = {
                            samples = 5,
                            delta = {
                                type = 'ntp',
                                seconds = math.floor(seconds),
                                fraction = math.floor(fractions),
                            },
                        },
                    }))
]]
                self:send_metadata(client, cinfo.slot, cjson.encode({
                        method = 'signalRate',
                        params = {
                            samples = cinfo.md.sample_rate,
                            delta = {
                                type = 'ntp',
                                seconds = 1,
                                fraction = 0,
                            },
                        },
                    }))
            end

            -- Send the signal "data" metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'data',
                    params = {
                        -- FIXME:  There is also a "TB" pattern for timestamp
                        -- per block.  That would be used for 'message_raw'.
                        pattern =
                            cinfo.md.data_type:find('^message_') and 'TV' or 'V',
                        endian = ffi.abi('le') and 'little' or 'big',
                        valueType =
                            value_types[ffi_string(cinfo.md.data_type)],
                        time = {
                            type = 'ntp',
                            size = 8,
                        },
                    },
                }))

            -- Tell the system to start sending data for this channel on the
            -- appropriate websocket client
            subscribed_channels[#subscribed_channels + 1] = {
                name = sig_name,
                cinfo = cinfo,
                want_min_max = false
            }
        else
            failed_list = failed_list or {}
            failed_list[#failed_list + 1] = sig_name
        end
    end
    if failed_list then
        return cjson.encode({
            jsonrpc = jsonrpc_version,
            error = {
                message = 'Invalid params',
                code = -32602,
                data = failed_list,
            },
            id = req_id,
        })
    end
    return cjson.encode({
        jsonrpc = jsonrpc_version,
        result = {},
        id = req_id,
    })
end

function ss:get_subscribed_channel_by_name(subscribed_channels, sig_name)
    for i, info in ipairs(subscribed_channels) do
        if info.name == sig_name then
            return subscribed_channels[i], i
        end
    end
end

-- Function called to process an incoming unsubscribe request for a stream
function ss:unsubscribe(json)
    local req_id = json.id
    local stream_id, err = self:get_stream_id(json.method, req_id)
    if not stream_id then
        return err
    end
    local failed_list
    local client = self:get_ws_client(stream_id)
    local ws = client.ws
    local ws_send = ws.send
    local pkt = client.s_pkt
    local subscribed_channels = client.subscribed_channels
    for _, sig_name in ipairs(json.params) do
        local e, i =
            self:get_subscribed_channel_by_name(subscribed_channels, sig_name)
        if e and i then
            local cinfo = self.channel_info[sig_name]

            -- Detach the pipe from the source if all have unsubscribed
            -- more clients using the channel
            cinfo.subscribers = cinfo.subscribers - 1
            if cinfo.subscribers == 0 then
                Pipe_detach_pipe(cinfo.source_pipe_list, cinfo.pipe)
                cinfo.pipe = nil
            end

            -- Send the unsubscribe metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'unsubscribe',
                }))

            -- Re-arm sending of first sample timestamp info if channel is
            -- resubscribed on this client
            client.fsts_sent[sig_name] = nil

            -- Tell the system to start sending data for this channel on the
            -- appropriate websocket client
            table.remove(subscribed_channels, i)
        else
            failed_list = failed_list or {}
            failed_list[#failed_list + 1] = sig_name
        end
    end
    if failed_list then
        return cjson.encode({
            jsonrpc = jsonrpc_version,
            error = {
                message = 'Invalid params',
                code = -32602,
                data = failed_list,
            },
            id = req_id,
        })
    end
    return cjson.encode({
        jsonrpc = jsonrpc_version,
        result = {},
        id = req_id,
    })
end

-- Handler for HTTP requests coming in from the stream control server
function ss.http_handler(loop, w, http)
    local self = w.me
    if not http.uri:match('^' .. self.http_server_info.httpPath) then
        w.streambuf:write("HTTP/1.1 404 NOT FOUND\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif http.method ~= self.http_server_info.httpMethod then
        w.streambuf:write("HTTP/1.1 400 INVALID METHOD\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not http.headers or not http.headers.content_type or
        not http.headers.content_type:find('application/json') then
        w.streambuf:write("HTTP/1.1 400 INVALID CONTENT TYPE\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not http.content then
        w.streambuf:write("HTTP/1.1 400 EMPTY CONTENT\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local cj = cjson.decode(http.content)
    if not cj then
        w.streambuf:write("HTTP/1.1 400 UNABLE TO DECODE JSON\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not cj.jsonrpc or cj.jsonrpc ~= jsonrpc_version then
        w.streambuf:write("HTTP/1.1 400 INVALID RPC VERSION\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local method = cj.method:match('^.+%.(.+)$')
    if not method or not self[method] then
        w.streambuf:write("HTTP/1.1 400 UNRECOGNIZED REQUEST\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local json = self[method](self, cj)
    w.streambuf:write(
        "HTTP/1.1 200 OK\r\n" ..
        "Content-Length: " .. #json + 2 .. "\r\n" ..
        "Content-Type:  application/json\r\n\r\n" ..
        json .. "\r\n")
    w.server.http_write_shutdown(loop, w)
end

-- Function used to add a channel to the servers list of channels available to
-- be served on the websocket connections.
function ss:add_channel(md, pipe_list)
    if not self.channel_info[md.name] then
        if md.sample_rate then md.sample_rate = tonumber(md.sample_rate) end
        self.channel_list[#self.channel_list + 1] = md.name
        self.channel_info[md.name] = {
            source_pipe_list = pipe_list,
            slot = #self.channel_list,
            md = md,
            is_async = md.data_type:find('^message_') and true
        }

        local ls_cname = md.name .. '_low_speed'
        self.channel_list[#self.channel_list + 1] = ls_cname
        self.channel_info[ls_cname] = {
            source_pipe_list = pipe_list,
            slot = #self.channel_list,
            md = md,
            is_low_speed = true,
        }
    end
end

-- Function used to send data from channels to each of the subscribed channels
-- on each of the websocket connections.
function ss:process_streams()
    -- Iterate through all of the clients
    local spl = ffi_sizeof('hbm_stream_packet_t')
    local processed_pipes = {}
    for cid, cinfo in ipairs(self.ws_client_list) do
        -- Iterate through all of the subscribed channels for the client
        -- FIXME:  Maybe make this a function to help the compiler out a bit
        for _, sig_info in ipairs(cinfo.subscribed_channels) do
            local new_frame
            -- Send all available data for this channel on the current client
            -- websocket connection.
            local chan_info = self.channel_info[sig_info.name]
            local is_async = chan_info.is_async
            local pipe = chan_info.pipe
            if pipe ~= nil then
                local frame
                local sid = chan_info.slot
                if not chan_info.is_low_speed then
                    -- Normal channels get all of the samples
                    frame = pipe.last_frame
                else
                    -- Low speed channels only get the last sample from the
                    -- last frame
                    frame = pipe.frame_list
                    if frame ~= nil then
                        new_frame = Frame_create(1, frame.element_size)
                        ffi_copy(new_frame.buf, ffi_cast('char *', frame.buf) +
                            ((frame.count - 1) * frame.element_size),
                            frame.element_size)
                        new_frame.count = 1
                        new_frame.hdr.ts = frame.hdr.ts
                        frame = new_frame
                    end
                end

                local tracked_min, tracked_max
                local have_min_max = false

                if sig_info.want_min_max then
                    if chan_info.md.data_type == 'sequential_float32' then
                        if MinMax.calculate_min_max_float32(
                                pipe, min_max_data) == 0 then
                            tracked_min = min_max_data[0].minimum
                            tracked_max = min_max_data[0].maximum
                            have_min_max = true
                        end
                    elseif chan_info.md.data_type == 'sequential_float64' then
                        if MinMax.calculate_min_max_float64(
                                pipe, min_max_data) == 0 then
                            tracked_min = min_max_data[0].minimum
                            tracked_max = min_max_data[0].maximum
                            have_min_max = true
                        end
                    end
                end

                -- Process all of the frames
                while frame ~= nil do
                    processed_pipes[sig_info.name] = pipe

                    -- If the first sample timestamp metadata has not been sent
                    -- for this channel on this client, send it
                    if not cinfo.fsts_sent[sig_info.name] and
                        tonumber(frame.hdr.ts) then
                        local fsts
                        if not is_async then
                            fsts = tonumber(frame.hdr.ts) /
                                tonumber(chan_info.md.sample_rate)
                        else
                            -- Use the timestamp from the first sample in the
                            -- frame
                            fsts = tonumber(ffi_cast('int64_t *', frame.buf)[0])
                            -- KLUDGE - 40MHz hardcoded!
                            fsts = fsts / 40000000
                        end
                        local s, f = math.modf(fsts)
                        self:send_metadata(cinfo, sid, cjson.encode({
                                method = 'time',
                                params = {
                                    stamp = {
                                        type = 'ntp',
                                        seconds = s,
                                        fraction = math.floor((f * (2^32))),
                                    },
                                },
                            }))
                        cinfo.fsts_sent[sig_info.name] = true
                    end

                    -- Send the frame if the 1st sample time stamp has been
                    -- sent
                    local ts
                    if cinfo.fsts_sent[sig_info.name] then
                        -- Async channel AND not the first frame sent
                        if is_async and not new_frame then
                            -- Make a copy of the frame
                            new_frame = Frame_create(frame.count,
                                frame.element_size)
                            ffi_copy(new_frame.buf, frame.buf,
                                frame.element_size * frame.count)
                            new_frame.count = frame.count
                            new_frame.hdr.ts = frame.hdr.ts

                            -- Convert the timestamps in the frame entries to
                            -- NTP time stamps
                            local p = ffi_cast('char *', new_frame.buf)
                            for i = 0, new_frame.count - 1 do
                                local ts =
                                    tonumber(ffi_cast('int64_t *', p)[0]) /
                                        40000000
                                local s, f = math.modf(ts)
                                local frac = math.floor((f * (2^32)))
                                ffi_cast('int32_t *', p)[0] = s
                                ffi_cast('uint32_t *', p)[1] = frac
                                p = p + frame.element_size
                            end

                            -- Send the frame on the stream
                            ts = new_frame.hdr.ts
                            local pkt = ffi_cast('hbm_stream_packet_t *',
                                new_frame.hdr)
                            pkt.hdr = htonl(C.HDR_DATA + sid)
                            pkt.length =
                                htonl(new_frame.count * new_frame.element_size)
                            cinfo.ws:send(pkt, (new_frame.count * new_frame.element_size) + spl)
                            Frame_delete(new_frame)
                            new_frame = nil
                        else
                            -- Sync channel frames just get sent
                            ts = frame.hdr.ts
                            local pkt = ffi_cast('hbm_stream_packet_t *',
                                frame.hdr)
                            pkt.hdr = htonl(C.HDR_DATA + sid)
                            pkt.length = htonl(frame.count * frame.element_size)
                            cinfo.ws:send(pkt,
                                (frame.count * frame.element_size) + spl)
                        end
                        frame.hdr.ts = ts
                    end
                    frame = frame.prev
                end

                if have_min_max then
                    self:send_metadata(cinfo, sid, cjson.encode({
                            method = 'min_max',
                            params = {
                                min = tracked_min,
                                max = tracked_max
                            },
                        }))
                end

                -- If a frame was created for the low speed channel delete it.
                if new_frame then Frame_delete(new_frame) end
            end
        end
    end

    -- Flush all of the pipes that were processed
    for _, p in pairs(processed_pipes) do
        Pipe_flush_frames(p)
    end
end

-- Function used to initialize streaming service
function ss:initialize(loop, opts)
    -- Seed the random number generator
    math.randomseed(os.time())

    opts = opts or {}
    opts.rate = opts.rate or .05

    -- Save off stuff
    self.loop = loop
    self.ws_client_list = {}

    self.channel_list = {}
    self.channel_info = {}

    -- Create an event based HTTP server for stream control
    self.http_server = http_server:new{
        ev_loop = loop,
        handler = self.http_handler,
        add_to_watchers = {
            me = self,
        }
    }

    -- FIXME?:  Should this be limited to the "external" interface only?
    self.http_w = self.http_server:listen({address = '*', port = opts.cport })
    self.http_port = self.http_w.port

    -- Save HTTP info needed when websocket clients connect
    self.http_server_info = {
        httpPath = '/stream_control',
        httpMethod = 'POST',
        httpVersion = '1.1',
        port = self.http_w.port,
    }

    self.ws_port = opts.port

    -- create a event based websocket server and start listening
    self.ws_server = tcp_server:new(ev.Loop.default, self.ws_port)
    self.ws_server:on_new_client(function(...) self:ws_connect(...) end)
    self.ws_server:start()

    -- Create a timer used to move sample data from the input pipes to the
    -- stream sockets
    self.data_writer = ev.Timer.new(function(l, t, e)
            self:process_streams()
        end, opts.rate, opts.rate):start(loop)
end

-- Terminate the websocket stuff
function ss:terminate(log, close_complete_func)
    -- For every client connected, close the connection and leave.  NOTICE
    -- this is intend for use just before using process termination
--[[
    if log then
        log:log(log.NOTICE, "Terminating HBM stream server")
    end
]]

    -- If there are no clients just return.  Call the callback function if one
    -- was given.
    if not self.ws_client_list or
        (self.ws_client_list and #self.ws_client_list == 0) then
        if close_complete_func then
            close_complete_func()
        end
        return
    end

    -- Close all of the client connections.  After the last client acknowledges
    -- the connection is closed, the callback is called (if given).
    for i, ci in ipairs(self.ws_client_list) do
        ci.ws:close()
    end
    if close_complete_func then close_complete_func() end
end

return ss
