require 'strict'

-- up_sampler.lua - Implements the Up Sampler computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local sample_clock = require 'sample_clock'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('factor=%lld', self.factor)
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
-- NOTE that output frames will be larger than input frames
-- by self.factor !!!
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    local out_frame_size =
        frame.bufsize(self.sample_rate, self.pipe_rate)

    -- Process each frame in input pipe.  There is only one input pipe.
    local in_count     -- length of data buf in current input frame
    local sample_type  -- C type string for frame's buf
    local in_vals      -- "pointer" to current input frame's buf
    local inbuf_index  -- sample index for current input frame
    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        in_count = f.count
        sample_type = utils.frame_data_type(f.element_size)
        in_vals = ffi_cast(sample_type, f.buf)
        inbuf_index = 0

        -- Need to output 'self.factor' frames, keeping the
        -- output frames the same size as the input frames.
        -- NOTE:  "frame rate" is determined by proxy node
        --   and all modules can assume the input frame size
        --   is correct for the "frame rate".
        while inbuf_index < in_count do
            if self.out_index == 0 then
                -- create frame for output
                self.current_out_frame =
                    Frame_create(out_frame_size, f.element_size)
                self.current_out_frame.hdr.ts = self.tot_count
                self.current_out_frame.ts = math.floor(self.tot_count)
                self.out_vals =
                    ffi_cast(sample_type, self.current_out_frame.buf)
            end
            for j = self.repeat_cnt, self.factor - 1 do
                self.out_vals[self.out_index] = in_vals[inbuf_index]
                self.out_index = self.out_index + 1
                self.repeat_cnt = self.repeat_cnt + 1
                if self.out_index >= out_frame_size then
                    break
                end
            end
            if self.repeat_cnt >= self.factor then
                self.repeat_cnt = 0
                inbuf_index = inbuf_index + 1
            end
            if self.out_index >= out_frame_size then
                self.current_out_frame.count = out_frame_size
                self.tot_count = self.tot_count + self.current_out_frame.count
                -- put output frame in output pipe(s)
                if not utils.put_frame(self.current_out_frame, self.pipe_list) then
                    Frame_delete(self.current_out_frame)
                end
                self.out_index = 0
                self.current_out_frame = nil
            end
        end

        -- delete input frame, get next one
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local up_sampler = setmetatable({}, {__index = mt})
    up_sampler.sie_collect_idx = first_sie_index
    utils.set_common_meta(up_sampler, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Up Sampler computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    up_sampler.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(up_sampler.pipe_rate and up_sampler.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(up_sampler.pipe_rate))
    --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    up_sampler.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(up_sampler.sample_rate and up_sampler.sample_rate > 0,
        "Invalid sample rate: " .. tostring(up_sampler.sample_rate))
    --   up sampler 'factor'
    up_sampler.factor = tonumber(stanza.factor)
    assert(up_sampler.factor and up_sampler.factor >= 1,
        "INVALID up sample factor: " .. tostring(up_sampler.factor))

    -- create input/output pipe structures
    up_sampler.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(up_sampler.in_pipe and up_sampler.in_pipe ~= nil,
        "failed to create input pipe")
    up_sampler.pipe_list = Pipe_new_pipe_list()
    assert(up_sampler.pipe_list and up_sampler.pipe_list ~= nil,
        "failed to create (output) pipe list")

    -- current output frame values
    up_sampler.current_out_frame = nil
    up_sampler.out_vals = nil
    up_sampler.out_index = 0
    -- nbr times current input sample has been used
    up_sampler.repeat_cnt = 0
    -- used for frames' time stamps (just a sample count for
    -- this module)
    up_sampler.tot_count = 0

    return up_sampler, 1
end

return m
